/******************************************************************************
*******************************************************************************
*               European Southern Observatory
*             VLTI MIDI Data Reduction Software
*
* Module name:  geometry.c
* Description:  Contains routines for all geometrical routines
*
* History:
* 14-Nov-03     (csabet) Created
*******************************************************************************
******************************************************************************/

/******************************************************************************
*   Compiler directives
******************************************************************************/

/******************************************************************************
*   Include files
******************************************************************************/
#include <math.h>
#include "geometry.h"
#include <stdio.h>
#include <cpl.h>
#include "midiGlobal.h"
#include "midiLib.h"

/**********************************************************
*   Constant definitions
**********************************************************/

/**********************************************************
*   Global Variables
**********************************************************/

/*============================ C O D E    A R E A ===========================*/


/******************************************************************************
*               European Southern Observatory
*            VLTI MIDI Data Reduction Software
*
* Module name:  computeUVW
* Input/Output: See function arguments to avoid duplication
* Description:  Computes the uvw coordinates. The base algorithm is taken from VINCI
*
* History:
* 13-Nov-03     (csabet) Created
******************************************************************************/
void computeUVW (
    double  t1x,    /*  In: Station coordinates relative to the array centre */
    double  t1y,    /*  In: Station coordinates relative to the array centre */
    double  t1z,    /*  In: Station coordinates relative to the array centre */
    double  t2x,    /*  In: Station coordinates relative to the array centre */
    double  t2y,    /*  In: Station coordinates relative to the array centre */
    double  t2z,    /*  In: Station coordinates relative to the array centre */
    double  time,   /*  In: time */
    double  RA,     /*  In: Right Ascension */
    double  Dec,    /*  In: Declanation */
    UVW     *uvw)   /*  Ou: uvw Coordinates */

{

    /*  Local Declarations
    --------------------*/
    const char  routine[] = "computeUVW";
    double         GST, LST, T0, T, UTdate, UTtime, julianDate, hourAngle, dx, dy, dz,
                ihar, decr, cha, sha, cdec, sdec, clat, slat;
    
    /*  Algorithm
    -----------*/
    if (diagnostic > 4)cpl_msg_info(cpl_func,"Invoking      routine   '%s' \n", routine);
    if (diagnostic > 4) fprintf(midiReportPtr, "Invoking      routine   '%s' \n", routine);

/*  TO_DO_MIDI_DRS  I am not sure about the "time" used in the following statement. csabet 14-Nov-03 */
/*                    In VINCI it is given by (julianDate = Batch_on_processed->scan_data[I].time + 2400000.5); */
/*                  Here I am using the UTC derived from the primary header of the FITS file */
//    csabet 29-Sep-04. We have decided to use MJD-OBS from the primary header

    julianDate = time + 2400000.5;

    if ((julianDate - floor (julianDate)) > 0.5)
        UTdate = floor (julianDate) + 0.5;
    else 
        UTdate = floor (julianDate) - 0.5;

    /*  Greenwich Mean Time (UTC) in decimal hours */
    UTtime = 24.0 * (julianDate - UTdate);

    T = (UTdate - 2451545.0) / 36525.0;
    T0 = 6.697374558 + (2400.051336 * T) + (0.000025862 * T * T) + (UTtime * 1.0027379093);
    GST = 24 * (T0 / 24.0 - floor (T0 / 24.0));
    
    /*  LST = local_sidereal */
    LST = GST - LONGITUDE_PARANAL * (12.0 / MIDI_PI);
    if (LST < 0) LST = LST + 24.0;
    LST = 24 * (LST / 24.0 - floor (LST / 24.0));

    /*  Beware: RA and Dec are in degrees ! */
    hourAngle = LST - RA * 12.0 / 180.0;

    dx = t2x - t1x;
    dy = t2y - t1y;
    dz = t2z - t1z;

    ihar = hourAngle * MIDI_PI / 12.0;
    decr = Dec * MIDI_PI / 180.0;
    cha = cos (ihar);
    sha = sin (ihar);
    cdec = cos (decr);
    sdec = sin (decr);
    clat = cos (LATITUDE_PARANAL);
    slat = sin (LATITUDE_PARANAL);

    uvw->uCoord = dx * cha - dy * (slat * sha) + dz * (clat * sha);
    uvw->vCoord = dx * (sha * sdec) + dy * (slat * cha * sdec + clat * cdec) - dz * (clat * cha * sdec - slat * cdec);

    return;
}
/*****************************************************************************/
