/*
 * lomb_scargel.c
 *
 *  Created on: Jul 30, 2009
 *      Author: agabasch
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <math.h>
#include <cpl.h>
#include "lomb_scargel.h"
#include "midiConst.h"

/*----------------------------------------------------------------------------*/
/**
  @brief    Implementation of the Scargle-Lomb algorithm ApJ 1982, 263, 835
  @param    minf       minimum of the analized frequency
  @param    maxf       maximum of the analized frequency
  @param    filling    steps to use between maxf and minf
  @param    x          the frequency/lambda/time
  @param    y          the frequency/lambda/time value
  @return   lomb_out   the Powerspectrum
 */
/*----------------------------------------------------------------------------*/

int lomb(double minf ,double maxf, int filling, cpl_array * x, cpl_array * y, cpl_array ** lomb_out)
{
   cpl_array    * filling_factor=NULL;
   cpl_array    * frequency_=NULL;
   cpl_array    * omega_=NULL;        
   cpl_array    * significance_ =NULL;
   
   cpl_array    * tau_ =NULL;
   cpl_array    * power_=NULL;

   double       * px=NULL; 
   double       * py=NULL; 
   double       * pomega_=NULL; 
   double       * ptau_=NULL; 
   double       * ppower_=NULL; 

   double         yvar_=0.; 
   double         ymean_=0.;
   int            i=0;
   int            x_i=0;
   double         o2_i=0;
   int            dimen_x=0;      
   int            omega_i=0;
   double         sum_sin=0;
   double         sum_cos=0;
   double         y2v =0.;
   double         ym = 0.;         
   double         o_i = 0.;
   double         t_i = 0.;
   double         s_y_c=0; 
   double         s_y_s=0.;
   double         s_c2 =0.; 
   double         s_s2 =0.;
   double         c =0.; 
   double         s =0.;
   double         dy = 0.;
   double         odx = 0.;
   
   
/*    const int n_=(ilogb(1.0+(maxf / minf)) * filling); */
   const int n_=filling;
   const double spacing=(double)(maxf-minf)/filling;
   
   filling_factor =cpl_array_new(n_, CPL_TYPE_DOUBLE);
   for(i=0;i<n_;i++){
      cpl_array_set_double(filling_factor, i, minf+(spacing*i)+DBL_EPSILON);
      
   }
   
   frequency_     =cpl_array_duplicate(filling_factor);
/*    cpl_array_multiply_scalar(frequency_,minf); */
   
   omega_        =cpl_array_duplicate(frequency_);
   cpl_array_multiply_scalar(omega_, 2.0*MIDI_PI);
   
   tau_          =cpl_array_new(n_, CPL_TYPE_DOUBLE);
   power_        =cpl_array_new(n_, CPL_TYPE_DOUBLE);
   significance_ =cpl_array_new(n_, CPL_TYPE_DOUBLE);
   
   cpl_array_fill_window(tau_, 0, n_, 0.);
   cpl_array_fill_window(power_, 0, n_, 0.);
   cpl_array_fill_window(significance_, 0, n_, 0.);
    
   px       =cpl_array_get_data_double(x);
   pomega_  =cpl_array_get_data_double(omega_);
   ptau_    =cpl_array_get_data_double(tau_);
   ppower_  =cpl_array_get_data_double(power_);
   
   dimen_x=cpl_array_get_size(x);
   
   sum_sin=0;
   sum_cos=0;
   for(omega_i=0;omega_i<n_;omega_i++){
      o2_i = (*pomega_) * 2.0;
      
      for(i=0;i<dimen_x;i++){
         sum_sin+=(sin(o2_i * (*px)));
         sum_cos+=(cos(o2_i * (*px)));
      }
      
      *ptau_ = atan( sum_sin / sum_cos ) / o2_i; 
      ++ptau_;
      ++pomega_;
      
   }
   
   ymean_=cpl_array_get_mean(y);
   yvar_=pow(cpl_array_get_stdev(y),2);
   y2v = 2.0 * yvar_;
   ym = ymean_;         

/*    Resetting to pointer to the first address */
   pomega_  =cpl_array_get_data_double(omega_);
   ptau_    =cpl_array_get_data_double(tau_);
   ppower_  =cpl_array_get_data_double(power_);
   
   px       =cpl_array_get_data_double(x);
   py       =cpl_array_get_data_double(y);

   for(omega_i=0;omega_i<n_;omega_i++){
      o_i = *pomega_;
      t_i = *ptau_;
      s_y_c=0; 
      s_y_s=0.;
      s_c2 =0.; 
      s_s2 =0.;
/*    Resetting to pointer to the first address */
      px       =cpl_array_get_data_double(x);
      py       =cpl_array_get_data_double(y);
      
      for(x_i=0;x_i<dimen_x;x_i++){
         dy = *py - ym;
         odx = (*px - t_i) * o_i;
         
         s = sin(odx);
         s_s2 += (s*s);
         s_y_s += dy * s;
         
         c = cos(odx);
         s_c2 += (c*c);
         s_y_c += dy * c;

         ++px;
         ++py;
      }

      *ppower_ = ( (s_y_c*s_y_c) / s_c2 +
                   (s_y_s*s_y_s) / s_s2 ) / y2v;
      ++ppower_;
      ++ptau_;
      ++pomega_;
      
   }
   
   
/*    cpl_msg_info(cpl_func,"n_: %d", n_); */
/*    cpl_msg_info(cpl_func,"%g %g %g",s_c2,s_s2,y2v); */

/*    for(i=0;i<n_;i++){ */
/*       cpl_msg_info(cpl_func,"filling_factor, frequency, omega, tau, POWER: %g, %g, %g, %g, %g",  */
/*                    cpl_array_get_double(filling_factor,i,NULL), */
/*                    cpl_array_get_double(frequency_,i,NULL), */
/*                    cpl_array_get_double(omega_,i,NULL), */
/*                    cpl_array_get_double(tau_,i,NULL), */
/*                    cpl_array_get_double(power_,i,NULL)); */
      
/*    } */

/*    for(i=0;i<n_;i++){ */
/*       cpl_msg_info(cpl_func,"frequency: %g  POWER: %g",  */

/*                    cpl_array_get_double(frequency_,i,NULL), */
/*                    cpl_array_get_double(power_,i,NULL)); */
      
/*    } */
   


   lomb_out[0]=frequency_;
   lomb_out[1]=power_;

   cpl_array_delete(filling_factor);
   cpl_array_delete(omega_);
   cpl_array_delete(tau_);
   cpl_array_delete(significance_);
   
   

   return 0;
}
