/******************************************************************************
*******************************************************************************
*               European Southern Observatory
*             VLTI MIDI Data Reduction Software
*
* Module name:  midiControl.c
* Description:  Contains the main program and routines for data reduction
*
* History:
* 10-Jan-05     (csabet) Created
*******************************************************************************
******************************************************************************/

/******************************************************************************
*   Compiler directives
******************************************************************************/

/******************************************************************************
*   Include files
******************************************************************************/

#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "cpl_init.h"
#include "midiGlobal.h"
#include "midiLib.h"
#include "midiControl.h"
#include "memoryHandling.h"
#include "procFrgHS.h"
#include "procFrgSP.h"
#include "procKappa.h"
#include "procAcq.h"
#include "procDetLin.h"
#include "procDetRon.h"
#include "procRefPix.h"
#include "procDspTrn.h"
#include "procWavCal.h"
#include "errorHandling.h"
#include "fileHandling.h"
#include "midiFitsClassification.h"
#include "calibration.h"
#include "diagnostics.h"
#include <cpl.h>
#include "midi_dfs.h"

/**********************************************************
*   Global Variables
**********************************************************/
int             rootName = 1;                                           // Output root name is given
int             diagnostic = 0;                                         // Set by command line argument DEBUG   (argumentFile.lst)
int             endOfTemplate = 0;                                      // Set by command line argument EOT   (argumentFile.lst)
int             plotFile = 2;                                           // Set by command line argument PLOT_FILE (argumentFile.lst)
int             plotDuration = 0;                                       // Set by command line argument PLOT_FILE (argumentFile.lst)
int             batchNumber = 0;                                        // It is used everywhere
FILE    *midiReportPtr=NULL;                            // File pointer for the program execution history
char    *midiReportName=NULL;                           // Name of event log file
FILE    *midiQcLogPtr=NULL;                                     // File pointer for the QC log file
char    *midiQcLogName=NULL;                            // Name of QC log file
char    outFileDir[MAX_STRING_LENGTH];          // Product directory name
char    outRootName[MAX_STRING_LENGTH];         // Root name for all output files
char    batchTemplate[MAX_STRING_LENGTH];       // A consistent Batch Template Name
char    givenName[MAX_STRING_LENGTH];           // User given  name for all output files
char    midiMessage[MAX_STRING_LENGTH];         // A global message holder
int             *badChannelList;                                        // An array of numOfChannels

/*============================ C O D E    A R E A ===========================*/


/******************************************************************************
*               European Southern Observatory
*             VLTI MIDI Data Reduction Software
*
* Module name:  executeDataReduction
* Input/Output: See function arguments to avoid duplication
* Description:  This is the main routine for the MIDI Data Reduction. It shall be
*                               invoked by one of the recipes. This rootine and the way user parameters
*                               are processed is not at the moment efficient. It has been adapted from the
*                               stand-alone version of the MIDI quickly to get the recipes running.
*
* History:
* 09-Jan-07     (csabet) Created
******************************************************************************/
void executeDataReduction (
   const char        *calibTempDir,  // In: CalibDb directory (full path-name)
   const char        *maskFileDir,   // In: Mask file directory (full path-name)
   const char        *productDir,    // In: Name of the product directory
   int               givenDispDur,   // In: Given plot duration
   const char        *inFitsList,    // In: Name of file containing input FITS files (full path-name)
   int               *error,         // Ou: Error status
   cpl_parameterlist *parlist,
   cpl_frameset      *frameset)
{
   
   //  Local Declarations
    //  ------------------
    const char  routine[] = "executeDataReduction";
    FILE                *inFitsClassifiedPtr=NULL, *inFitsBatchPtr=NULL, *tempReportPtr=NULL;
    char                *stringTemp=NULL, *tempReportName=NULL, *fileNameStr=NULL,
                                *classification=NULL;
    MidiFiles   *fileNames=NULL;
    UserOptions *options=NULL;
    FILE                *signaturePtr=NULL;

    //  Algorithm
    //  ---------
//      cpl_init ();
        *error = 0;

    //  Allocate memory
    tempReportName = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
    options = (UserOptions *) calloc (1, sizeof (UserOptions));
    fileNames = callocMidiFiles (MAX_STRING_LENGTH);

    // Initialisation
    sprintf (tempReportName, "midiPipelineTempReport.log");
    tempReportPtr = fopen (tempReportName, "w");
    sprintf (outFileDir, "%s", productDir);
        plotDuration = givenDispDur;
        if (plotDuration == 0) plotFile = 1;
        sprintf (givenName, "%s", "MIDI");
//      cpl_msg_info(cpl_func,"Name of the FITS list = %s \n", inFitsList);
        

        getMaskFileFromSof(fileNames,frameset);
        
        


    getParameters (calibTempDir, maskFileDir, inFitsList, tempReportPtr, options, fileNames, error);
        if (*error)
        {
                cpl_msg_info(cpl_func,"\n");
                fprintf (tempReportPtr, "\n");
                fclose (tempReportPtr);
                free (tempReportName);
                free (options);
                free (fileNames);
                return;
        }



    //  Allocate memory
    stringTemp = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
    midiQcLogName = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
    midiReportName = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
    fileNameStr = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
    classification = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));

    //  Classify FITS files
    //  -------------------
    classifyFitsFiles (options, fileNames, tempReportPtr, error,
    		parlist, frameset);
        if (*error)
        {
                if (*error > 1) 
                {
                        sprintf (midiMessage, "No FITS files to process. Check log file: %s", tempReportName);
                        midiReportWarning (tempReportPtr, routine, __FILE__, __LINE__, midiMessage);
                }
                cpl_msg_info(cpl_func,"\n");
                fprintf (tempReportPtr, "\n");
                fclose (tempReportPtr);
                freeMidiFiles (fileNames);
                free (options);
                free (stringTemp);
                free (tempReportName);
                free (midiQcLogName);
                free (midiReportName);
                free (fileNameStr);
                free (classification);
                return;
        }

    /*====================================================================================
    At this point a file containing the names of input fits files should be available. Here
    the following assumptions are made:

        1.  Interferometry FITS files may be split into more than one file
        2.  Data Organizer's list of files are consecutive in types. Mixed split files are 
                        not allowed
        3.  All file categories are processed separately and thier results are stored in 
                        either memory or relevant output FITS files.
        4.  A Process is only invoked when necessary data is available
        5.  Photometry FITS files can also be split. However the number of Photometry A files
                        must be equal to the number of Photometry B files.
        6.  The order of Photometry and Interferometry in the above list is not important. 
    ====================================================================================*/

    //  Open the classified list of FITS files for reading
    if ((inFitsClassifiedPtr = fopen (fileNames->inFitsClassified, "r")) == NULL)
        midiReportError (tempReportPtr, routine, __FILE__, __LINE__, "Cannot open list of classified input FITS files");

    //  Loop through the list of classified FITS files and process 
    //  each batch. The first record  in the file is the name of the template
    while (fgets (stringTemp, MAX_STRING_LENGTH, inFitsClassifiedPtr) != NULL)
    {
                //      Extract cleaned up batchTemplate string and save it
                removeNewLine (stringTemp, batchTemplate);

        //      Create a batch file
        sprintf (fileNames->inFitsBatch, "%s", "inputFitsBatch.log");
        if ((inFitsBatchPtr = fopen (fileNames->inFitsBatch, "w")) == NULL)
            midiReportError (tempReportPtr, routine, __FILE__, __LINE__, "Cannot open batch FITS files");

        //      Copy the rest of files from the classified-list into the batch-list until the END_OF_BATCH
        while (fgets (stringTemp, MAX_STRING_LENGTH, inFitsClassifiedPtr) != NULL)
        {
            sprintf (classification, "%s", "");
            sscanf (stringTemp, "%s%s", fileNameStr, classification);
            if (strcmp (fileNameStr, "END_OF_BATCH") != 0)
                fprintf (inFitsBatchPtr, "%s %s\n", fileNameStr, classification);
            else
                break;
        }
        fclose (inFitsBatchPtr);

        //      Increment batch counter
                batchNumber=1;
                        
                //      Create output file names
        createOutputFileNames (fileNames, tempReportPtr, error);
                if (*error)
                {
                        midiReportWarning (tempReportPtr, routine, __FILE__, __LINE__, "This batch is rejected");
                        continue;   //  Abandon the current batch and get the next batch
                }
        handoverReport (tempReportName, &tempReportPtr);

        //      Create Mask FITS file
        //      ---------------------
        if (options->maskMode == COMPUTED)
        {
            createMaskFile (fileNames, error);
            if (*error)
            {
                                midiReportWarning (tempReportPtr, routine, __FILE__, __LINE__, "This batch is rejected");
                midiRejectBatch ();
                continue;   //  Abandon the current batch and get the next batch
            }
        }

                //      Process templates
                //      -----------------
                if ((strcmp (batchTemplate, "HIGH_SENS_CALIB") == 0) ||
                        (strcmp (batchTemplate, "HIGH_SENS_SCIENCE") == 0))
                {
                   if (options->drsFrg) procFrgHS (fileNames, options, error, parlist, frameset);
                        if (*error) midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "This batch is rejected");
                }
                else if (strcmp (batchTemplate, "SCI_PHOT_KAPPA") == 0)
                {
                        if (options->drsKap)
                        {
                                //      Write a signature. Midas checks to see which template is running
                                signaturePtr = fopen ("MIDI_sig_spk.log", "w");
                            fclose (signaturePtr);
                                procKappa (fileNames, options, error);
                                if (*error) midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "This batch is rejected");
                        }
                }
                else if ((strcmp (batchTemplate, "SCI_PHOT_CALIB") == 0) ||
                        (strcmp (batchTemplate, "SCI_PHOT_SCIENCE") == 0))
                {
                        if (options->drsFrg)
                        {
                                if (strcmp (batchTemplate, "SCI_PHOT_CALIB") == 0)
                                    signaturePtr = fopen ("MIDI_sig_spc.log", "w");
                                else
                                        signaturePtr = fopen ("MIDI_sig_sps.log", "w");
                                fclose (signaturePtr);
                                procFrgSP (fileNames, options, error, parlist, frameset);
                                if (*error) midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "This batch is rejected");
                        }
                }
                else if (strcmp (batchTemplate, "ACQ") == 0)
                {
                        if (options->drsAcq) 
                                procAcq (fileNames, error);
                        if (*error) midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, 
                                "This batch is rejected");
                }
                else if (strcmp (batchTemplate, "DETLIN") == 0)
                {
                        if (options->tecLin) 
                                procDetLin (fileNames, error);
                        if (*error) midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, 
                                "This batch is rejected");
                }
                else if (strcmp (batchTemplate, "DETRON") == 0)
                {
                        if (options->tecRon) 
                                procDetRon (fileNames, error);
                        if (*error) midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, 
                                "This batch is rejected");
                }
                else if (strcmp (batchTemplate, "DSPTRN") == 0)
                {
                        if (options->tecTrn) 
                                procDspTrn (fileNames, error);
                        if (*error) midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, 
                                "This batch is rejected");
                }
                else if (strcmp (batchTemplate, "REFPIX") == 0)
                {
                        if (options->tecPix) 
                                procRefPix (fileNames, error);
                        if (*error) midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, 
                                "This batch is rejected");
                }
                else if (strcmp (batchTemplate, "WAVECAL") == 0)
                {
                        if (options->tecWav) 
                                procWaveCal (options->tecWav, fileNames, error);
                        if (*error) midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, 
                                "This batch is rejected");
                }
                else 
                {
                        sprintf (midiMessage, "Cannot process template: %s", batchTemplate);
                        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, midiMessage);
                        *error = 1;
                }

                //      Close log files for this batch
                if (*error) midiRejectBatch ();
                else midiCloseBatch ();
                
                if (strcmp (batchTemplate, "SCI_PHOT_KAPPA") == 0)
                {
                   /* The empty dummy file MIDI_b1_spk.qc.dummy has to be erased  */ 
                   remove (midiQcLogName);
                   
                   continue;
                }
                   break;

   }
 
    //  Close temporary files and release memory
    //  ----------------------------------------
    if (diagnostic) cpl_msg_info(cpl_func,"\nClosing temporary files \n");
    if (diagnostic) cpl_msg_info(cpl_func,"----------------------- \n");
    if (diagnostic) cpl_msg_info(cpl_func,"\n\n");
    fclose (inFitsClassifiedPtr);
    fclose (tempReportPtr);
    remove (fileNames->inFitsClassified);
    remove (fileNames->inFitsBatch);
    remove (tempReportName);
 
    //  Release remaining memory
    freeMidiFiles (fileNames);
    free (options);
    free (stringTemp);
    free (tempReportName);
    free (midiQcLogName);
        free (midiReportName);
    free (fileNameStr);
    free (classification);
//      cpl_end ();
        
    return;
}
/*****************************************************************************/


/******************************************************************************
*               European Southern Observatory
*             VLTI MIDI Data Reduction Software
*
* Module name:  midiRejectBatch
* Input/Output: See function arguments to avoid duplication
* Description:  Garbage collector
*
* History:
* 17-May-05     (csabet) Created
******************************************************************************/
void midiRejectBatch (void)
{
    //  Local Declarations
    //  ------------------
    const char  routine[] = "midiRejectBatch";

    //  Algorithm
    //  ---------
    if (diagnostic > 2) cpl_msg_info(cpl_func,"Invoking      routine   '%s' \n", routine);
    if (diagnostic > 2) fprintf (midiReportPtr, "Invoking      routine   '%s' \n", routine);

        //      Close log File and release memory for this batch
        //      ------------------------------------------------
        if (diagnostic) cpl_msg_info(cpl_func,"Created Log files: %s \n", midiReportName);
        fprintf (midiReportPtr, "Created Log files: %s \n", midiReportName);
        if (diagnostic) cpl_msg_info(cpl_func,"\n");
        fprintf (midiReportPtr, "\n");
        if (midiReportPtr)
        {
                fclose (midiReportPtr); 
                midiReportPtr = NULL;
        }
        if (midiQcLogPtr)
        {
                fclose (midiQcLogPtr); 
                midiQcLogPtr = NULL;
                remove (midiQcLogName);
        }

        return;
}
/*****************************************************************************/


/******************************************************************************
*               European Southern Observatory
*             VLTI MIDI Data Reduction Software
*
* Module name:  midiCloseBatch
* Input/Output: See function arguments to avoid duplication
* Description:  Garbage collector
*
* History:
* 17-May-05     (csabet) Created
******************************************************************************/
void midiCloseBatch (void)
{
    //  Local Declarations
    //  ------------------
    const char  routine[] = "midiCloseBatch";

    //  Algorithm
    //  ---------
    if (diagnostic > 2) cpl_msg_info(cpl_func,"Invoking      routine   '%s' \n", routine);
    if (diagnostic > 2) fprintf (midiReportPtr, "Invoking      routine   '%s' \n", routine);

        //      Close log File and release memory for this batch
        //      Close files used in this batch
        cpl_msg_info(cpl_func,"Created Log files: %s  %s \n", midiReportName, midiQcLogName);
        fprintf (midiReportPtr, "Created Log files: %s  %s \n", midiReportName, midiQcLogName);
        cpl_msg_info(cpl_func,"\n\n");
        fprintf (midiReportPtr, "\n\n");
        if (midiReportPtr)
        {
                fclose (midiReportPtr); 
                midiReportPtr = NULL;
        }
        if (midiQcLogPtr)
        {
                fclose (midiQcLogPtr); 
                midiQcLogPtr = NULL;
        }

        return;
}
/*****************************************************************************/


/******************************************************************************
*               European Southern Observatory
*             VLTI MIDI Data Reduction Software
*
* Module name:  getparameters
* Input/Output: See function arguments to avoid duplication
* Description:  Gets user options
*
* History:
* 24-Sep_03     (csabet) Removed OPDstep. It shall be derived from the FITS header
* 03-Mar-03     (csabet) Created
******************************************************************************/
void getParameters (
        const char            *calibTempDir,  // In: CalibDb directory
        const char            *maskFileDir,   // In: Mask file directory
        const char            *inFitsList,    // In: Name of file containing input FITS files
        FILE            *filePtr,               // In: Pointer to a temporary open file
        UserOptions     *options,               // IO: Pointer to the user options
        MidiFiles       *fileNames,             // IO: Pointer to midi file structure
        int                     *error)                 // Ou: Error status

{

    //  Local Declarations
    //  ------------------
    const char  routine[] = "getParameters";
    FILE        *inFitsListPtr=NULL, *inFitsClassifiedPtr=NULL;
    char        *argument=NULL, *value=NULL, *tempStr=NULL;

    //  Algorithm
    //  ---------
        *error = 0;

    //  Allocate memory
    tempStr = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
    argument = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
    value = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));

    //  Set default variables
    //  ---------------------
    sprintf (fileNames->inFileDir, "%s", "");
    sprintf (fileNames->trfNameRead, "%s", "/tmp/MIDI.trf.dummy");
    sprintf (fileNames->trfNameWrite, "%s", "/tmp/MIDI.trf.dummy");
    sprintf (fileNames->wclNameRead, "%s", "/tmp/MIDI.wcl.dummy");
    sprintf (fileNames->wclNameWrite, "%s", "/tmp/MIDI.wcl.dummy");
    sprintf (fileNames->trrNameRead, "%s", "/tmp/MIDI.trr.dummy");
    sprintf (fileNames->trrNameWrite, "%s", "/tmp/MIDI.trr.dummy");
    sprintf (fileNames->calibDbDir, "%s", calibTempDir);
    sprintf (fileNames->maskFileDir, "%s", maskFileDir);
    sprintf (fileNames->inFitsList, "%s", inFitsList);
        options->processing = DISPERSED;
        options->drsFrg = 1;
        options->drsAcq = 1;
        options->drsKap = 0;
        options->tecRon = 1;
        options->tecLin = 1;
        options->tecPix = 1;
        options->tecTrn = 1;
        options->tecWav = 1;
/*     options->maskMode = SELECTED; */
    options->maskMode = SPECIFIED;
        
    //  Display banner, time and date
    displayBanner (filePtr);

    if (diagnostic > 2) cpl_msg_info(cpl_func,"Invoking      routine   '%s' \n", routine);

    //  Create names for calibrator data files
    sprintf (fileNames->calibDbName, "%scalibrator_database.dat", fileNames->calibDbDir);
    sprintf (fileNames->waveCalibName, "%swaveCalib", fileNames->calibDbDir);
    sprintf (fileNames->transferRatiosName, "%stransferRatios", fileNames->calibDbDir);
    sprintf (fileNames->grismFoilName, "%sgrism_foil_4lines.dat", fileNames->calibDbDir);
    sprintf (fileNames->prismFoilName, "%sprism_foil_4lines.dat", fileNames->calibDbDir);

    //  Create a name for the classified file
    strcpy (fileNames->inFitsClassified, "inputFitsClassified.log");

//      cpl_msg_info(cpl_func,"Name of the FITS list = %s \n", fileNames->inFitsList);
    //  Check if the input FITS file list exists and it is not empty
    if (((inFitsListPtr = fopen(fileNames->inFitsList, "r")) != NULL) &&
        (fscanf (inFitsListPtr, "%s\n", argument) != EOF))
    {
        //      Create a file for the classified list
        rewind (inFitsListPtr);
        inFitsClassifiedPtr = fopen (fileNames->inFitsClassified, "w");

        while (fgets (tempStr, MAX_STRING_LENGTH, inFitsListPtr) != NULL)
        {
            //  Clear string
            sprintf (value, "%s", "");
            sscanf (tempStr, "%s%s", argument, value);
            fprintf (inFitsClassifiedPtr, "%s %s\n", argument, value);
        }
        fclose (inFitsListPtr);
        fclose (inFitsClassifiedPtr);
    }
    else
    {
       cpl_msg_info(cpl_func,"\n<<Error>>: List of FITS file is not given or the list is empty \n");
       cpl_msg_info(cpl_func,"\n");
        fprintf(filePtr, "\n<<Error>>: List of FITS file is not given or the list is empty\n");
        fprintf(filePtr, "\n");
                *error = 1;
        if (argument) free (argument);
        if (tempStr) free (tempStr);
        if (value) free (value);
                return;
    }

    //  Report the contents of the list
    inFitsClassifiedPtr = fopen (fileNames->inFitsClassified, "r");
    cpl_msg_info(cpl_func,"\n");
    cpl_msg_info(cpl_func,"The FITS list contains the following files:\n");
    cpl_msg_info(cpl_func,"-------------------------------------------\n");
    fprintf (filePtr, "\n");
    fprintf (filePtr, "Given FITS list contains the following files:\n");
    fprintf (filePtr, "---------------------------------------------\n");
    while (fgets (tempStr, MAX_STRING_LENGTH, inFitsClassifiedPtr) != NULL)
    {
        //      Clear string
        sprintf (value, "%s", "");
        sscanf (tempStr, "%s%s", argument, value);
        cpl_msg_info(cpl_func,"%s %s\n", argument, value);
        fprintf (filePtr, "%s %s\n", argument, value);
    }
    fclose (inFitsClassifiedPtr);

    //  Release memory
    if (argument) free (argument);
    if (tempStr) free (tempStr);
    if (value) free (value);

    return;
}
/******************************************************************************/




/******************************************************************************
*               European Southern Observatory
*             VLTI MIDI Data Reduction Software
*
* Module name:  createOutputFileNames
* Input/Output: See function arguments to avoid duplication
* Description:  Creates names for the product FITS files
*
* History:
* 03-Mar-03     (csabet) Created
******************************************************************************/
void createOutputFileNames (
    MidiFiles   *fileNames,     // IO: Pointer to midi files structure
    FILE        *filePtr,
    int                 *error)

{

    //  Local Declarations
    //  ------------------
    const char      routine[] = "createOutputFileNames";
    FILE            *inFitsBatchPtr=NULL;
    size_t          stringLength;
    char                *fileTemp=NULL, *classification=NULL, *pipeFilePrefix=NULL, 
                                        *fileNameStr=NULL, *suffix=NULL, *pipeRootName=NULL;

    //  Algorithm
    //  ---------
    if (diagnostic > 2) cpl_msg_info(cpl_func,"Invoking      routine   '%s' \n", routine);
    if (diagnostic > 2) fprintf (filePtr, "Invoking      routine   '%s' \n", routine);
    
    //  Reset parameters
    *error = 0;
    
    //  Allocate memory
    fileTemp = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
    suffix = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
    fileNameStr = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
    classification = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
        pipeFilePrefix = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
        pipeRootName = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
        
        //      Create the prefix for the PIPEFILE
    sprintf (pipeFilePrefix, "r.");

        //      Get the pipeline root name (pipeRootName) from the first FITS file in the batch
        if ((inFitsBatchPtr = fopen(fileNames->inFitsBatch, "r")) == NULL)
                midiReportError (filePtr, routine, __FILE__, __LINE__, "Cannot open FITS file list");

        fgets (fileTemp, MAX_STRING_LENGTH, inFitsBatchPtr);
        sprintf (classification, "%s", "");
        sscanf (fileTemp, "%s%s", fileNameStr, classification);
        stringLength = strlen (fileNameStr);
        strncpy (pipeRootName, fileNameStr, stringLength-5);

        //      Strip off the path name if any and Create PIPEFILE and 
        //      ARCHFILE. Always from the root name of the first FITS file 
        removePathName (pipeRootName, filePtr);
        if (inFitsBatchPtr) fclose (inFitsBatchPtr);
        sprintf (fileNames->archFileName, "%s.fits", pipeRootName);
        sprintf (fileNames->pipeFileName, "%s%s_tpl_0000.fits", pipeFilePrefix, pipeRootName);

        //      Create template suffix
        if (strcmp (batchTemplate, "HIGH_SENS_CALIB") == 0)
        {
                sprintf (suffix, "_b%d_hsc", batchNumber);
        }
        else if (strcmp (batchTemplate, "HIGH_SENS_SCIENCE") == 0)
        {
                sprintf (suffix, "_b%d_hss", batchNumber);
        }
        else if (strcmp (batchTemplate, "SCI_PHOT_CALIB") == 0)
        {
                sprintf (suffix, "_b%d_spc", batchNumber);
        }
        else if (strcmp (batchTemplate, "SCI_PHOT_SCIENCE") == 0)
        {
                sprintf (suffix, "_b%d_sps", batchNumber);
        }
        else if (strcmp (batchTemplate, "SCI_PHOT_KAPPA") == 0)
        {
                sprintf (suffix, "_b%d_spk", batchNumber);
        }
        else if (strcmp (batchTemplate, "ACQ") == 0)
        {
                sprintf (suffix, "_b%d_acq", batchNumber);
        }
        else if (strcmp (batchTemplate, "DETLIN") == 0)
        {
                sprintf (suffix, "_b%d_lin", batchNumber);
        }
        else if (strcmp (batchTemplate, "DETRON") == 0)
        {
                sprintf (suffix, "_b%d_ron", batchNumber);
        }
        else if (strcmp (batchTemplate, "DSPTRN") == 0)
        {
                sprintf (suffix, "_b%d_trn", batchNumber);
        }
        else if (strcmp (batchTemplate, "REFPIX") == 0)
        {
                sprintf (suffix, "_b%d_pix", batchNumber);
        }
        else if (strcmp (batchTemplate, "WAVECAL") == 0)
        {
                sprintf (suffix, "_b%d_wav", batchNumber);
        }
        else
        {
                *error = 1;
                sprintf (midiMessage, "Unknown Template: %s", batchTemplate);
                midiReportWarning (filePtr, routine, __FILE__, __LINE__, midiMessage);
                free (fileTemp);
                free (suffix);
                free (classification);
                free (fileNameStr);
                free (pipeFilePrefix);
                free (pipeRootName);
                return;
        }
        if (!rootName) sprintf (outRootName, "%s%s", pipeRootName, suffix);
        else sprintf (outRootName, "%s%s", givenName, suffix);
        
        sprintf (midiReportName, "%s%s.log", outFileDir, outRootName);
        sprintf (midiQcLogName, "%s%s.qc.dummy", outFileDir, outRootName);
        sprintf (fileNames->trfHistoryName, "%s%s.trfh.dummy", outFileDir, outRootName);
        sprintf (fileNames->wclHistoryName, "%s%s.wclh.dummy", outFileDir, outRootName);
        sprintf (fileNames->trrHistoryName, "%s%s.trrh.dummy", outFileDir, outRootName);
        sprintf (fileNames->outFitsName, "%s%s.pro.fits", outFileDir, outRootName);
        sprintf (fileNames->outWaterfallName, "%s%s.waf.fits", outFileDir, outRootName);
        sprintf (fileNames->outWaterpowerName, "%s%s.wap.dat", outFileDir, outRootName);
        sprintf (fileNames->outAveImageName, "%s%s.aim.fits", outFileDir, outRootName);
        sprintf (fileNames->outChSpectrumName, "%s%s.chs.fits", outFileDir, outRootName);
                
    //  Release memory
    if (fileTemp) free (fileTemp);
    if (suffix) free (suffix);
    if (classification) free (classification);
    if (fileNameStr) free (fileNameStr);
        if (pipeFilePrefix) free (pipeFilePrefix);
        if (pipeRootName) free (pipeRootName);

    return;
}
/******************************************************************************/



/******************************************************************************
*               European Southern Observatory
*             VLTI MIDI Data Reduction Software
*
* Module name:  displayBanner
* Input/Output: See function arguments to avoid duplication
* Description:  Displays the ESO Banner
*
* History:
* 23-Jul-03     (csabet) Created
******************************************************************************/
void displayBanner (
    FILE    *filePtr)

{

    /*  Local Declarations
    --------------------*/
    const char  routine[] = "displayBanner";
    time_t      now;
    struct tm   *newTime=NULL;
    char        *currentTime=NULL;

    /*  Algorithm
    -----------*/
    if (diagnostic > 2) cpl_msg_info(cpl_func,"Invoking      routine   '%s' \n", routine);
    if (diagnostic > 2) fprintf (filePtr, "Invoking      routine   '%s' \n", routine);

    /*  Allocate memory */
    currentTime = (char *) calloc (MIN_STRING_LENGTH, sizeof (char));

    now = time(NULL);
    newTime = gmtime (&now);
    strftime (currentTime, MIN_STRING_LENGTH, "%a %d %b %Y at %H:%M:%S", newTime);

    cpl_msg_info(cpl_func,"\n\n\n\n");
    cpl_msg_info(cpl_func,"   ,----------------------------------------------.\n");
    cpl_msg_info(cpl_func,"   |   *       European Southern Observatory      |\n");
    cpl_msg_info(cpl_func,"   | *E S*          VLTI DMD, Garching            |\n");
    cpl_msg_info(cpl_func,"   |   O        MIDI Data Reduction Program       |\n");
    cpl_msg_info(cpl_func,"   |   *            Version: %s\n", MIDI_PIPE_VERSION);
    cpl_msg_info(cpl_func,"   |                                              |\n");
    cpl_msg_info(cpl_func,"   |               For support e-mail to:         |\n");
    cpl_msg_info(cpl_func,"   |         pballest@eso.org or agabasch@eso.org   |\n");
    cpl_msg_info(cpl_func,"   `----------------------------------------------'\n");
    cpl_msg_info(cpl_func,"\nMIDI pipeline started on: %s  GMT \n", currentTime);

    fprintf (filePtr, "\n\n\n\n");
    fprintf (filePtr, "   ,----------------------------------------------.\n");
    fprintf (filePtr, "   |   *       European Southern Observatory      |\n");
    fprintf (filePtr, "   | *E S*          VLTI DMD, Garching            |\n");
    fprintf (filePtr, "   |   O        MIDI Data Reduction Program       |\n");
    fprintf (filePtr, "   |   *            Version: %s\n", MIDI_PIPE_VERSION);
    fprintf (filePtr, "   |                                              |\n");
    fprintf (filePtr, "   |               For support e-mail to:         |\n");
    fprintf (filePtr, "   |         pballest@eso.org or agabasch@eso.org   |\n");
    fprintf (filePtr, "   `----------------------------------------------'\n");
    fprintf (filePtr, "\nMIDI pipeline started on: %s  GMT \n", currentTime);

    /*  Release memory */
    if (currentTime) free (currentTime);

    return;
}
/******************************************************************************/



/******************************************************************************
*               European Southern Observatory
*             VLTI MIDI Data Reduction Software
*
* Module name:  handoverReport
* Input/Output: See function arguments to avoid duplication
* Description:  At the begining of the DRS before user parameters have been
*               processed all reports are logged in a temporary
*               file. Afterwards the content of this file is transfered to an
*               actual report file the name of which is user defined..
*
* History:
* 22-Jul-03     (csabet) Created
******************************************************************************/
void handoverReport (
    char        *fileName,  /*  IO: Pointer to the temporary file */
    FILE        **filePtr)
{

    /*  Local Declarations
    --------------------*/
    const char  routine[] = "handoverReport";
    char        *tempStr=NULL;
    time_t      now;
    struct tm   *newTime=NULL;
    char        *currentTime=NULL;
    struct stat buf;

    /*  Algorithm
    -----------*/
    if (diagnostic > 2) cpl_msg_info(cpl_func,"Invoking      routine   '%s' \n", routine);
    if (diagnostic > 2) fprintf(*filePtr, "Invoking      routine   '%s' \n", routine);
        
    /*  Allocate memory */
    tempStr = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
    currentTime = (char *) calloc (MIN_STRING_LENGTH, sizeof (char));

        //      Now open the following global files for writing
    if (stat (midiReportName, &buf) == 0) remove (midiReportName);       // If the file exist delete it
    midiReportPtr = fopen(midiReportName, "w");
    if (!midiReportPtr) midiReportError (*filePtr, routine, __FILE__, __LINE__, "Cannot create output event log file");

    if (stat (midiQcLogName, &buf) == 0) remove (midiQcLogName);        // If the file exist delete it
    midiQcLogPtr = fopen(midiQcLogName, "w");
    if (!midiQcLogPtr) midiReportError (*filePtr, routine, __FILE__, __LINE__, "Cannot create output QC log file");

    fclose (*filePtr);

    *filePtr = fopen (fileName, "r");
    while (fgets (tempStr, MAX_STRING_LENGTH, *filePtr) != NULL)
        fprintf (midiReportPtr, "%s", tempStr);

    now = time(NULL);
    newTime = gmtime (&now);
    strftime (currentTime, MIN_STRING_LENGTH, "%a %d %b %Y at %H:%M:%S", newTime);
    cpl_msg_info(cpl_func,"\nMIDI Data Reduction for batch  %d  started on: %s  GMT \n", batchNumber, currentTime);
    fprintf (midiReportPtr, "\nMIDI Data Reduction for batch  %d  started on: %s  GMT \n", batchNumber, currentTime);

    fclose (*filePtr);
    *filePtr = fopen (fileName, "a");

    /*  Release memory */
    if (tempStr) free (tempStr);
    if (currentTime) free (currentTime);

    return;
}
/******************************************************************************/

/*----------------------------------------------------------------------------*/
/**
  @brief    Gets the Mask/Profile file from the SOF    

  Gets the Mask/Profile file from the SOF and saves the results in fileNames->maskFileName 
 */
/*----------------------------------------------------------------------------*/

void getMaskFileFromSof(MidiFiles * fileNames, cpl_frameset  * frameset)
{
   
   
   cpl_frame             * cur_frame;
   char                  * tag_firstframe=NULL;
   char                  * tag_mask=NULL;
   char                  * tag=NULL;
   char                    gris_name[100]= ""; 
   cpl_propertylist      * plist;
   int                     i;
   
   /* Select the right profile/mask from the SOF and write the
    * variable fileNames->maskFileName accordingly */


  /* Identify the RAW and CALIB frames in the input frameset */
//   cpl_ensure_code(midi_dfs_set_groups(frameset) == CPL_ERROR_NONE,
//                   cpl_error_get_code());
   
   if (midi_dfs_set_groups(frameset) != CPL_ERROR_NONE) {
      cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                            "Identifying calib and/or raw frames failed");
      return;
   }
   
   cur_frame = cpl_frameset_get_position(frameset, 0);
   if (cur_frame == NULL) {
      cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                            "SOF does not have any file");
      return;
   }
   
/* Skip the calibration files and use the first raw file  */  
   
   for(i=0; i<cpl_frameset_get_size(frameset);i++){
/*       cpl_msg_info(cpl_func, "Reading %s",cpl_frame_get_filename(cur_frame)); */
/*       cpl_msg_info(cpl_func,"group: %d",cpl_frame_get_group(cur_frame)); */
      if((cpl_frame_get_group(cur_frame) == CPL_FRAME_GROUP_RAW)){
         break;
      }
      else {
         cur_frame = cpl_frameset_get_position(frameset, i);
         
      }
      
   }

   if (cur_frame == NULL) {
      cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                            "SOF does not have any raw files");
      cpl_msg_error(cpl_func, "SOF does not have any raw files");
      return;
   }
   

   tag_firstframe = (char*)cpl_frame_get_tag(cur_frame);
   plist = cpl_propertylist_load(cpl_frame_get_filename(cur_frame), 0);
   /*       Checking if PRISM or GRISM */
   if (cpl_propertylist_has(plist, "ESO INS GRIS NAME") == 1)
   {
      strcpy(gris_name,cpl_propertylist_get_string(plist, "ESO INS GRIS NAME"));
/*       cpl_msg_info(cpl_func,"gris_name: %s",gris_name); */
   }
           
   /*  Computing the tag for the right mask */

   if ((strcmp(tag_firstframe, MIDI_HIGH_SENS_CALIB)==0 || 
        strcmp(tag_firstframe, MIDI_HIGH_SENS_SCIENCE)==0)  && strcmp(gris_name,"GRISM")==0) {
      tag_mask = cpl_sprintf("%s", "KAPPA_HS_MASK_GRISM");  
   }
   
   else if ((strcmp(tag_firstframe, MIDI_HIGH_SENS_CALIB)==0 || 
             strcmp(tag_firstframe, MIDI_HIGH_SENS_SCIENCE)==0)  && strcmp(gris_name,"PRISM")==0) {
      tag_mask = cpl_sprintf("%s", "KAPPA_HS_MASK_PRISM");  
   }
   
   else if ((strcmp(tag_firstframe, MIDI_SCI_PHOT_CALIB)==0 || 
             strcmp(tag_firstframe, MIDI_SCI_PHOT_SCIENCE)==0)  && strcmp(gris_name,"GRISM")==0) {
      tag_mask = cpl_sprintf("%s", "KAPPA_SP_MASK_GRISM");  
   }

   else if ((strcmp(tag_firstframe, MIDI_SCI_PHOT_CALIB)==0 || 
             strcmp(tag_firstframe, MIDI_SCI_PHOT_SCIENCE)==0)  && strcmp(gris_name,"PRISM")==0) {
      tag_mask = cpl_sprintf("%s", "KAPPA_SP_MASK_PRISM");  
   }

   else {
      tag_mask = cpl_sprintf("%s", "NO MASK"); 
   }

      

   for(i=0; i<cpl_frameset_get_size(frameset);i++)
   {
       cur_frame = cpl_frameset_get_position(frameset, i);
       /*      Check the right tags */
      tag = (char*)cpl_frame_get_tag(cur_frame);
/*       cpl_msg_info(cpl_func,"tag: %s tag_mask: %s",tag,tag_mask); */
      if (strcmp(tag, tag_mask)==0) {
/*                  fileNames->maskFileName=cpl_sprintf("%s",cpl_frame_get_filename(cur_frame)); */
         sprintf(fileNames->maskFileName,"%s",cpl_frame_get_filename(cur_frame));
      }
   }
           
/*    cpl_msg_info(cpl_func,"tag_mask: %s",  tag_mask); */
/*    cpl_msg_info(cpl_func,"fileNames->maskFileName: %s", fileNames->maskFileName); */

   /*  ToDo: Free the memory */
   cpl_propertylist_delete(plist);
   cpl_free(tag_mask);          
   return ;
}
