/******************************************************************************
*******************************************************************************
*               European Southern Observatory
*             VLTI MIDI Data Reduction Software
*
* Module name:  preProcFrgHS.c
* Description:  Contains routines for pre processing
*
* History:
* 13-Jul-05     (csabet) Created
*******************************************************************************
******************************************************************************/

/******************************************************************************
*   Compiler directives
******************************************************************************/

/******************************************************************************
*   Include files
******************************************************************************/
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <cpl.h>
#include <math.h>
#include "midiGlobal.h"
#include "midiLib.h"
#include "memoryHandling.h"
#include "errorHandling.h"
#include "midiFitsUtility.h"
#include "diagnostics.h"
#include "preProcFrgHS.h"
#include "qfits.h"

/**********************************************************
*   Constant definitions
**********************************************************/

/**********************************************************
*   Global Variables
**********************************************************/

/*============================ C O D E    A R E A ===========================*/



/******************************************************************************
*               European Southern Observatory
*            VLTI MIDI Data Reduction Software
*
* Module name:  preProcFrgHS
* Input/Output: See function arguments to avoid duplication
* Description:  Prepares the data for both DISPERSED as well as UNDISPERSED modes of 
*                processing
*
* History:
* 21-Jul-03     (csabet) Created
* 25 aug 04     (JM) Modified by combining several other modules to handle both DISPERSED & UNDISPERSED modes
* 16-Jan-05        (csabet) Cleaned up as much as possible, removed GOTO instructions! and integrated into MIDI
******************************************************************************/
void preProcFrgHS (
    UserOptions        *options,            // In: User parameters
    FilterData        *filterInfo,        // IO: Pointer to the filter data structure
    MidiFiles        *fileNames,            // In: Pointer to the MIDI file structure
    CompressedData    *compressedInterf,    // Ou: Pointer to the compressed interferometry data structure
    CompressedData    *compressedPhotomA,    // Ou: Pointer to the compressed photom A data structure
    CompressedData    *compressedPhotomB,    // Ou: Pointer to the compressed photom B data structure
    ImageFormat        *formatInterf,        // In: Interf size parameters
    ImageFormat        *formatPhotomA,        // In: PhotA size parameters
    ImageFormat        *formatPhotomB,        // In: PhotB size parameters
    int                *error)                // Ou: Error status
{

    /*  Local Declarations
    --------------------*/
    const char            routine[] = "preProcFrgHS";
    FILE                *inFitsBatchPtr = NULL;
    int                    newFile, extNumOfImagingDataFile, extNumOfImagingDataMask, R;
    char                *maskFile, *cleanString, *classification, *stringTemp, *obsTech, *fileName, *title;
    unsigned int        loopCount = 0;
    ImageFormat            *localFormat;

    /*  Algorithm
    -----------*/
    if (diagnostic > 4)cpl_msg_info(cpl_func,"Invoking      routine   '%s' \n", routine);
    if (diagnostic > 4) fprintf(midiReportPtr, "Invoking      routine   '%s' \n", routine);

   cpl_msg_info(cpl_func,"\nReducing data of batch  %d \n", batchNumber);
   cpl_msg_info(cpl_func,"---------------------- \n");
    fprintf (midiReportPtr, "\nReducing data of batch  %d \n", batchNumber);
    fprintf (midiReportPtr, "---------------------- \n");

    /*  Allocate memory */
    obsTech = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
    stringTemp = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
    classification = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
    cleanString = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
    localFormat = callocImageFormat ();
    maskFile = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));

    /*  Reset status */
    *error = 0;
    compressedInterf->exists = 0;
    compressedPhotomA->exists = 0;
    compressedPhotomB->exists = 0;
    sprintf (obsTech, "UNKNOWN");
    newFile = 1;

    /*  Open the list of files */
    if ((inFitsBatchPtr = fopen (fileNames->inFitsBatch, "r")) == NULL)
    {
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, 
            "Cannot open input FITS file list. No compression has been carried out for this batch");
        free (cleanString);
        free (localFormat);
        free (stringTemp);
        free (classification);
        freeImageFormat (localFormat);
        free (maskFile);
        free (obsTech);
        *error = 1;
        return;
    }

    //  Loop through the list of files and compress data
    while (fgets (stringTemp, MAX_STRING_LENGTH, inFitsBatchPtr) != NULL)
    {
        sprintf (classification, "%s", "");
        sscanf (stringTemp, "%s%s", fileNames->inFitsName, classification);
       cpl_msg_info(cpl_func,"\nProcessing file   %s \n", fileNames->inFitsName);
        fprintf (midiReportPtr, "\nProcessing file   %s \n", fileNames->inFitsName);

        //    Get 'extNumOfImagingDataFile' extension number of IMAGING_DATA in input file
        extNumOfImagingDataFile  = findImagingDataExtension (fileNames->inFitsName, TAB_IMAGING_DATA, error);
        if (*error) break;

        //    Get Image Format parameters
        if (extNumOfImagingDataFile > 0)
        {
            getImageFormat (fileNames->inFitsName, extNumOfImagingDataFile, localFormat, error);
            if (*error) break;
        }
        else localFormat->hasData = 0;

        //    Get filter data and mask file name from the first FITS file
        if (loopCount == 0)
        {
            getFilterData (fileNames->inFitsName, filterInfo, error);
            if (*error) break;
            else fprintf (midiReportPtr, "First file of batch = %s (QCLOG)\n", fileNames->inFitsName);

            //    Select mask file
            selectMask (options->maskMode, fileNames, maskFile, error);
            if (*error) break;

            //    Get 'extNumOfImagingDataMask' extension number of IMAGING_DATA in mask file
            extNumOfImagingDataMask = findImagingDataExtension (maskFile, MASK_IMAGING_DATA, error);
            if (*error) break;

           cpl_msg_info(cpl_func,"Mask File is   %s\n", maskFile);
            fprintf( midiReportPtr, "Mask File is   %s\n", maskFile);
        }    

        //    Check if the file has data
        if (localFormat->hasData)
        {
            //    Check "Observation Technique". We expect either INTERFEROMETRY or IMAGE
            if (strcmp (localFormat->obsTech, "INTERFEROMETRY") == 0)
            {
                //    Check previous TYPE
                if (strcmp (obsTech, "INTERF") != 0) newFile = 1;
                
                //    Set the TYPE flag
                sprintf (obsTech, "INTERF");
                
                //    Check consistency
                if ((formatInterf->numOfDetectorRegions != localFormat->numOfDetectorRegions) ||
                    (formatInterf->iXWidth != localFormat->iXWidth) || (formatInterf->iYWidth != localFormat->iYWidth))
                {
                    midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Expected format is incorrect");
                    *error = 1;
                    break;
                }
                
                //    Send Interferometry file for compression
                organiseFrgHS (newFile, obsTech, fileNames->inFitsName, maskFile, 
                    extNumOfImagingDataFile, extNumOfImagingDataMask, localFormat, formatInterf, 
                    compressedInterf, error);

                if (*error)    break;
                newFile = 0;    // So the next time through, will know not to start at frame 0
            }
            else if ((strcmp (localFormat->obsTech, "IMAGE,WINDOW,CHOPNOD") == 0) &&
                (strcmp(localFormat->shutterId,"AOPEN") == 0))
            {
                //    Check previous TYPE
                if (strcmp (obsTech, "PHOTOMA") != 0) newFile = 1;
                
                //    Set the TYPE flag
                sprintf (obsTech, "PHOTOMA");
                
                //    Check consistency
                if ((formatPhotomA->numOfDetectorRegions != localFormat->numOfDetectorRegions) ||
                    (formatPhotomA->iXWidth != localFormat->iXWidth) || (formatPhotomA->iYWidth != localFormat->iYWidth))
                {
                    midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Expected format is incorrect");
                    *error = 1;
                    break;
                }
                
                //    Send Photometry A file for compression
                organiseFrgHS (newFile, obsTech, fileNames->inFitsName, maskFile, 
                    extNumOfImagingDataFile, extNumOfImagingDataMask, localFormat, formatPhotomA, 
                    compressedPhotomA, error);

                if (*error)    break;
                newFile = 0;    // So the next time through, will know not to start at frame 0
            }
            else if ((strcmp (localFormat->obsTech, "IMAGE,WINDOW,CHOPNOD") == 0) &&
                (strcmp(localFormat->shutterId,"BOPEN") == 0))
            {
                //    Check previous TYPE
                if (strcmp (obsTech, "PHOTOMB") != 0) newFile = 1;
                
                //    Set the TYPE flag
                sprintf (obsTech, "PHOTOMB");
                
                //    Check consistency
                if ((formatPhotomB->numOfDetectorRegions != localFormat->numOfDetectorRegions) ||
                    (formatPhotomB->iXWidth != localFormat->iXWidth) || (formatPhotomB->iYWidth != localFormat->iYWidth))
                {
                    midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Expected format is incorrect");
                    *error = 1;
                    break;
                }
                
                //    Send Photometry B file for compression
                organiseFrgHS (newFile, obsTech, fileNames->inFitsName, maskFile, 
                    extNumOfImagingDataFile, extNumOfImagingDataMask, localFormat, formatPhotomB, 
                    compressedPhotomB, error);

                if (*error)    break;
                newFile = 0;    // So the next time through, will know not to start at frame 0
            }
            else
            {
                midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Unknown Catg, Type or Tech");
                *error = 1;
                break;
            }
        }
        else
        {
            if (diagnostic)
            {
                sprintf (midiMessage, "No data tables in %s. Not processed", fileNames->inFitsName);
                midiReportInfo (midiReportPtr, routine, __FILE__, __LINE__, midiMessage);
            }
        }
        loopCount++;
    }

    //    Close files and release memory
    fclose (inFitsBatchPtr);
    free (cleanString);
    free (stringTemp);
    free (classification);
    freeImageFormat (localFormat);
    free (maskFile);
    free (obsTech);

    //    Check if reduction has been carried out
   cpl_msg_info(cpl_func,"\nCompression status \n");
   cpl_msg_info(cpl_func,"------------------ \n");
    if (compressedInterf->exists)cpl_msg_info(cpl_func,"Created Compressed Interferometry Data\n");
    if (compressedPhotomA->exists)cpl_msg_info(cpl_func,"Created Compressed Photometry A Data\n");
    if (compressedPhotomB->exists)cpl_msg_info(cpl_func,"Created Compressed Photometry B Data\n");
   cpl_msg_info(cpl_func,"\n");

    fprintf (midiReportPtr, "\nCompression status \n");
    fprintf (midiReportPtr, "------------------ \n");
    if (compressedInterf->exists) fprintf (midiReportPtr, "Created Compressed Interferometry Data\n");
    if (compressedPhotomA->exists) fprintf (midiReportPtr, "Created Compressed Photometry A Data\n");
    if (compressedPhotomB->exists) fprintf (midiReportPtr, "Created Compressed Photometry B Data\n");
    fprintf (midiReportPtr, "\n");

    if (!(compressedInterf->exists) || 
        !(compressedPhotomA->exists) || !(compressedPhotomB->exists) || *error)
    {
        *error = 1;
        sprintf (midiMessage, 
            "Cannot continue. Need the following compressed data: Interf, PhotomA, PhotomB");
        midiReportWarning (midiReportPtr, 
            routine, __FILE__, __LINE__, midiMessage);
        return;
    }

    //    Display images
    if (diagnostic > 1 && plotFile)
    {
        fileName = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
        title = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));

        for (R = 0; R < formatPhotomA->numOfRegionsToProcess; R++)
        {
            sprintf (fileName, "3dPhotomADATA%d", R+1);
            sprintf (title, "Photometry A DATA %d (Masked and Sky removed)", R+1);
            midiCreatePlotFile3D (fileName, title, "X", "Y", "Flux", 0, 
                compressedPhotomA->image[R], formatPhotomA->iXWidth, formatPhotomA->iYWidth, "lines", "3");
        }

        for (R = 0; R < formatPhotomB->numOfRegionsToProcess; R++)
        {
            sprintf (fileName, "3dPhotomBDATA%d", R+1);
            sprintf (title, "Photometry B DATA %d (Masked and Sky removed)", R+1);
            midiCreatePlotFile3D (fileName, title, "X", "Y", "Flux", 0, 
                compressedPhotomB->image[R], formatPhotomB->iXWidth, formatPhotomB->iYWidth, "lines", "3");
        }

        free (fileName);
        free (title);
    }

    return;
}
/*****************************************************************************/


/******************************************************************************
*               European Southern Observatory
*            VLTI MIDI Data Reduction Software
*
* Module name:  organiseFrgHS
* Input/Output: See function arguments to avoid duplication
* Description:  Prepares HIGH_SENS files for compression
*
* History:
* 03-Nov-05     (csabet) Created
******************************************************************************/
void organiseFrgHS (
    int                    newFile,                    // In: Flag indicating continuation
    char                *obsTech,                    // In: Whether Interf or Photom
    char                *fitsFile,                    // In: Name of the input FITS file
    char                *maskFile,                    // In: Name of the Mask file
    int                    extNumOfImagingDataFile,    // In: Extention number
    int                    extNumOfImagingDataMask,    // In: Extention number
    ImageFormat            *localFormat,                // In: Local split file format
    ImageFormat            *formatMax,                    // In: Maximum file format
    CompressedData        *compressed,                // In: Compressed data
    int                    *error)                        // Ou: Error status
{

    //    Local Declarations
    //    ------------------
    const char        routine[] = "organiseFrgHS";
    int                frame0;

    //    Algorithm
    //    ---------
    if (diagnostic > 4)cpl_msg_info(cpl_func,"Invoking      routine   '%s' \n", routine);
    if (diagnostic > 4) fprintf(midiReportPtr, "Invoking      routine   '%s' \n", routine);

    //    Initialise
    *error = 0;

    //    Compress data
    frame0 = compressFrgHS (newFile, obsTech, fitsFile, maskFile, extNumOfImagingDataFile,
        extNumOfImagingDataMask, compressed, localFormat, formatMax->numOfFrames, error);
    if (*error) return;

    //    Report compression status
    if (diagnostic)
    {
        sprintf (midiMessage, "\nWrote %d frames into %s, starting at frame %d\nThus now %d frames total,"
            " expecting %d altogether.  Error=%d \n\n", localFormat->numOfFrames, obsTech, frame0, 
            localFormat->numOfFrames+frame0, formatMax->numOfFrames, *error);
        midiReportInfo (midiReportPtr, routine, __FILE__, __LINE__, midiMessage);
    }

    //    Set the flags
    if ((frame0 + localFormat->numOfFrames) == formatMax->numOfFrames)  // ALL have been read in
    {
        compressed->exists = 1;    // Mark it as completed

        //    Now with data read in, display some of the parameters:
        if (plotFile && (strcmp (obsTech, "INTERF") == 0) && diagnostic)
        {
            midiCreatePlotFile2D ("Time", "TIME", "Frame", "Time", 0, compressed->time, 0, 
                formatMax->numOfFrames, 1, 0);
            midiCreatePlotFile2D ("BigDelayLine", "Net VLTI Delay Line Positions", "Frame", 
                "Net Position", 0, compressed->bigDL, 0, formatMax->numOfFrames, 1, 0);
            midiCreatePlotFile2D ("PiezoDelayLine", "Net Piezo Delay Line Position", "Frame", "Net Delay", 0, 
                compressed->localOPD, 0, formatMax->numOfFrames, 1, 0);
        }
    }
            
    return;
}
/*****************************************************************************/



/******************************************************************************
*               European Southern Observatory
*            VLTI MIDI Data Reduction Software
*
* Module name:    compressFrgHS
* Input/Output:    See function arguments to avoid duplication
* Description:    This routine prepares the interferomerty data for dispersed
*               AND undispersed processing. In DISPERSED mode iDispFringe points to 
*                a 3D structure. One dimension is time (frame), the second dimension is 
*                wave (produced by collapsing each X, Y image into one line) and the 
*                third dimension is the number of sub-regions.
* History:
* 14-Aug-03        (csabet) Created. Derived from pballest July 2003
* 25-Aug-04        (JM) Modified by combining several other modules to handle both DISPERSED & UNDISPERSED modes
* 16-Jan-05        (csabet) Cleaned and added numerous features. This module is still far too big. Will 
*                modularise later
******************************************************************************/
int compressFrgHS (                                // Ou: Status. Returns number of first frame compressed; -1 if error
    int                    newSet,                    // In: Flag indicating the arrival of a new set of FITS files
    char                *obsTech,                // In: Whether Interf or Photom
    char                *inFitsFile,            // In: Name of the input FITS file
    char                *maskFile,                // In: Name of the mask file
    int                    extNumOfImagingDataFile,// In: Extension number of the IMAGING_DATA in input file
    int                    extNumOfImagingDataMask,// In: Extension number of the IMAGING_DATA in mask file
    CompressedData        *compressed,            // In: Compressed data
    ImageFormat            *localFormat,            // In: Local split file format
    int                    numOfFramesMax,            // In: Maximum number of frames
    int                    *error)                    // Ou: Error status
{

    //    Local Declarations
    //    ------------------
    const char        routine[] = "compressFrgHS";
    qfits_table        *pTable  = NULL, *pMask = NULL;
    short int        **inData, *inSteppingPhase;
    float            **inMask;
    float            accum, current;
    float            *normalization; // Normalization factor. Becomes an array in X for the normalization of the 
                                    // compressed data, valid for each region then overwritten
    int                frameOffset, indexOPD = - 1, indexLOCALOPD = - 1, indexTIME = - 1, maxstep;
    char            **inTARTYP = NULL, *tempStr, fitsColumnString[10], *dataName, *title=NULL, *fileString=NULL;
    double            *inTIME = NULL; 
    int                frame0 = - 1;     // Gets set to a valid number UNLESS routine terminates in error
    double            (*inLOCALOPD)[2] = NULL, (*inOPD)[2] = NULL; 
    static int        aprioriSteppingPhase, iF, chopInterval, channelSelected = 0;
    static double    zeroTime ;
    int                i, k, fps, *foundData, foundSteppingPhase = 0, indexSteppingPhase, scalingOffset, *indexData , 
                    *indexMask, maskWidthX, maskWidthY,    maskSubWindow, F, X, Y, R, 
                    *indexTARTYP, *foundTARTYP = 0, startframe=0;
    int                tartypMult = 2;    // Signifies that there are 2 (instead of 1!) characters per tartyp value (only 
                                    // first is useful). In future, if the problem producing the MIDI files is solved, 
                                    // can be changed to 1 (only for newer files) or made variable (with some way to 
                                    // determine whether it is an "old" type file). Note: this problem really has 
                                    // nothing to do with Qfits, but with the (unintended) way the files are written.
    int                i2, fst, snd, found;
    char            lastType;
    
    //    Algorithm
    //    ---------
    if (diagnostic > 4)cpl_msg_info(cpl_func,"Invoking      routine   '%s' \n", routine);
    if (diagnostic > 4) fprintf (midiReportPtr, "Invoking      routine   '%s' \n", routine);

    //    Reset status
    *error = 0;
    fps = localFormat->framesPerScan;

    //    Allocate memory
    inData = (short int **) calloc (localFormat->numOfDetectorRegions, sizeof (short int *));
    inTARTYP = (char **) calloc (localFormat->numOfDetectorRegions, sizeof (char *));
    inMask = (float **) calloc (localFormat->numOfRegionsToProcess, sizeof (float *));
    foundData = (int *) calloc (localFormat->numOfDetectorRegions, sizeof (int));
    indexData = (int *) calloc (localFormat->numOfDetectorRegions, sizeof (int));
    foundTARTYP = (int *) calloc (localFormat->numOfDetectorRegions, sizeof (int));
    indexTARTYP = (int *) calloc (localFormat->numOfDetectorRegions, sizeof (int));
    indexMask = (int *) calloc (localFormat->numOfDetectorRegions, sizeof (int));
    normalization = (float *) calloc (localFormat->iXWidth, sizeof (float));
    dataName = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
        
    //    Initialise the indices for a new set of FITS files. The
    //    files may be split the indices ensure continuity of split files
    if (newSet)
    {
        aprioriSteppingPhase = 0;
        iF = 0;
    }

    //    Make sure the sum of the split-files numOfFrames is not more than the allowed maximum
    if ((iF + localFormat->numOfFrames) > numOfFramesMax)
        localFormat->numOfFrames = numOfFramesMax - iF;

    //    Open IMAGING_DATA = INDEX 2
    pMask = qfits_table_open (maskFile, extNumOfImagingDataMask);
    if (!pMask)
    {
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot load IMAGING_DATA");
        free (inData);
        free (inTARTYP);
        free (inMask);
        free (foundData);
        free (indexData);
        free (foundTARTYP);
        free (indexTARTYP);
        free (indexMask);
        free (normalization);
        free (dataName);
        *error = 1;
        return (-1);
    }
    pTable = qfits_table_open (inFitsFile, extNumOfImagingDataFile);
    if (!pTable)
    {
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot load IMAGING_DATA");
        qfits_table_close (pMask);
        free (inData);
        free (inTARTYP);
        free (inMask);
        free (foundData);
        free (indexData);
        free (foundTARTYP);
        free (indexTARTYP);
        free (indexMask);
        free (normalization);
        free (dataName);
        *error = 1;
        return (-1);
    }

    //    Get data table information
    for (R = 0; R < localFormat->numOfDetectorRegions; R++)
    {
        foundData[R] = 0;
        indexData[R] = 0;
    }
    for (i = 0; i < pTable->nc; i++)
    {
        for (R = 0; R < localFormat->numOfDetectorRegions; R++)
        {
            sprintf (dataName, "DATA%d", R+1);
            if (strcmp (pTable->col[i].tlabel, dataName) == 0)
            {
                foundData[R] = 1;
                indexData[R] = i;
                if (diagnostic) 
                {
                   cpl_msg_info(cpl_func,"Found 'DATA%d' at column %d in data file %s \n", R+1, i+1, inFitsFile);
                    fprintf(midiReportPtr, "Found 'DATA%d' at column %d in data file %s \n", R+1, i+1, inFitsFile);
                }
            }
            
            sprintf (dataName, "TARTYP%d", R+1);
            if (strcmp (pTable->col[i].tlabel, dataName) == 0)
            {
                foundTARTYP[R] = 1;
                indexTARTYP[R] = i;
                if (diagnostic) 
                {
                   cpl_msg_info(cpl_func,"Found 'TARTYP%d' at column %d in data file %s \n", R+1, i+1, inFitsFile);
                    fprintf(midiReportPtr, "Found 'TARTYP%d' at column %d in data file %s \n", R+1, i+1, inFitsFile);
                }
            }
        }
        if (strcmp (pTable->col[i].tlabel, "STEPPING_PHASE") == 0)
        {
            foundSteppingPhase = 1;
            indexSteppingPhase = i;
        }
        if (strcmp (pTable->col[i].tlabel, "OPD") == 0)
        {
            indexOPD = i;
        }
        if (strcmp (pTable->col[i].tlabel, "LOCALOPD") == 0)
        {
            indexLOCALOPD = i;
        }
        if (strcmp (pTable->col[i].tlabel, "TIME") == 0)
        {
            indexTIME = i;
        }
    }

    //    Now issue warnings
    if (foundSteppingPhase == 0)
    {
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot find STEPPING_PHASE in data FITS file");
        *error = 1;
    }
    if (indexOPD < 0)
    {
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot find column for OPD in data FITS file");
        *error = 1;
    }
    if (indexLOCALOPD < 0)
    {
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot find column for LOCALOPD in data FITS file");
        *error = 1;
    }
    if (indexTIME < 0)
    {
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot find column for TIME in data FITS file");
        *error = 1;
    }
    for (R = 0; R < localFormat->numOfDetectorRegions; R++)
    {
        if (foundData[R] == 0)
        {
            midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, 
                "Cannot find requested DATA column in data FITS file");
            *error = 1;
        }
        if (foundTARTYP[R] == 0)
        {
            midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, 
                "Cannot find requested TARTYP column in data FITS file");
            *error = 1;
        }
    }
    
    //    Get mask data table information
    for (R = 0; R < localFormat->numOfDetectorRegions; R++)
    {
        foundData[R] = 0;
        indexMask[R] = 0;
    }
    for (i = 0; i < pMask->nc; i++)
    {
        for (R = 0; R < localFormat->numOfDetectorRegions; R++)
        {
            sprintf (dataName, "DATA%d", R+1);
            if (strcmp (pMask->col[i].tlabel, dataName) == 0)
            {
                foundData[R] = 1;
                indexMask[R] = i;
                if (diagnostic) 
                {
                   cpl_msg_info(cpl_func,"Found 'DATA%d' at column %d in mask file %s \n", R+1, i+1, maskFile);
                    fprintf(midiReportPtr, "Found 'DATA%d' at column %d in mask file %s \n", R+1, i+1, maskFile);
                }
            }
        }
    }

    //    Now issue warnings
    for (R = 0; R < localFormat->numOfDetectorRegions; R++)
    {
        if (foundData[R] == 0)
        {
            midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, 
                "Cannot find requested DATA column in mask FITS file");
           *error = 1;
        }
    }

    //    Get subwindow format for the mask file
    sprintf (fitsColumnString, "TDIM%d", indexMask[1]+1);    // Point it to the column of a valid region
    tempStr = qfits_query_ext (maskFile, fitsColumnString, extNumOfImagingDataMask);
    sscanf (tempStr, "'(%d,%d) '", &maskWidthX, &maskWidthY);
    maskSubWindow = maskWidthX * maskWidthY;
    if (newSet && diagnostic)cpl_msg_info(cpl_func,"Mask sub-window size = %d\n", maskSubWindow);
    if (diagnostic)cpl_msg_info(cpl_func,"Data sub-window size = %d\n", localFormat->subWindowSize);
    if (newSet) fprintf (midiReportPtr, "Mask sub-window size = %d\n", maskSubWindow);
    fprintf (midiReportPtr, "Data sub-window size = %d\n", localFormat->subWindowSize);
    if (maskSubWindow != localFormat->subWindowSize)
    {
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Mask has incompatible sub window size");
        *error = 1;
    }

    //    Get MAXSTEP value
    tempStr = qfits_query_ext (inFitsFile, "MAXSTEP", extNumOfImagingDataFile);
    if (tempStr != NULL)
    { 
        if (diagnostic)cpl_msg_info(cpl_func,"MAXSTEP = %s\n", tempStr);
        fprintf( midiReportPtr, "MAXSTEP = %s\n", tempStr);
        sscanf (tempStr, "%d", &maxstep);
    }
    else
    {
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot read MAXSTEP");
        *error = 1;
    }

    //    Read column COL_STEPPING_PHASE and check compatibility
    inSteppingPhase = (short int*)qfits_query_column (pTable, indexSteppingPhase, NULL);
    if (inSteppingPhase[0] != ((aprioriSteppingPhase % localFormat->framesPerScan) + 1))
    {
        sprintf (midiMessage, "Incorrect Stepping Phase. Expected %d to %d. Instead found %d to %d", 
            aprioriSteppingPhase+1, maxstep, inSteppingPhase[0], maxstep-1);
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, midiMessage);
        *error = 1;
    }

    //    Release memory and exit this module if there has been a warning. No point going any further
    if (*error)
    {
        qfits_table_close (pTable);
        qfits_table_close (pMask);
        free (inData);
        free (inTARTYP);
        free (inMask);
        free (foundData);
        free (indexData);
        free (foundTARTYP);
        free (indexTARTYP);
        free (indexMask);
        free (inSteppingPhase);
        free (normalization);
        free (dataName);
        return (-1);
    }

    //    Read all regions from input files
    for (R = 0; R < localFormat->numOfDetectorRegions; R++)
    {
        inData[R] = (short int*) qfits_query_column (pTable, indexData[R], NULL);
        inTARTYP[R] = (char *) qfits_query_column (pTable, indexTARTYP[R], NULL);
        
        found = 0;
        for (F = 0; F < localFormat->numOfFrames; F++)
        {
            //    Correct the initial anomalies
            if (newSet && (strcmp (obsTech, "INTERF") != 0))
            {
                if (inTARTYP[R][F*tartypMult] == 'U')
                    found = 1;
                if (!found)
                    inTARTYP[R][F*tartypMult] = 'U';
            }

            //    Load data and correct tartype problem
            (compressed->tarType)[iF+F] = inTARTYP[R][F*tartypMult];
                
            if (R > 0) 
            {
                if (inTARTYP[R][F*tartypMult] != inTARTYP[R-1][F*tartypMult])
                {
                    sprintf (midiMessage, "Incompatible Interferometry TARTYP%d & TARTYP%d at frame %d", R, R-1, iF+F);
                    midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, midiMessage);

                    //    All channels are affected
                    for (i = 0; i < localFormat->iXWidth; i++)
                        compressed->rejectList[i][iF+F] |= BSL_TARTYPE_CROSS;
                    
                    //    badScanList is only meaningful for Interferometry data
                    if (strcmp (obsTech, "INTERF") == 0) compressed->badScanList[(iF+F)/fps] |= BSL_TARTYPE_CROSS;
                }
            }
        }
    }

    //    Get TIME
    if (indexTIME >= 0)    // Valid ...
    {
        startframe = 0;
        inTIME = (double *) qfits_query_column (pTable, indexTIME, NULL);
        //    Stuff it into time, subtracting off the zero time
        if (iF == 0) zeroTime = inTIME[0];

        //    Fix a bug found in the data:
        if (iF == 0)
        {
            for (startframe = 0; startframe < ARB_NUM_OF_FRAMES; startframe++)
            {
                if ((zeroTime = inTIME[startframe]) >  1.0)    // Should be more like 53000 or something!
                break;
            }
        }
        if (startframe)
        {
            if (diagnostic) 
            {
               cpl_msg_info(cpl_func,"\nLOOK: frames 0 - %d had ZERO for their time field!!\n", startframe-1);
                fprintf(midiReportPtr, "\nLOOK: frames 0 - %d had ZERO for their time field!!\n", startframe-1);
            }
        }

        for (F = startframe; F < localFormat->numOfFrames; F++)
        {
            compressed->time[iF+F] = (float)(inTIME[F] - zeroTime);
            if (isnan ((compressed->time[iF+F])))
            {
                sprintf (midiMessage, "inTIME has an INVALID value at frame %d", iF+F);
                midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, midiMessage);

                //    All channels and regions are affected
                for (i = 0; i < localFormat->iXWidth; i++)
                    compressed->rejectList[i][iF+F] |= BSL_TIME_ERROR;
                    
                //    badScanList is only meaningful for Interferometry data
                if (strcmp (obsTech, "INTERF") == 0) compressed->badScanList[(iF+F)/fps] |= BSL_TIME_ERROR;
            }
        }
    }

    //    Get LOCALOPD
    if (indexLOCALOPD >= 0)    // Valid ...
    {
        inLOCALOPD = (double (*)[2]) qfits_query_column (pTable, indexLOCALOPD, NULL);
        //    Add up the 2 delay lines, and stuff it into localOPD:
        for (F = 0; F < localFormat->numOfFrames; F++)
        {
            compressed->localOPD[iF+F] = (float)(inLOCALOPD[F][0] + inLOCALOPD[F][1]);
            if (isnan ((compressed->localOPD[iF+F])))
            {
                sprintf (midiMessage, "localOPD has an INVALID value at frame %d", iF+F);
                midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, midiMessage);

                //    All channels are affected
                for (i = 0; i < localFormat->iXWidth; i++)
                    compressed->rejectList[i][iF+F] |= BSL_LOCALOPD_ERROR;
                    
                //    badScanList is only meaningful for Interferometry data
                if (strcmp (obsTech, "INTERF") == 0) compressed->badScanList[(iF+F)/fps] |= BSL_LOCALOPD_ERROR;
            }
        }
    }

    //    Get OPD
    if (indexOPD >= 0)    // Valid ...
    {
        inOPD = (double (*)[2]) qfits_query_column (pTable, indexOPD, NULL);
        //    Add up the 2 delay lines, and stuff it into bigDL:
        for (F = 0; F < localFormat->numOfFrames; F++)
        {
            (compressed->bigDL)[iF+F] = (float)(inOPD[F][0] + inOPD[F][1]);
            if (isnan ((compressed->bigDL)[iF+F]))
            {
                sprintf (midiMessage, "bigDL has an INVALID value at frame %d", iF+F); 
                midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, midiMessage);

                //    All channels are affected
                for (i = 0; i < localFormat->iXWidth; i++)
                    compressed->rejectList[i][iF+F] |= BSL_OPD_ERROR;

                //    badScanList is only meaningful for Interferometry data
                if (strcmp (obsTech, "INTERF") == 0) compressed->badScanList[(iF+F)/fps] |= BSL_OPD_ERROR;
            }
        }
    }
    
    //    Compute chopping frequency. (Diagnostic)
    if (newSet && (strcmp (obsTech, "INTERF") != 0))
    {
        lastType = ' ';
        chopInterval = 0;
        for (F = 0; F < localFormat->numOfFrames; F++)
        {
            if (compressed->tarType[F] == 'U' && (lastType == 'T' || lastType == 'S'))
            {
                break;
            }
            else
            {
                lastType = compressed->tarType[F];
                chopInterval++;
            }
        }
    }
    
    //    Read all subregions from mask file
    for (R = 0; R < localFormat->numOfDetectorRegions; R++)
    {
        inMask[R] = (float*) qfits_query_column (pMask, indexMask[R], NULL);

        //    Display mask data
        if (diagnostic > 1 && plotFile && newSet && (strcmp(obsTech, "INTERF") == 0))
        {
            fileString = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
            title = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
            sprintf (fileString, "3dMaskDATA%d", R+1);
            sprintf (title, "3D Mask DATA %d", R+1);
            midiCreatePlotFile3D (fileString, title, "X", "Y", "Flux", 0, 
                inMask[R], localFormat->iXWidth, localFormat->iYWidth, "lines", "3");
            free (fileString);
            free (title);
        }
    }
    
    //    Select valid channels
    if ((newSet || !channelSelected) && (strcmp(obsTech, "INTERF") == 0))
    {
        selectChannels (0, localFormat->numOfDetectorRegions, localFormat, inMask);
        channelSelected = 1;
    }
    
    //    Get the scaling offset
    for (i = 14; i < 25; i++)
    {
        sprintf (dataName, "TZERO%d", i);
        tempStr = qfits_query_ext (inFitsFile, dataName, extNumOfImagingDataFile);
        if (tempStr != NULL)
        {
            if (diagnostic)cpl_msg_info(cpl_func,"Scaling Offset = %s\n", tempStr);
            if (diagnostic) fprintf (midiReportPtr, "Scaling Offset = %s\n", tempStr);
            sscanf (tempStr, "%d", &scalingOffset);
            break;
        }
    }
    if (tempStr == NULL)
    {
        scalingOffset = 0;
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot read Scaling Offset. It is set to 0");
    }

    //    ACTUAL COMPRESSION OF DATA HERE
    for (R = 0; R < localFormat->numOfDetectorRegions; R++)
    {    
        //    Determine the normalization for each X (DISPERSED mode)
        for (X = 0; X < localFormat->iXWidth; X++)
        {
            accum = 0.0F;
            for (Y = 0; Y < localFormat->iYWidth; Y++)
                accum +=  inMask[R][Y * localFormat->iXWidth + X];

            if (accum > 0.0)
                normalization[X] = 1.F/accum;
            else
                normalization[X] = 1.F;
        }

        for (F = 0; F < localFormat->numOfFrames; F++)
        {
            frameOffset = F * localFormat->subWindowSize;
            for (X = 0; X < localFormat->iXWidth; X++)
            {
                accum = 0.0F;
                for (Y = 0; Y < localFormat->iYWidth; Y++)
                {
                    k = Y * localFormat->iXWidth + X;
                    i = frameOffset + k;

                    //    If data is bad reject it
/*
                    if (isnan (inData[R][i]))
                    {
                        sprintf (midiMessage, "inData has an INVALID value at frame %d", (iF+F));
                        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, midiMessage);
 
                        compressed->rejectList[X][iF+F] |= BSL_DATA_ERROR;
                        
                        //    badScanList is only meaningful for Interferometry data
                        if (strcmp (obsTech, "INTERF") == 0) compressed->badScanList[(iF+F)/fps] |= BSL_DATA_ERROR;
                    }
                    else
*/
                    {
                        current = (inData[R][i] + scalingOffset) * inMask[R][k];
                        accum += current;
                    }
                }

                //    Accumulate further 
                (((compressed->iDispFringe)[R])[X])[iF+F] = accum * normalization[X];
                if (R == 0) 
                {
                    (compressed->iFringe1)[iF+F] += accum;
                    (compressed->iFringe)[iF+F] += accum;
                }
                if (R == 1)
                {
                    (compressed->iFringe2)[iF+F] += accum;
                    (compressed->iFringe)[iF+F] -= accum;
                }
            }
        }

        //    Diagnostic
        if (diagnostic > 1 && (strcmp (obsTech, "INTERF") != 0))
        {
            for (F = 0; F < localFormat->numOfFrames-chopInterval; F++)
            {
                fst = F * localFormat->subWindowSize;
                snd = (F+chopInterval) * localFormat->subWindowSize;
                if (compressed->tarType[F] != 'U')
                {
                    for (Y = 0; Y < localFormat->iYWidth; Y++)
                    {
                        for (X = 0; X < localFormat->iXWidth; X++)
                        {
                            k = X * localFormat->iYWidth + Y;
                            i = fst + k;
                            i2 = snd + k;
                            if ((compressed->tarType[F] == 'T') && (compressed->tarType[F+chopInterval] == 'S'))
                                current = (inData[R][i] + scalingOffset) - (inData[R][i2] + scalingOffset);
                            else if ((compressed->tarType[F] == 'S') && (compressed->tarType[F+chopInterval] == 'T'))
                                current = (inData[R][i2] + scalingOffset) - (inData[R][i] + scalingOffset);
                        
                            //    Load
                            compressed->image[R][k] += current;
                        }
                    }
                }
            }
            F += chopInterval;
        }
    }

    aprioriSteppingPhase = inSteppingPhase[F-1];    // F-1 because of the above loop
    if (aprioriSteppingPhase == localFormat->framesPerScan) aprioriSteppingPhase = 0;

    frame0 = iF;    // To output on return, below
    iF += localFormat->numOfFrames;    // Set it correctly for the next call to this function

    //    Clean up now:
    for (R = 0; R < localFormat->numOfDetectorRegions; R++) 
    {
        free (inData[R]);
        free (inTARTYP[R]);
        free (inMask[R]);
    }
    if (pMask) qfits_table_close (pMask);
    if (pTable) qfits_table_close (pTable);
    if (inLOCALOPD) free(inLOCALOPD);
    if (inOPD) free(inOPD);
    if (inTIME) free(inTIME);
    free (inData);
    free (inTARTYP);
    free (inMask);
    free (inSteppingPhase);
    free (foundData);
    free (indexData);
    free (foundTARTYP);
    free (indexTARTYP);
    free (indexMask);
    free (normalization);
    free (dataName);

    return (frame0);
}
/*****************************************************************************/

