/******************************************************************************
*******************************************************************************
*               European Southern Observatory
*             VLTI MIDI Data Reduction Software
*
* Module name:  procKappa.c
* Description:  Contains routines for Transfer Ratios computations
*
* History:
* 03-Dec-05     (csabet) Created
*******************************************************************************
******************************************************************************/

/******************************************************************************
*   Compiler directives
******************************************************************************/

/******************************************************************************
*   Include files
******************************************************************************/
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <cpl.h>
#include <stdlib.h>
#include <stdlib.h>
#include <math.h>
#include "midiGlobal.h"
#include "fft.h"
#include "midiLib.h"
#include "memoryHandling.h"
#include "errorHandling.h"
#include "diagnostics.h"
#include "imageProcessing.h"
#include "calibration.h"
#include "statistics.h"
#include "fitsAnalysisKappa.h"
#include "preProcKappa.h"
#include "procKappa.h"
#include "procFrgHS.h"
#include "createProdKappa.h"

/**********************************************************
*   Constant definitions
**********************************************************/

/**********************************************************
*   Global Variables
**********************************************************/

/*============================ C O D E    A R E A ===========================*/




/******************************************************************************
*               European Southern Observatory
*            VLTI MIDI Data Reduction Software
*
* Module name:  procKappa
* Input/Output: See function arguments to avoid duplication
* Description:    This is the main routine for calculating Transfer Ratios
*                If options->drsKap = 1, this process creates the Transfer Ratios file
*                in current directory.
*                If options->drsKap = 2, this process updates the Transfer Ratios file
*                in the database as well as creating a file in the current directory
*
*                The relevant file structures are:
*
*                Photometry A files have the following configurations:
*                DATA1 = Photometry A     data
*                DATA2 = Interferometry     data
*                DATA3 = Interferometry    data
*                DATA4 = NULL
*
*                Photometry B files have the following configurations:
*                DATA1 = NULL
*                DATA2 = Interferometry     data
*                DATA3 = Interferometry    data
*                DATA4 = Photometry B     data
*
* History:
* 05-Dec-05     (csabet) Created
******************************************************************************/
void procKappa (
    MidiFiles    *fileNames,    // In: Pointer to midi files structure
    UserOptions    *options,    // In: Pointer to user options
    int            *error)        // Ou: Error status
{

    //    Local Declarations
    //    ------------------
    const char      routine[] = "procKappa";
    ImageFormat        *formatInterfA=NULL, *formatInterfB=NULL, 
                    *formatPhotomA=NULL, *formatPhotomB=NULL;
    CompressedData    *compressedInterfA=NULL, *compressedInterfB=NULL, 
                    *compressedPhotomA=NULL, *compressedPhotomB=NULL;
    TransferRatios    *trr;
    FilterData        *filterInfo=NULL;

    //    Algorithm
    //    ---------
    if (diagnostic > 4)cpl_msg_info(cpl_func,"Invoking      routine   '%s' \n", routine);
    if (diagnostic > 4) fprintf(midiReportPtr, "Invoking      routine   '%s' \n", routine);

    //    Reset status
    *error = 0;
    if (options->processing == UNDISPERSED)
    {
        *error = 1;
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, 
            "'UNDISPERSED' is an invalid processing request for SCI_PHOT data");
        return;
    }

    //    Allocate memory
    formatInterfA = callocImageFormat ();
    formatInterfB = callocImageFormat ();
    formatPhotomA = callocImageFormat ();
    formatPhotomB = callocImageFormat ();
    
    //    Analyse Input files
    //    -------------------
    analyseFitsKappa (fileNames, formatInterfA, formatInterfB, 
        formatPhotomA, formatPhotomB, error);
    if (*error)
    {
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot analyse FITS files");
        freeImageFormat (formatInterfA);
        freeImageFormat (formatInterfB);
        freeImageFormat (formatPhotomA);
        freeImageFormat (formatPhotomB);
        return;
    }

    //    Reduce Data
    //    -----------
    //  Allocate memory for data compression
    compressedInterfA = callocMidiCompressed (formatInterfA);
    compressedInterfB = callocMidiCompressed (formatInterfB);
    compressedPhotomA = callocMidiCompressed (formatPhotomA);
    compressedPhotomB = callocMidiCompressed (formatPhotomB);
    badChannelList = (int *) calloc (formatInterfA->iXWidth, sizeof (int)); 
    filterInfo = callocFilterInfo();

    //    Compress data
    preProcKappa (options, filterInfo, fileNames, compressedInterfA, compressedInterfB, 
        compressedPhotomA, compressedPhotomB, formatInterfA, formatInterfB, formatPhotomA, 
        formatPhotomB, error);
    if (*error)
    {
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot compress FITS files");
        freeCompressedData (formatInterfA, compressedInterfA);
        freeCompressedData (formatInterfB, compressedInterfB);
        freeCompressedData (formatPhotomA, compressedPhotomA);
        freeCompressedData (formatPhotomB, compressedPhotomB);
        free (badChannelList);
        freeImageFormat (formatInterfA);
        freeImageFormat (formatInterfB);
        freeImageFormat (formatPhotomA);
        freeImageFormat (formatPhotomB);
        freeFilterInfo (filterInfo);
        return;
    }

    //    Check Delay Lines for jumps within scans
    checkDelayLineJumps ("INTERFA", formatInterfA, compressedInterfA);
    checkDelayLineJumps ("INTERFB", formatInterfB, compressedInterfB);
    checkDelayLineJumps ("PHOTOMA", formatPhotomA, compressedPhotomA);
    checkDelayLineJumps ("PHOTOMB", formatPhotomB, compressedPhotomB);

    //    Check Delay Line Consistencies
    checkDelayLineConsistencies ("INTERFA", "PHOTOMA", compressedInterfA, compressedPhotomA, 
        formatInterfA, formatPhotomA, error);
    checkDelayLineConsistencies ("INTERFB", "PHOTOMB", compressedInterfB, compressedPhotomB, 
        formatInterfB, formatPhotomB, error);
    if (*error)
    {
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Kappa calculation abandonned");
        freeCompressedData (formatInterfA, compressedInterfA);
        freeCompressedData (formatInterfB, compressedInterfB);
        freeCompressedData (formatPhotomA, compressedPhotomA);
        freeCompressedData (formatPhotomB, compressedPhotomB);
        free (badChannelList);
        freeImageFormat (formatInterfA);
        freeImageFormat (formatInterfB);
        freeImageFormat (formatPhotomA);
        freeImageFormat (formatPhotomB);
        freeFilterInfo (filterInfo);
        return;
    }

    //    Check Chopping Statistics. For QCLOG diagnostics
    //    ------------------------------------------------
    reportInterfChopping ("INTERFA", formatInterfA, compressedInterfA);
    reportInterfChopping ("INTERFB", formatInterfB, compressedInterfB);
    reportPhotomChopping ("PHOTOMA", formatPhotomA, compressedPhotomA);
    reportPhotomChopping ("PHOTOMB", formatPhotomB, compressedPhotomB);
    displayInterfChoppingDisp ("INTERFA", formatInterfA, compressedInterfA);
    displayInterfChoppingDisp ("INTERFB", formatInterfB, compressedInterfB);
    displayPhotomChoppingDisp ("PHOTOMA", formatPhotomA, compressedPhotomA);
    displayPhotomChoppingDisp ("PHOTOMB", formatPhotomB, compressedPhotomB);

    //    Remove Sky Background and display Target with Sky removed
    cleanupKappa (options->processing, formatInterfA, formatInterfB, formatPhotomA, formatPhotomB, 
        compressedInterfA, compressedInterfB, compressedPhotomA, compressedPhotomB, error);
    if (*error)
    {
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Kappa calculation abandonned");
        freeCompressedData (formatInterfA, compressedInterfA);
        freeCompressedData (formatInterfB, compressedInterfB);
        freeCompressedData (formatPhotomA, compressedPhotomA);
        freeCompressedData (formatPhotomB, compressedPhotomB);
        free (badChannelList);
        freeImageFormat (formatInterfA);
        freeImageFormat (formatInterfB);
        freeImageFormat (formatPhotomA);
        freeImageFormat (formatPhotomB);
        freeFilterInfo (filterInfo);
        return;
    }

    //    Compute the Transfer Ratios using dispersed data
    trr = callocTransferRatios (formatInterfA->iXWidth);
    midiComputeKappaCoeff (compressedInterfA, compressedInterfB, compressedPhotomA, 
        compressedPhotomB, formatInterfA, formatInterfB, formatPhotomA, formatPhotomB, trr, error);
    if (*error > 1)
    {
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Kappa calculation abandonned");
        freeCompressedData (formatInterfA, compressedInterfA);
        freeCompressedData (formatInterfB, compressedInterfB);
        freeCompressedData (formatPhotomA, compressedPhotomA);
        freeCompressedData (formatPhotomB, compressedPhotomB);
        free (badChannelList);
        freeImageFormat (formatInterfA);
        freeImageFormat (formatInterfB);
        freeImageFormat (formatPhotomA);
        freeImageFormat (formatPhotomB);
        freeTransferRatios (trr);
        freeFilterInfo (filterInfo);
        return;
    }
    
    //    Create product Transfer Ratios files
    //    ------------------------------------
    reportRejectList (formatInterfA, compressedInterfA);

/* At the moment the Kappa matrix calculation is done by midi_kappamatrix */
/* In order to avoid an empty dummy file MIDI_b1_spk.qc.dummy the next function  */
/* is not executed any more */

/*     createKappaProd (options->drsKap, fileNames, formatInterfA, trr, error); */

    //    Release memory used in this batch
    freeCompressedData (formatInterfA, compressedInterfA);
    freeCompressedData (formatInterfB, compressedInterfB);
    freeCompressedData (formatPhotomA, compressedPhotomA);
    freeCompressedData (formatPhotomB, compressedPhotomB);
    free (badChannelList);
    freeImageFormat (formatInterfA);
    freeImageFormat (formatInterfB);
    freeImageFormat (formatPhotomA);
    freeImageFormat (formatPhotomB);
    freeTransferRatios (trr);
    freeFilterInfo (filterInfo);
        
    return;
}
/*****************************************************************************/


/******************************************************************************
*               European Southern Observatory
*            VLTI MIDI Data Reduction Software
*
* Module name:  cleanupKappa
* Input/Output: See function arguments to avoid duplication
* Description:  Removes sky etc from the data
*
* History:
* 06-Dec-05     (csabet) Created
******************************************************************************/
void cleanupKappa (
    enum ProcessingMode    processing,
    ImageFormat            *formatInterfA,        // In: Interf size parameters
    ImageFormat            *formatInterfB,        // In: Interf size parameters
    ImageFormat            *formatPhotomA,        // In: PhotA size parameters
    ImageFormat            *formatPhotomB,        // In: PhotB size parameters
    CompressedData        *compressedInterfA,    // Ou: Pointer to the compressed interferometry data structure
    CompressedData        *compressedInterfB,    // Ou: Pointer to the compressed interferometry data structure
    CompressedData        *compressedPhotomA,    // Ou: Pointer to the compressed photom A data structure
    CompressedData        *compressedPhotomB,    // Ou: Pointer to the compressed photom B data structure
    int                    *error)                // Ou: Error status
{

    //    Local Declarations
    //    ------------------
    const char    routine[] = "cleanupKappa";
    int            transitions, f, X, R, n, localError;
    double        *array, fluxAve;
    float        *arrayF;
    char        *fileName, *title;

    //    Algorithm
    //    ---------
    if (diagnostic > 4)cpl_msg_info(cpl_func,"Invoking      routine   '%s' \n", routine);
    if (diagnostic > 4) fprintf(midiReportPtr, "Invoking      routine   '%s' \n", routine);

   cpl_msg_info(cpl_func,"\nCleaning up KAPPA files for batch  %d \n", batchNumber);
   cpl_msg_info(cpl_func,"--------------------------------- \n");
    fprintf (midiReportPtr, "\nCleaning up KAPPA files for batch  %d \n", batchNumber);
    fprintf (midiReportPtr, "--------------------------------- \n");

    //    Initialise
    localError = 0;
    *error = 0;
    transitions = 0;
    
    //    Check Interferometry A chopping cycle
    //    -------------------------------------    
    transitions = countTransitions ('T', formatInterfA->numOfFrames, compressedInterfA->tarType);
    sprintf (midiMessage, "Target Transitions for Interf A = %d", transitions); 
    midiReportInfo (midiReportPtr, routine, __FILE__, __LINE__, midiMessage);
    if (transitions < 10)
    {
        sprintf (midiMessage, "Interferometry A data is not chopped"); 
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, midiMessage);
        *error = 1;
    }
    
    //    Check Interferometry B chopping cycle
    //    -------------------------------------    
    transitions = countTransitions ('T', formatInterfB->numOfFrames, compressedInterfB->tarType);
    sprintf (midiMessage, "Target Transitions for Interf B = %d", transitions); 
    midiReportInfo (midiReportPtr, routine, __FILE__, __LINE__, midiMessage);
    if (transitions < 10)
    {
        sprintf (midiMessage, "Interferometry B data is not chopped"); 
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, midiMessage);
        *error = 1;
    }

    //    Check PhotomA chopping cycle
    //    ----------------------------    
    transitions = countTransitions ('T', formatPhotomA->numOfFrames, compressedPhotomA->tarType);
    sprintf (midiMessage, "Target Transitions for Photom A = %d", transitions); 
    midiReportInfo (midiReportPtr, routine, __FILE__, __LINE__, midiMessage);
    if (transitions < 10)
    {
        sprintf (midiMessage, "Photometry A data is not chopped"); 
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, midiMessage);
        *error = 1;
    }

    //    Check PhotomB chopping cycle
    //    ----------------------------
    transitions = countTransitions ('T', formatPhotomB->numOfFrames, compressedPhotomB->tarType);
    sprintf (midiMessage, "Target Transitions for Photom B = %d", transitions); 
    midiReportInfo (midiReportPtr, routine, __FILE__, __LINE__, midiMessage);
    if (transitions < 10)
    {
        sprintf (midiMessage, "Photometry B data is not chopped"); 
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, midiMessage);
        *error = 1;
    }

    //    if No error so far
    if (!(*error))
    {
        formatInterfA->chopped = 1;
        formatInterfB->chopped = 1;
        formatPhotomA->chopped = 1;
        formatPhotomB->chopped = 1;
    }
    else 
    {
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot cleanup KAPPA files");
        return;
    }
 
    //    Correct benign tarType inconsistencies
    //    --------------------------------------
    correctTarType ("INTERFA", compressedInterfA->tarType, compressedInterfA->time, formatInterfA->numOfFrames, &localError);
    if (localError) *error = 1;
    correctTarType ("INTERFB", compressedInterfB->tarType, compressedInterfB->time, formatInterfB->numOfFrames, &localError);
    if (localError) *error = 1;
    correctTarType ("PHOTOMA", compressedPhotomA->tarType, compressedPhotomA->time, formatPhotomA->numOfFrames, &localError);
    if (localError) *error = 1;
    correctTarType ("PHOTOMB", compressedPhotomB->tarType, compressedPhotomB->time, formatPhotomB->numOfFrames, &localError);
    if (*error || localError)
    {
        *error = 1;
        return;
    }
    
    //    Remove sky background
    //    ---------------------
    removeSkyBackground ("INTERFA", processing, formatInterfA, compressedInterfA, &localError);
    if (localError) *error = 1;
    removeSkyBackground ("INTERFB", processing, formatInterfB, compressedInterfB, &localError);
    if (localError) *error = 1;
    removeSkyBackground ("PHOTOMA", processing, formatPhotomA, compressedPhotomA, &localError);
    if (localError) *error = 1;
    removeSkyBackground ("PHOTOMB", processing, formatPhotomB, compressedPhotomB, &localError);
    if (formatInterfA->numOfFrames != formatPhotomA->numOfFrames || 
        formatInterfA->numOfFrames != formatPhotomB->numOfFrames ||
        formatInterfA->numOfFrames != formatInterfB->numOfFrames || *error || localError)
    {
        *error = 1;
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Unequal numOfFrames");
        return;
    }

    //    Reject frames whose flux are <= 0 or are not 'T'
    for (R = 0; R < formatInterfA->numOfRegionsToProcess; R++)
    {
        for (X = 0; X < formatInterfA->iXWidth; X++)
        {
            if (badChannelList[X])
                continue;
                    
            for (f = 0; f < formatInterfA->numOfFrames; f++)
            {
                if ((compressedInterfA->tarType[f] == 'T') && 
                    (((compressedInterfA->iDispFringe)[R])[X])[f] <= 0.0)
                {
                    compressedInterfA->rejectList[X][f] |= BSL_DATA_ERROR;
                    if (diagnostic > 3) 
                    {
                        sprintf (midiMessage, "Negative Flux at: R %d    X %3d    F %4d         InterfA %10.2f", 
                            R, X, f, (((compressedInterfA->iDispFringe)[R])[X])[f]);
                        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, midiMessage);
                    }
                }
                //    Reject frames which are not 'T'
                if (compressedInterfA->tarType[f] != 'T')
                    compressedInterfA->rejectList[X][f] |= BSL_SKY_SCAN;
            }
        }
    }

    for (R = 0; R < formatInterfB->numOfRegionsToProcess; R++)
    {
        for (X = 0; X < formatInterfB->iXWidth; X++)
        {
            if (badChannelList[X])
                continue;
                    
            for (f = 0; f < formatInterfB->numOfFrames; f++)
            {
                if ((compressedInterfB->tarType[f] == 'T') && 
                    (((compressedInterfB->iDispFringe)[R])[X])[f] <= 0.0)
                {
                    compressedInterfB->rejectList[X][f] |= BSL_DATA_ERROR;
                    if (diagnostic > 3) 
                    {
                        sprintf (midiMessage, "Negative Flux at: R %d    X %3d    F %4d         InterfB %10.2f", 
                            R, X, f, (((compressedInterfB->iDispFringe)[R])[X])[f]);
                        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, midiMessage);
                    }
                }
                if (compressedInterfB->tarType[f] != 'T')
                    compressedInterfB->rejectList[X][f] |= BSL_SKY_SCAN;
            }
        }
    }

    for (R = 0; R < formatPhotomA->numOfRegionsToProcess; R++)
    {
        for (X = 0; X < formatPhotomA->iXWidth; X++)
        {
            if (badChannelList[X])
                continue;
                    
            for (f = 0; f < formatPhotomA->numOfFrames; f++)
            {
                if ((compressedPhotomA->tarType[f] == 'T') && 
                    (((compressedPhotomA->iDispFringe)[R])[X])[f] <= 0.0)
                {
                    compressedPhotomA->rejectList[X][f] |= BSL_DATA_ERROR;
                    if (diagnostic > 3) 
                    {
                        sprintf (midiMessage, "Negative Flux at: R %d    X %3d    F %4d         PhotomA %10.2f", 
                            R, X, f, (((compressedPhotomA->iDispFringe)[R])[X])[f]);
                        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, midiMessage);
                    }
                }
                if (compressedPhotomA->tarType[f] != 'T')
                    compressedPhotomA->rejectList[X][f] |= BSL_SKY_SCAN;
            }
        }
    }

    for (R = 0; R < formatPhotomB->numOfRegionsToProcess; R++)
    {
        for (X = 0; X < formatPhotomB->iXWidth; X++)
        {
            if (badChannelList[X])
                continue;
                    
            for (f = 0; f < formatPhotomB->numOfFrames; f++)
            {
                if ((compressedPhotomB->tarType[f] == 'T') && 
                    (((compressedPhotomB->iDispFringe)[R])[X])[f] <= 0.0)
                {
                    compressedPhotomB->rejectList[X][f] |= BSL_DATA_ERROR;
                    if (diagnostic > 3) 
                    {
                        sprintf (midiMessage, "Negative Flux at: R %d    X %3d    F %4d         PhotomB %10.2f", 
                            R, X, f, (((compressedPhotomB->iDispFringe)[R])[X])[f]);
                        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, midiMessage);
                    }
                }
                if (compressedPhotomB->tarType[f] != 'T')
                    compressedPhotomB->rejectList[X][f] |= BSL_SKY_SCAN;
            }
        }
    }
    
    //    Display data with background removed
    if (plotFile && diagnostic > 1)
    {
        array = (double *) calloc (formatInterfA->numOfFrames, sizeof (double));

        fluxAve = 0.0;
        n = 0;
        for (f = 0; f < formatInterfA->numOfFrames; f++)
        {
            if (compressedInterfA->tarType[f] == 'T')
            {
                array[n] = compressedInterfA->iFringe1[f];
                fluxAve += array[n];
                n++;
            }
        }
        fluxAve /= n;
       cpl_msg_info(cpl_func,"InterfA DATA2 Average Flux = %f \n", fluxAve);
        midiCreatePlotFileDouble2D ("InterfACleanDATA2", "Interferometry A DATA2 (Sky Removed)", 
            "Frame", "Flux", 0, array, 0, n, 1);

        fluxAve = 0.0;
        n = 0;
        for (f = 0; f < formatInterfA->numOfFrames; f++)
        {
            if (compressedInterfA->tarType[f] == 'T')
            {
                array[n] = compressedInterfA->iFringe2[f];
                fluxAve += array[n];
                n++;
            }
        }
        fluxAve /= n;
       cpl_msg_info(cpl_func,"InterfA DATA3 Average Flux = %f \n", fluxAve);
        midiCreatePlotFileDouble2D ("InterfACleanDATA3", "Interferometry A DATA3 (Sky Removed)", 
            "Frame", "Flux", 0, array, 0, n, 1);

        n = 0;
        for (f = 0; f < formatInterfA->numOfFrames; f++)
        {
            if (compressedInterfA->tarType[f] == 'T')
            {
                array[n] = compressedInterfA->iFringe[f];
                n++;
            }
        }
        midiCreatePlotFileDouble2D ("InterfACleanDATA2_DATA3", "Interferometry A DATA2-DATA3 (Sky Removed)", 
            "Frame", "Flux", 0, array, 0, n, 1);

        fluxAve = 0.0;
        n = 0;
        for (f = 0; f < formatInterfB->numOfFrames; f++)
        {
            if (compressedInterfB->tarType[f] == 'T')
            {
                array[n] = compressedInterfB->iFringe1[f];
                fluxAve += array[n];
                n++;
            }
        }
        fluxAve /= n;
       cpl_msg_info(cpl_func,"InterfB DATA2 Average Flux = %f \n", fluxAve);
        midiCreatePlotFileDouble2D ("InterfBCleanDATA2", "Interferometry B DATA2 (Sky Removed)", 
            "Frame", "Flux", 0, array, 0, n, 1);

        fluxAve = 0.0;
        n = 0;
        for (f = 0; f < formatInterfB->numOfFrames; f++)
        {
            if (compressedInterfB->tarType[f] == 'T')
            {
                array[n] = compressedInterfB->iFringe2[f];
                fluxAve += array[n];
                n++;
            }
        }
        fluxAve /= n;
       cpl_msg_info(cpl_func,"InterfB DATA3 Average Flux = %f \n", fluxAve);
        midiCreatePlotFileDouble2D ("InterfBCleanDATA3", "Interferometry B DATA3 (Sky Removed)", 
            "Frame", "Flux", 0, array, 0, n, 1);

        n = 0;
        for (f = 0; f < formatInterfB->numOfFrames; f++)
        {
            if (compressedInterfB->tarType[f] == 'T')
            {
                array[n] = compressedInterfB->iFringe[f];
                n++;
            }
        }
        midiCreatePlotFileDouble2D ("InterfBCleanDATA2_DATA3", "Interferometry B DATA2-DATA3 (Sky Removed)", 
            "Frame", "Flux", 0, array, 0, n, 1);

        fluxAve = 0.0;
        n = 0;
        for (f = 0; f < formatPhotomA->numOfFrames; f++)
        {
            if (compressedPhotomA->tarType[f] == 'T')
            {
                array[n] = compressedPhotomA->iFringe1[f];
                fluxAve += array[n];
                n++;
            }
        }
        fluxAve /= n;
       cpl_msg_info(cpl_func,"PhotomA DATA1 Average Flux = %f \n", fluxAve);
        midiCreatePlotFileDouble2D ("PhotomACleanDATA1", "Photometry A DATA1 (Sky Removed)", 
            "Frame", "Flux", 0, array, 0, n, 1);

        fluxAve = 0.0;
        n = 0;
        for (f = 0; f < formatPhotomB->numOfFrames; f++)
        {
            if (compressedPhotomB->tarType[f] == 'T')
            {
                array[n] = compressedPhotomB->iFringe1[f];
                fluxAve += array[n];
                n++;
            }
        }
        fluxAve /= n;
       cpl_msg_info(cpl_func,"PhotomB DATA4 Average Flux = %f \n", fluxAve);
        midiCreatePlotFileDouble2D ("PhotomBCleanDATA4", "Photometry B DATA4 (Sky Removed)", 
            "Frame", "Flux", 0, array, 0, n, 1);

        free (array);
    }

    //    Display for each channel
    if (plotFile && diagnostic > 4)
    {
        arrayF = (float *) calloc (formatInterfA->numOfFrames, sizeof (float));
        fileName = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
        title = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
        
        for (R = 0; R < formatInterfA->numOfRegionsToProcess; R++)
        {
            for (X = 0; X < formatInterfA->iXWidth; X++)
            {
                if (badChannelList[X])
                    continue;
                    
                n = 0;
                for (f = 0; f < formatInterfA->numOfFrames; f++)
                {
                    if (compressedInterfA->tarType[f] == 'T' && !(compressedInterfA->rejectList[X][f]))
                    {
                        arrayF[n] = compressedInterfA->iDispFringe[R][X][f];
                        n++;
                    }
                }
                sprintf (fileName, "InterfACleanDATA%d_X%d", R+2, X+1);
                sprintf (title, "Interferometry A DATA%d for Channel %d (Sky Removed)", R+2, X+1);
                midiCreatePlotFile2D (fileName, title, "Frame", "Flux", 1, arrayF, 0, n, 1, 0);
            }
        }
        for (R = 0; R < formatInterfB->numOfRegionsToProcess; R++)
        {
            for (X = 0; X < formatInterfB->iXWidth; X++)
            {
                if (badChannelList[X])
                    continue;
                    
                n = 0;
                for (f = 0; f < formatInterfB->numOfFrames; f++)
                {
                    if (compressedInterfB->tarType[f] == 'T' && !(compressedInterfB->rejectList[X][f]))
                    {
                        arrayF[n] = compressedInterfB->iDispFringe[R][X][f];
                        n++;
                    }
                }
                sprintf (fileName, "InterfBCleanDATA%d_X%d", R+2, X+1);
                sprintf (title, "Interferometry B DATA%d for Channel %d (Sky Removed)", R+2, X+1);
                midiCreatePlotFile2D (fileName, title, "Frame", "Flux", 1, arrayF, 0, n, 1, 0);
            }
        }
        for (R = 0; R < formatPhotomA->numOfRegionsToProcess; R++)
        {
            for (X = 0; X < formatPhotomA->iXWidth; X++)
            {
                if (badChannelList[X])
                    continue;
                    
                n = 0;
                for (f = 0; f < formatPhotomA->numOfFrames; f++)
                {
                    if (compressedPhotomA->tarType[f] == 'T' && !(compressedPhotomA->rejectList[X][f]))
                    {
                        arrayF[n] = compressedPhotomA->iDispFringe[R][X][f];
                        n++;
                    }
                }
                sprintf (fileName, "PhotomACleanDATA1_X%d", X+1);
                sprintf (title, "Photometry A DATA1 for Channel %d (Sky Removed)", X+1);
                midiCreatePlotFile2D (fileName, title, "Frame", "Flux", 1, arrayF, 0, n, 1, 0);
            }
        }
        for (R = 0; R < formatPhotomB->numOfRegionsToProcess; R++)
        {
            for (X = 0; X < formatPhotomB->iXWidth; X++)
            {
                if (badChannelList[X])
                    continue;
                    
                n = 0;
                for (f = 0; f < formatPhotomB->numOfFrames; f++)
                {
                    if (compressedPhotomB->tarType[f] == 'T' && !(compressedPhotomB->rejectList[X][f]))
                    {
                        arrayF[n] = compressedPhotomB->iDispFringe[R][X][f];
                        n++;
                    }
                }
                sprintf (fileName, "PhotomBCleanDATA4_X%d", X+1);
                sprintf (title, "Photometry B DATA4 for Channel %d (Sky Removed)", X+1);
                midiCreatePlotFile2D (fileName, title, "Frame", "Flux", 1, arrayF, 0, n, 1, 0);
            }
        }
        free (fileName);
        free (title);
        free (arrayF);
    }

    //    Finalise rejection list
    finaliseRejectListPhotom ("Interferometry A", formatInterfA, compressedInterfA, error);
    if (*error) return;
    finaliseRejectListPhotom ("Interferometry B", formatInterfB, compressedInterfB, error);
    if (*error) return;
    finaliseRejectListPhotom ("Photometry A", formatPhotomA, compressedPhotomA, error);
    if (*error) return;
    finaliseRejectListPhotom ("Photometry B", formatPhotomB, compressedPhotomB, error);
    if (*error) return;

    return;
}
/*****************************************************************************/


/******************************************************************************
*               European Southern Observatory
*            VLTI MIDI Data Reduction Software
*
* Module name:  midiComputeKappaCoeff
* Input/Output: See function arguments to avoid duplication
* Description:  Computes the transfer matrix. The FITS files have the following
*                configuration:
*
*                Photometry A files (AOPEN) have the following configurations:
*                DATA1 = Photometry A     PA    data
*                DATA2 = Interferometry     IA1    data
*                DATA3 = Interferometry    IA2    data
*                DATA4 = NULL            
*
*                Photometry B files (BOPEN) have the following configurations:
*                DATA1 = NULL
*                DATA2 = Interferometry     IB1    data
*                DATA3 = Interferometry    IB2    data
*                DATA4 = Photometry B     PB    data
*                
*                In general we have:
*                PA = k11 * TFA + k12 * TFB
*                I1 = k21 * TFA + k22 * TFB
*                I2 = k31 * TFA + k32 * TFB
*                PB = k41 * TFA + k42 * TFB
*
*                Where
*                TFA and TFB, the total flux through the telescopes are given by:
*                TFA = PA + IA1 + IA2
*                TFB = PB + IB1 + IB2
*
*                When AOPEN then B = 0. Hence
*                k11 = PA / TFA
*                k12 = 0
*                k21 = IA1 / TFA
*                k31 = IA2 / TFA
*
*                When BOPEN then TFA = 0. Hence
*                k22 = IB1 / TFB
*                k32 = IB2 / TFB
*                k41 = 0
*                k42 = PB / TFB
*
*                And the ratios are:
*                krA1 = k21/k11
*                krA2 = k31/k11
*                krB1 = k22/k42
*                krB2 = k32/k42
*
* History:
* 06-Oct-05     (csabet) Created
******************************************************************************/
void midiComputeKappaCoeff (
    CompressedData    *compressedIA,    // In: Compressed interferometry data
    CompressedData    *compressedIB,    // In: Compressed interferometry data
    CompressedData    *compressedPA,    // In: Photometry data
    CompressedData    *compressedPB,    // In: Photometry data
    ImageFormat        *formatIA,        // In: Image size of interferometry data
    ImageFormat        *formatIB,        // In: Image size of interferometry data
    ImageFormat        *formatPA,        // In: Applies to A photom files
    ImageFormat        *formatPB,        // In: Applies to B photom files
    TransferRatios    *trr,            // Ou: Transfer Ratios
    int                *error)            // Ou: Status
{
 
    //  Local Declarations
    //    ------------------
    const char            routine[] = "midiComputeKappaCoeff";
    int                    f, x, numOfFrames, nzc, minFrames;
    float                *kappArray,    *totalFluxA, *totalFluxB, variance;
    KappaCoefficients    *kc;
    char                *string;
 
    //    Algorithm
    //    ---------
    if (diagnostic > 4)cpl_msg_info(cpl_func,"Invoking      routine   '%s' \n", routine);
    if (diagnostic > 4) fprintf(midiReportPtr, "Invoking      routine   '%s' \n", routine);
 
   cpl_msg_info(cpl_func,"\nComputing the Kappa Coefficients for batch  %d \n", batchNumber);
   cpl_msg_info(cpl_func,"------------------------------------------ \n");
    fprintf (midiReportPtr, "\nComputing the Kappa Coefficients for batch  %d \n", batchNumber);
    fprintf (midiReportPtr, "------------------------------------------ \n");
 
    //  Reset status
    *error = 0;
    
    //    Exception
    if ((formatIA->numOfFrames != formatIB->numOfFrames) || 
        (formatIA->numOfFrames != formatPA->numOfFrames) ||
        (formatIA->numOfFrames != formatPB->numOfFrames) ||
        (formatPA->numOfFrames != formatPB->numOfFrames))
    {
        *error = 2;
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Inconsistent frame counts");
        return;
    }
    numOfFrames = formatIA->numOfFrames;
    minFrames = MIN_FRAMES_PER_SCAN * compressedIA->numOfTargetChops;


    //    Allocate memory
    kappArray = (float *) calloc (numOfFrames, sizeof (float));
    totalFluxA = (float *) calloc (numOfFrames, sizeof (float));
    totalFluxB = (float *) calloc (numOfFrames, sizeof (float));
    string = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
    kc = callocKappaCoefficients (formatIA->iXWidth);
    
    //    Compute kappa coefficients for each wavelength
    //    ----------------------------------------------
   cpl_msg_info(cpl_func,"Channel    k11 sig11     k21 sig21     k31 sig31     k22 sig22     k32 sig32     k42 sig42 \n");
    fprintf (midiReportPtr, "Channel    k11 sig11     k21 sig21     k31 sig31     k22 sig22     k32 sig32     k42 sig42 \n");
    for (x = 0; x < formatIA->iXWidth; x++)
    {
        if (badChannelList[x])
        {
           cpl_msg_info(cpl_func,"%3d    <-- %s --> \n", x, UNAV);
            fprintf (midiReportPtr, "%3d    <-- %s --> \n", x, UNAV);
            continue;
        }
            
        //    Compute total flux through each telescope
        for (f = 0; f < numOfFrames; f++)
        {
            if ((compressedIA->tarType[f] == 'T') && 
                (compressedPA->tarType[f] == 'T') &&
                !(compressedIA->rejectList[x][f]) &&
                !(compressedPA->rejectList[x][f]))
            {
                totalFluxA[f] = compressedIA->iDispFringe[0][x][f] + compressedIA->iDispFringe[1][x][f] + 
                    compressedPA->iDispFringe[0][x][f];
                
            }
            if ((compressedIB->tarType[f] == 'T') && 
                (compressedPB->tarType[f] == 'T') && 
                !(compressedIB->rejectList[x][f]) &&
                !(compressedPB->rejectList[x][f]))
            {
                totalFluxB[f] = compressedIB->iDispFringe[0][x][f] + compressedIB->iDispFringe[1][x][f] + 
                    compressedPB->iDispFringe[0][x][f];
            }
        }
        
        //    kappa11 computation
        nzc = -1;
        for (f = 0; f < numOfFrames; f++)
        {
            if ((compressedPA->tarType[f] == 'T') && 
                (((compressedPA->iDispFringe[0])[x])[f] > 0.0) &&
                (totalFluxA[f] > 0.0) && 
                !(compressedPA->rejectList[x][f]))
            {
                nzc++;
                kappArray[nzc] = ((compressedPA->iDispFringe[0])[x])[f] / totalFluxA[f];
            }
        }
        if (nzc > minFrames)
        {
            kc->k11[x] = signalMean (kappArray, 0, nzc);
            variance = signalVariance (kappArray, 0, nzc, &(kc->sig11[x]));
            if (plotFile && diagnostic > 2)
            {
                sprintf (string, "Kappa11 Array, Channel %d", x);
                midiCreatePlotFile2D ("Kappa11Array", string, "Frame", "Flux", 1, kappArray, 0, nzc, 1, 1);
            }
        }
        else
        {
            badChannelList[x] |= BSL_DATA_ERROR;
           cpl_msg_info(cpl_func,"%3d    <-- %s --> \n", x, UNAV);
            fprintf (midiReportPtr, "%3d    <-- %s --> \n", x, UNAV);
            continue;
        }

        //    kappa21 computation
        nzc = -1;
        for (f = 0; f < numOfFrames; f++)
        {
            if ((compressedIA->tarType[f] == 'T') && 
                (((compressedIA->iDispFringe[0])[x])[f] > 0.0) &&
                (totalFluxA[f] > 0.0) &&
                !(compressedIA->rejectList[x][f]))
            {
                nzc++;
                kappArray[nzc] = ((compressedIA->iDispFringe[0])[x])[f] / totalFluxA[f];
            }
        }
        if (nzc > minFrames)
        {
            kc->k21[x] = signalMean (kappArray, 0, nzc);
            variance = signalVariance (kappArray, 0, nzc, &(kc->sig21[x]));
            if (plotFile && diagnostic > 2)
            {
                sprintf (string, "Kappa21 Array, Channel %d", x);
                midiCreatePlotFile2D ("Kappa21Array", string, "Frame", "Flux", 1, kappArray, 0, nzc, 1, 1);
            }
        }
        else
        {
            badChannelList[x] |= BSL_DATA_ERROR;
           cpl_msg_info(cpl_func,"%3d    <-- %s --> \n", x, UNAV);
            fprintf (midiReportPtr, "%3d    <-- %s --> \n", x, UNAV);
            continue;
        }

        //    kappa31 computation
        nzc = -1;
        for (f = 0; f < numOfFrames; f++)
        {
            if ((compressedIA->tarType[f] == 'T') && 
                (((compressedIA->iDispFringe[1])[x])[f] > 0.0) &&
                (totalFluxA[f] > 0.0) &&
                !(compressedIA->rejectList[x][f]))
            {
                nzc++;
                kappArray[nzc] = ((compressedIA->iDispFringe[1])[x])[f] / totalFluxA[f];
            }
        }
        if (nzc > minFrames)
        {
            kc->k31[x] = signalMean (kappArray, 0, nzc);
            variance = signalVariance (kappArray, 0, nzc, &(kc->sig31[x]));
            if (plotFile && diagnostic > 2)
            {
                sprintf (string, "Kappa31 Array, Channel %d", x);
                midiCreatePlotFile2D ("Kappa31Array", string, "Frame", "Flux", 1, kappArray, 0, nzc, 1, 1);
            }
        }
        else
        {
            badChannelList[x] |= BSL_DATA_ERROR;
           cpl_msg_info(cpl_func,"%3d    <-- %s --> \n", x, UNAV);
            fprintf (midiReportPtr, "%3d    <-- %s --> \n", x, UNAV);
            continue;
        }

        //    kappa22 computation
        nzc = -1;
        for (f = 0; f < numOfFrames; f++)
        {
            if ((compressedIB->tarType[f] == 'T') && 
                (((compressedIB->iDispFringe[0])[x])[f] > 0.0) &&
                (totalFluxB[f] > 0.0) &&
                !(compressedIB->rejectList[x][f]))
            {
                nzc++;
                kappArray[nzc] = ((compressedIB->iDispFringe[0])[x])[f] / totalFluxB[f];
            }
        }
        if (nzc > minFrames)
        {
            kc->k22[x] = signalMean (kappArray, 0, nzc);
            variance = signalVariance (kappArray, 0, nzc, &(kc->sig22[x]));
            if (plotFile && diagnostic > 2)
            {
                sprintf (string, "Kappa22 Array, Channel %d", x);
                midiCreatePlotFile2D ("Kappa22Array", string, "Frame", "Flux", 1, kappArray, 0, nzc, 1, 1);
            }
        }
        else
        {
            badChannelList[x] |= BSL_DATA_ERROR;
           cpl_msg_info(cpl_func,"%3d    <-- %s --> \n", x, UNAV);
            fprintf (midiReportPtr, "%3d    <-- %s --> \n", x, UNAV);
            continue;
        }

        //    kappa32 computation
        nzc = -1;
        for (f = 0; f < numOfFrames; f++)
        {
            if ((compressedIB->tarType[f] == 'T') && 
                (((compressedIB->iDispFringe[1])[x])[f] > 0.0) &&
                (totalFluxB[f] > 0.0) &&
                !(compressedIB->rejectList[x][f]))
            {
                nzc++;
                kappArray[nzc] = ((compressedIB->iDispFringe[1])[x])[f] / totalFluxB[f];
            }
        }
        if (nzc > minFrames)
        {
            kc->k32[x] = signalMean (kappArray, 0, nzc);
            variance = signalVariance (kappArray, 0, nzc, &(kc->sig32[x]));
            if (plotFile && diagnostic > 2)
            {
                sprintf (string, "Kappa32 Array, Channel %d", x);
                midiCreatePlotFile2D ("Kappa32Array", string, "Frame", "Flux", 1, kappArray, 0, nzc, 1, 1);
            }
        }
        else
        {
            badChannelList[x] |= BSL_DATA_ERROR;
           cpl_msg_info(cpl_func,"%3d    <-- %s --> \n", x, UNAV);
            fprintf (midiReportPtr, "%3d    <-- %s --> \n", x, UNAV);
            continue;
        }

        //    kappa42 computation
        nzc = -1;
        for (f = 0; f < numOfFrames; f++)
        {
            if ((compressedPB->tarType[f] == 'T') && 
                (((compressedPB->iDispFringe[0])[x])[f] > 0.0) &&
                (totalFluxB[f] > 0.0) &&
                !(compressedPB->rejectList[x][f]))
            {
                nzc++;
                kappArray[nzc] = ((compressedPB->iDispFringe[0])[x])[f] / totalFluxB[f];
            }
        }
        if (nzc > minFrames)
        {
            kc->k42[x] = signalMean (kappArray, 0, nzc);
            variance = signalVariance (kappArray, 0, nzc, &(kc->sig42[x]));
            if (plotFile && diagnostic > 2)
            {
                sprintf (string, "Kappa42 Array, Channel %d", x);
                midiCreatePlotFile2D ("Kappa42Array", string, "Frame", "Flux", 1, kappArray, 0, nzc, 1, 1);
            }
        }
        else
        {
            badChannelList[x] |= BSL_DATA_ERROR;
           cpl_msg_info(cpl_func,"%3d    <-- %s --> \n", x, UNAV);
            fprintf (midiReportPtr, "%3d    <-- %s --> \n", x, UNAV);
            continue;
        }
       cpl_msg_info(cpl_func,"%3d        %1.2f %1.2f     %1.2f %1.2f     %1.2f %1.2f     %1.2f %1.2f     %1.2f %1.2f     %1.2f %1.2f  \n", 
            x, kc->k11[x], kc->sig11[x], kc->k21[x], kc->sig21[x], kc->k31[x], kc->sig31[x], kc->k22[x], kc->sig22[x], 
            kc->k32[x], kc->sig32[x], kc->k42[x], kc->sig42[x]);
        fprintf (midiReportPtr, "%3d        %1.2f %1.2f     %1.2f %1.2f     %1.2f %1.2f     %1.2f %1.2f     %1.2f %1.2f     %1.2f %1.2f  \n", 
            x, kc->k11[x], kc->sig11[x], kc->k21[x], kc->sig21[x], kc->k31[x], kc->sig31[x], kc->k22[x], kc->sig22[x], 
            kc->k32[x], kc->sig32[x], kc->k42[x], kc->sig42[x]);
    }

    //    Smooth the coefficients. Not sure if this is the right moment for filtering. A
    //    more suitable place would probably be filtering "kappArray[nzc]" at frame ensemble level.
    smoothKappaCoefficiets (formatIA, kc, error);
    if (*error)
    {
        freeKappaCoefficients (kc);
        free (kappArray);
        free (totalFluxA);
        free (totalFluxB);
        free (string);
        return;
    }
    
    //    Now compute the ratios
    computeTransferRatios (formatIA, kc, trr, error);

    //    Release memory
    freeKappaCoefficients (kc);
    free (kappArray);
    free (totalFluxA);
    free (totalFluxB);
    free (string);

    return;
}
/*****************************************************************************/



/******************************************************************************
*               European Southern Observatory
*            VLTI MIDI Data Reduction Software
*
* Module name:  computeTransferRatios
* Input/Output: See function arguments to avoid duplication
* Description:  Computes the transfer ratios and the associated errors as follows:
*
*                krA1 = k21/k11
*                krA2 = k31/k11
*                krB1 = k22/k42
*                krB2 = k32/k42
*
* History:
* 02-Feb-06     (csabet) Created
******************************************************************************/
void computeTransferRatios (
    ImageFormat            *format,    // In: Image size of interferometry data
    KappaCoefficients    *kc,        // In: Kappa Coefficients
    TransferRatios        *tr,        // Ou: Transfer Ratios
    int                    *error)        // Ou: Status
{
 
    //  Local Declarations
    //    ------------------
    const char    routine[] = "computeTransferRatios";
    int            x, n;
     
    //    Algorithm
    //    ---------
    if (diagnostic > 4)cpl_msg_info(cpl_func,"Invoking      routine   '%s' \n", routine);
    if (diagnostic > 4) fprintf(midiReportPtr, "Invoking      routine   '%s' \n", routine);
 
    //    Reset
    *error = 0;
    tr->exists = 0;
    
    //    Check sanity. Kappa coefficients must lie between 0.0 and 1.0 exclusively
   cpl_msg_info(cpl_func,"Channel    ka1 siga1         ka2 siga2         kb1 sigb1         kb2 sigb2 \n");
    fprintf (midiReportPtr, "Channel    ka1 siga1         ka2 siga2         kb1 sigb1         kb2 sigb2 \n");
    n = 0;
    for (x = 0; x < format->iXWidth; x++)
    {
        if (badChannelList[x])
        {
           cpl_msg_info(cpl_func,"%3d    <-- %s --> \n", x, UNAV);
            fprintf (midiReportPtr, "%3d    <-- %s --> \n", x, UNAV);
            continue;
        }
            
        if (((kc->k11[x] <= 0.0) || (kc->k11[x] >= 1.0)) || 
            ((kc->k21[x] <= 0.0) || (kc->k21[x] >= 1.0)) ||
            ((kc->k31[x] <= 0.0) || (kc->k31[x] >= 1.0)) || 
            ((kc->k22[x] <= 0.0) || (kc->k22[x] >= 1.0)) ||
            ((kc->k32[x] <= 0.0) || (kc->k32[x] >= 1.0)) || 
            ((kc->k42[x] <= 0.0) || (kc->k42[x] >= 1.0)))
        {
            badChannelList[x] |= BSL_DATA_ERROR;
            if (diagnostic > 1)
            {
                sprintf (midiMessage, 
                    "Cannot compute Transfer Ratios for Channel %3d. Values of Kappa Matrix are unreasonable", x);
                midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, midiMessage);
            }
           cpl_msg_info(cpl_func,"%3d    <-- %s --> \n", x, UNAV);
            fprintf (midiReportPtr, "%3d    <-- %s --> \n", x, UNAV);
        }
        else
        {
            n++;
            tr->ka1[x] = kc->k21[x] / kc->k11[x];
            tr->siga1[x] = kc->sig21[x] / fabs(kc->k11[x]) + kc->sig11[x] * (fabs(kc->k21[x]) / (kc->k11[x] * kc->k11[x]));
            tr->ka2[x] = kc->k31[x] / kc->k11[x];
            tr->siga2[x] = kc->sig31[x] / fabs(kc->k11[x]) + kc->sig11[x] * (fabs(kc->k31[x]) / (kc->k11[x] * kc->k11[x]));
            tr->kb1[x] = kc->k22[x] / kc->k42[x];
            tr->sigb1[x] = kc->sig22[x] / fabs(kc->k42[x]) + kc->sig42[x] * (fabs(kc->k22[x]) / (kc->k42[x] * kc->k42[x]));
            tr->kb2[x] = kc->k32[x] / kc->k42[x];
            tr->sigb2[x] = kc->sig32[x] / fabs(kc->k42[x]) + kc->sig42[x] * (fabs(kc->k32[x]) / (kc->k42[x] * kc->k42[x]));

           cpl_msg_info(cpl_func,"%3d        %6.2f %6.2f      %6.2f %6.2f      %6.2f %6.2f      %6.2f %6.2f \n", 
                x, tr->ka1[x], tr->siga1[x], tr->ka2[x], tr->siga2[x], tr->kb1[x], tr->sigb1[x], tr->kb2[x], tr->sigb2[x]);
            fprintf (midiReportPtr, "%3d        %6.2f %6.2f      %6.2f %6.2f      %6.2f %6.2f      %6.2f %6.2f \n",
                x, tr->ka1[x], tr->siga1[x], tr->ka2[x], tr->siga2[x], tr->kb1[x], tr->sigb1[x], tr->kb2[x], tr->sigb2[x]);
        }
    }

    if (!n)
    {
        tr->exists = 0;
        *error = 1;
        return;
    }    

    if (plotFile && diagnostic > 1)
    {
        midiCreatePlotFile2D ("Ka1Ratio", "Ka1 Ratio", "Channel", "Flux", 1, tr->ka1, 0, format->iXWidth, 1, 0);
        midiCreatePlotFile2D ("Ka1Sig", "Ka1 Error", "Channel", "Flux", 1, tr->siga1, 0, format->iXWidth, 1, 0);
 
        midiCreatePlotFile2D ("Ka2Ratio", "Ka2 Ratio", "Channel", "Flux", 1, tr->ka2, 0, format->iXWidth, 1, 0);
        midiCreatePlotFile2D ("Ka2Sig", "Ka2 Error", "Channel", "Flux", 1, tr->siga2, 0, format->iXWidth, 1, 0);
 
        midiCreatePlotFile2D ("Kb1Ratio", "Kb1 Ratio", "Channel", "Flux", 1, tr->kb1, 0, format->iXWidth, 1, 0);
        midiCreatePlotFile2D ("Kb1Sig", "Kb1 Error", "Channel", "Flux", 1, tr->sigb1, 0, format->iXWidth, 1, 0);
 
        midiCreatePlotFile2D ("Kb2Ratio", "Kb2 Ratio", "Channel", "Flux", 1, tr->kb2, 0, format->iXWidth, 1, 0);
        midiCreatePlotFile2D ("Kb2Sig", "Kb2 Error", "Channel", "Flux", 1, tr->sigb2, 0, format->iXWidth, 1, 0);
    }
    
    tr->exists = 1;
    
    return;
}
/*****************************************************************************/
 
 
/******************************************************************************
*               European Southern Observatory
*            VLTI MIDI Data Reduction Software
*
* Module name:  smoothKappaCoefficiets
* Input/Output: See function arguments to avoid duplication
* Description:  Not sure if this routine is required. A better method would
*                probably be to filter the kappa elements at frame enseble stage.
*                Then compute the associated errors and propagate the errors to 
*                the ratios.
*
* History:
* 07-Mar-06     (csabet) Created
******************************************************************************/
void smoothKappaCoefficiets (
    ImageFormat            *format,    // In: Image size of interferometry data
    KappaCoefficients    *kc,        // In: Kappa Coefficients
    int                    *error)
{
 
    //  Local Declarations
    //    ------------------
    const char    routine[] = "smoothKappaCoefficiets";
    int            x, n;
     
    //    Algorithm
    //    ---------
    if (diagnostic > 4)cpl_msg_info(cpl_func,"Invoking      routine   '%s' \n", routine);
    if (diagnostic > 4) fprintf(midiReportPtr, "Invoking      routine   '%s' \n", routine);

    //    Reset
    *error = 0;
     
    //    First reject noisy values
    n = 0;
    for (x = 0; x < format->iXWidth; x++)
    {
        if (!(badChannelList[x]))
        {
            if ((kc->sig11[x] > KAPPA_NOISE_LIMIT * kc->k11[x]) || (kc->sig21[x] > KAPPA_NOISE_LIMIT * kc->k21[x]) ||
                (kc->sig31[x] > KAPPA_NOISE_LIMIT * kc->k31[x]) || (kc->sig22[x] > KAPPA_NOISE_LIMIT * kc->k22[x]) ||
                (kc->sig32[x] > KAPPA_NOISE_LIMIT * kc->k32[x]) || (kc->sig42[x] > KAPPA_NOISE_LIMIT * kc->k42[x]))
                badChannelList[x] |= BSL_DATA_ERROR;
            else
                n++;
        }
    }

    //    Plot
    if (plotFile && diagnostic > 1 && n)
    {
        midiCreatePlotFile2D ("Kappa11VersX", "k11 Versus X", "X", "k11", 1, kc->k11, 0, format->iXWidth, 1, 0);
        midiCreatePlotFile2D ("Kappa21VersX", "k21 Versus X", "X", "k21", 1, kc->k21, 0, format->iXWidth, 1, 0);
        midiCreatePlotFile2D ("Kappa31VersX", "k31 Versus X", "X", "k31", 1, kc->k31, 0, format->iXWidth, 1, 0);
        midiCreatePlotFile2D ("Kappa22VersX", "k22 Versus X", "X", "k22", 1, kc->k22, 0, format->iXWidth, 1, 0);
        midiCreatePlotFile2D ("Kappa32VersX", "k32 Versus X", "X", "k32", 1, kc->k32, 0, format->iXWidth, 1, 0);
        midiCreatePlotFile2D ("Kappa42VersX", "k42 Versus X", "X", "k42", 1, kc->k42, 0, format->iXWidth, 1, 0);
    }

    if (!n)
    {
        *error = 1;
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "High error bars on all channels");
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Computation of the Kappa Coefficients is abandonned");
        return;
    }

    for (x = 0; x < format->iXWidth; x++)
    {
        //    If it is other than a masked error then no kappa matrix should be computed
        if (badChannelList[x] & 1023)
        {
            *error = 1;
            return;
        }
    }
    
    //    Now smooth the result. (Not sure yet if filtering is wise)    
    midiReportTbd (midiReportPtr, routine, __FILE__, __LINE__, "Routine TBD");
    
    return;
}
/*****************************************************************************/
