
/******************************************************************************
*******************************************************************************
*               European Southern Observatory
*          VLTI MIDI Maintenance Templates Software
*
* Module name:  procRefPix.c
* Description:  Contains routines for processing the templates
*
* History:      
* 16-Jun-04     (csabet) created
*******************************************************************************
******************************************************************************/

/******************************************************************************
*   Compiler directives
******************************************************************************/

/******************************************************************************
*   Include files
******************************************************************************/
#include <math.h>
#include <stdio.h>
#include <cpl.h>
#include "midiGlobal.h"
#include "midiLib.h"
#include "imageProcessing.h"
#include "memoryHandling.h"
#include "createProdRefPix.h"
#include "procRefPix.h"
#include "errorHandling.h"
#include "midiFitsUtility.h"
#include "fitsAnalysisTec.h"
#include "diagnostics.h"
#include "qfits.h"

/**********************************************************
*   Constant definitions
**********************************************************/

/**********************************************************
*   Global Variables 
**********************************************************/

/*============================ C O D E    A R E A ===========================*/



/******************************************************************************
*               European Southern Observatory
*          VLTI MIDI Maintenance Templates Software
*
* Module name:  procRefPix
* Input/Output: See function arguments to avoid duplication
* Description:  Evaluates the reference positions of the VLTI beams on the MIDI
*                detector for the fine positioning of the target. This is the main 
*                routine for this application.
*
* History:      
* 03-May-05     (csabet) Created
******************************************************************************/
void procRefPix (
    MidiFiles    *fileNames,    // In: Pointer to file names
    int            *error)        // Ou: Error status
{

    //    Local Declarations
    //    ------------------
    const char      routine[] = "procRefPix";
    ImageFormat        *format=NULL;
    ReferencePixel     *refPix=NULL;
    int                numOfExposures, maxNumOfBeams;
    FILE            *signaturePtr=NULL;
    
    //    Algorithm
    //    ---------
    if (diagnostic > 4)cpl_msg_info(cpl_func,"Invoking      routine   '%s' \n", routine);
    if (diagnostic > 4) fprintf (midiReportPtr, "Invoking      routine   '%s' \n", routine);

    //    Write a signature
    signaturePtr = fopen ("MIDI_sig_pix.log", "w");
    fclose (signaturePtr);

    //    Reset status
    *error = 0;
    numOfExposures = 0;
    maxNumOfBeams = 0;
    
    //    Allocate memory
    format = callocImageFormat ();
    
    analyseFitsRefPix (fileNames, format, &numOfExposures, &maxNumOfBeams, error);
    if (*error)
    {
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot analyse REFPIX");
        freeImageFormat (format);
        return;
    }
    
    refPix = callocRefPix (numOfExposures, maxNumOfBeams);
    computeRefPix (fileNames, refPix, error);
    if (*error)
    {
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot process REFPIX");
        freeRefPix (refPix);
        freeImageFormat (format);
        return;
    }
    
    createRefPixProd (fileNames, format, refPix, error);
    if (*error)    midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot create REFPIX products");

    //    Release memory
    freeRefPix (refPix);
    freeImageFormat (format);

    return; 
}
/*****************************************************************************/



/******************************************************************************
*               European Southern Observatory
*          VLTI MIDI Maintenance Templates Software
*
* Module name:  computeRefPix
* Input/Output: See function arguments to avoid duplication
* Description:  Evaluates the reference positions of the VLTI beams on the MIDI
*                detector for the fine positioning of the target. It also monitors
*                and evaluates the stability of the beams inside the MIDI dewar
*                For every frame of the beam compute the coordinates of the brightest
*                pixel (x1, y1). This is found by a 2D gaussian fit or any suitable
*                centroiding method. The output for beam 1 (I1) is N coordinates
*                where N is the number of frames. Carry out the same for beam 2 (I2)
*                and obtain a second set of coordinates. A number of correlation may be
*                carried out to assess the quality of the output. If DATA1 and DATA2
*                are combined into DATA1 then we need to know the sub-image coordinates.
*
* History:      
* 11-Aug-04     (csabet) Created
******************************************************************************/
void computeRefPix (
    MidiFiles        *fileNames,    // In: Pointer to the MIDI file structure
    ReferencePixel     *refPix,    // Ou: Pointer to the reference pixel data structure
    int                *error)        // Ou:    Error status
{
  
    //    Local Declarations
    //    ------------------
    const char  routine[] = "computeRefPix";
    char                        *fileTemp, *classification;
    FILE                    *inFitsBatchPtr;
    ImageFormat             *format;
    int                     localError, fileNumber, extNumOfImagingDataFile;
    
    //    Algorithm
    //    ---------
    if (diagnostic > 4)cpl_msg_info(cpl_func,"Invoking      routine   '%s' \n", routine);
    if (diagnostic > 4) fprintf (midiReportPtr, "Invoking      routine   '%s' \n", routine);

    //    Allocate memory
    classification = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
    fileTemp = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
    format = callocImageFormat ();

    //    Reset status
    *error = 0;
    localError = 0;
    fileNumber = 0;
    refPix->exists = 0;
    format->hasData = 0;
    
    //    Open the list of files
    if ((inFitsBatchPtr = fopen (fileNames->inFitsBatch, "r")) == NULL)
    {
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, 
            "Cannot open input FITS file list\n                 No data preprocessing has been carried out for this batch");
        freeImageFormat (format);
        free (fileTemp);
        free (classification);
        *error = 1;
        return;
    }

    //    Loop through the files and analyse
    while (fgets (fileTemp, MAX_STRING_LENGTH, inFitsBatchPtr) != NULL)
    {
        sprintf (classification, "%s", "");
        sscanf (fileTemp, "%s%s", fileNames->inFitsName, classification);
        if (diagnostic)cpl_msg_info(cpl_func,"\n   Processing file   %s \n", fileNames->inFitsName);
        fprintf(midiReportPtr, "\n   Processing file   %s \n", fileNames->inFitsName);

        //    Get 'extNumOfImagingDataFile' extension number of IMAGING_DATA in input file
        extNumOfImagingDataFile  = findImagingDataExtension (fileNames->inFitsName, TAB_IMAGING_DATA, &localError);
        if (localError) 
        {
            *error = 1;
            break;
        }

        //    Get Image Format parameters
        if (extNumOfImagingDataFile > 0)
        {
            getImageFormat (fileNames->inFitsName, extNumOfImagingDataFile, format, &localError);
            if (localError) 
            {
                *error = 1;
                break;
            }
        }
        else format->hasData = 0;

        //    Check if the file has data
        if (format->hasData)
        {
            /*  Check Categ, Tech and Type and then compute the image size */
            if ((strcmp (format->obsCatg, "CALIB") == 0) &&
            (strcmp (format->obsTech, "IMAGE") == 0) &&
            (strcmp (format->obsType, "FMTCHCK") == 0))
            {
                //    Save beam combiner and sutter ID
                sprintf (refPix->exposure[fileNumber].shutterId, "%s", format->shutterId);
                sprintf (refPix->exposure[fileNumber].beamCombiner, "%s", format->beamCombiner);
                if ((strcmp (format->beamCombiner, "HIGH_SENS") == 0) || 
                    (strcmp (format->beamCombiner, "OPEN") == 0))
                    refPix->exposure[fileNumber].numOfBeams = 2;
                else if (strcmp (format->beamCombiner, "SCI_PHOT") == 0) 
                    refPix->exposure[fileNumber].numOfBeams = 3;

                if (diagnostic)cpl_msg_info(cpl_func,"   Beam combiner for file %d      = %s\n", fileNumber+1, 
                    refPix->exposure[fileNumber].beamCombiner);
                if (diagnostic)cpl_msg_info(cpl_func,"   Shutter ID for file %d         = %s\n", fileNumber+1, 
                    refPix->exposure[fileNumber].shutterId);
                if (diagnostic)cpl_msg_info(cpl_func,"   Number of beams for file %d    = %d\n", fileNumber+1, 
                    refPix->exposure[fileNumber].numOfBeams);
                fprintf (midiReportPtr, "   Beam combiner for file %d      = %s\n", fileNumber+1, 
                    refPix->exposure[fileNumber].beamCombiner);
                fprintf (midiReportPtr, "   Shutter ID for file %d         = %s\n", fileNumber+1, 
                    refPix->exposure[fileNumber].shutterId);
                fprintf (midiReportPtr, "   Number of beams for file %d    = %d\n", fileNumber+1, 
                    refPix->exposure[fileNumber].numOfBeams);

                //    Calculate Image Centroids
                procCentroidsRefPix (fileNumber, fileNames->inFitsName, extNumOfImagingDataFile, 
                    format, refPix, &localError);
                if (localError) *error = 1;
                        
                //    Set the flags
                refPix->exists = 1;
                
                //    Increment file number but make sure it is not greater than the allocated
                fileNumber++;
                if (fileNumber > refPix->numOfExposures)
                {
                    *error = 1;
                    break;
                }                
            }
            else
                midiReportInfo (midiReportPtr, routine, __FILE__, __LINE__, "The above file is not suitable for this task");
        }
        else
        {
            if (diagnostic)
            {
                sprintf (midiMessage, "No data tables in %s. Not processed", fileNames->inFitsName);
                midiReportInfo (midiReportPtr, routine, __FILE__, __LINE__, midiMessage);
            }
        }
    }

    //    Check if processing has been carried out
    if ((refPix->exists) && !(*error) && (fileNumber == refPix->numOfExposures))
    {
       cpl_msg_info(cpl_func,"\nReference Pixel Position Inventry: \n");
       cpl_msg_info(cpl_func,"================================= \n");
       cpl_msg_info(cpl_func,"   Expected number of data files  = %d\n", refPix->numOfExposures);
       cpl_msg_info(cpl_func,"   Number of data files processed = %d\n", fileNumber);
       cpl_msg_info(cpl_func,"\n");

        fprintf (midiReportPtr, "\nReference Pixel Position Inventry: \n");
        fprintf (midiReportPtr, "================================= \n");
        fprintf (midiReportPtr, "   Expected number of data files  = %d\n", refPix->numOfExposures);
        fprintf (midiReportPtr, "   Number of data files processed = %d\n", fileNumber);
        fprintf (midiReportPtr, "\n");

        //    Beam analysis
        //    -------------
        evaluateBeamPosition ();
    }

    if (*error || localError)
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot create an Image Centroid Data for this batch");

    //    Close the file list
    fclose (inFitsBatchPtr);
    
    //    Release memory
    freeImageFormat (format);
    free (fileTemp);
    free (classification);    
    
    return; 
}
/*****************************************************************************/



/******************************************************************************
*               European Southern Observatory
*          VLTI MIDI Maintenance Templates Software
*
* Module name:  procCentroidsRefPix
* Input/Output: See function arguments to avoid duplication
* Description:  Main routine for computing the averaged image and the coordinates of 
*                the targets
*                The structure of the centroid data is as follows:
*                refPix->centrod[0]    beam 1 of OPEN        file 1    ABOPEN (Imaging)
*                refPix->centrod[1]    beam 2 of OPEN        file 1    ABOPEN (Imaging)
*                refPix->centrod[2]    beam 1 of HIGH_SENS    file 2    AOPEN
*                refPix->centrod[3]    beam 2 of HIGH_SENS    file 2    AOPEN
*                refPix->centrod[4]    beam 1 of HIGH_SENS    file 3    BOPEN
*                refPix->centrod[5]    beam 2 of HIGH_SENS    file 3    BOPEN
*                refPix->centrod[6]    beam 1 of SCI_PHOT    file 4    AOPEN
*                refPix->centrod[7]    beam 2 of SCI_PHOT    file 4    AOPEN
*                refPix->centrod[8]    beam 3 of SCI_PHOT    file 4    AOPEN
*                refPix->centrod[9]    beam 1 of SCI_PHOT    file 5    BOPEN
*                refPix->centrod[10]    beam 2 of SCI_PHOT    file 5    BOPEN
*                refPix->centrod[11]    beam 3 of SCI_PHOT    file 5    BOPEN
*
*
* History:      
* 11-Aug-04     (csabet) Created
* 28-Aug-06     (csabet) Modified
******************************************************************************/
void procCentroidsRefPix ( 
    int                fileNumber,        //    In:    File number
    char            *inFitsName,    //  In: Name of file to process
    int             extensionNumber,//    In:    Extension number of the IMAGE_DATA
    ImageFormat     *format,        //    In:    Pointer to the image format
    ReferencePixel     *refPix,        //  Ou: Pointer to the reference pixel data structure
    int                *error)            //    Ou:    Error status

{
  
    //  Local Declarations
    //    ------------------
    const char  routine[] = "procCentroidsRefPix";
    qfits_table *pTable  = NULL;
    short int   *inData;
    char        *tempStr, *string, *title, *dataName;
    int         i, foundData = 0, scalingOffset, indexData, subWindowSize;
    float        *aveImage;


    //  Algorithm
    //    ---------
    if (diagnostic > 4)cpl_msg_info(cpl_func,"Invoking      routine   '%s' \n", routine);
    if (diagnostic > 4) fprintf (midiReportPtr, "Invoking      routine   '%s' \n", routine);
    
    //    Reset status
    *error = 0;
    
    //  Open IMAGING_DATA = INDEX 2
    pTable = qfits_table_open (inFitsName, extensionNumber);
    if (!pTable)
    {
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot load IMAGING_DATA");
        *error = 1;
        return;
    }
        
    //    Get data table information
    for (i = 0; i < pTable->nc; i++)
    {
        if (strcmp (pTable->col[i].tlabel, "DATA1") == 0)
        {
            foundData = 1;
            indexData = i;
        }
    }
    if (foundData == 0)
    {
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot find requested columns in data FITS file");
        qfits_table_close (pTable);
        *error = 1;
        return;
    }

    //  Read column DATA
    inData = (short int*) qfits_query_column (pTable, indexData, NULL); 

    //    Read Scaling Offset
    dataName = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
    for (i = 14; i < 25; i++)
    {
        sprintf (dataName, "TZERO%d", i);
        tempStr = qfits_query_ext (inFitsName, dataName, extensionNumber);
        if (tempStr != NULL)
        {
            if (diagnostic)cpl_msg_info(cpl_func,"Scaling Offset = %s\n", tempStr);
            if (diagnostic) fprintf (midiReportPtr, "Scaling Offset = %s\n", tempStr);
            sscanf (tempStr, "%d", &scalingOffset);
            break;
        }
    }
    if (tempStr == NULL)
    {
        scalingOffset = 0;
        midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot read Scaling Offset. It is set to 0");
    }
    free (dataName);
    
    //    Create averaged image for the whole file. This also allows a check on existence of a target
    subWindowSize = format->iXWidth * format->iYWidth;
    aveImage = (float *) calloc (subWindowSize, sizeof(float));
    createAveragedImage (inData, scalingOffset,    format, aveImage);
    
    //    Create an image FITS file
    title = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
    string = (char *) calloc (MAX_STRING_LENGTH, sizeof (char));
    sprintf (title, "AveImg%d", fileNumber+1);
    sprintf (string, "file %d", fileNumber+1);
    createFitsImage (string, title, inFitsName, format->iXWidth, format->iYWidth, aveImage);
    
    //    Create plot file
    sprintf (string, "3dAveImg%d", fileNumber+1);
    sprintf (title, "Averaged Image, Beam Combiner %s", format->beamCombiner);
    if (plotFile) midiCreatePlotFile3D (string, title, "X", "Y", "Flux", 
        0, aveImage, format->iXWidth, format->iYWidth, "lines", "3");
    free (string);
    free (title);
    
    //    Compute the centroid coordinates of targets in the given image 
    computeTargetCentroidsRefPix (fileNumber, format, aveImage, refPix, error);
    if (*error) midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot determine target centroid");
    
    //    Release memory
    qfits_table_close (pTable);
    free (inData);
    free (aveImage);

    return; 
}
/*****************************************************************************/



/******************************************************************************
*               European Southern Observatory
*          VLTI MIDI Maintenance Templates Software
*
* Module name:  computeTargetCentroidsRefPix
* Input/Output: See function arguments to avoid duplication
* Description:     Using a 2D Gaussian fit find the coordinates of the target in the given averaged 
*                image. The number of targets depend on the type of observation. For 
*                Interferometry (HIGH_SENS or OPEN) there are 2 groups of 3-pin-holes. The 
*                coordinates of the central pinhole for each group are always given. 
*                Additionally the size of a search window is also given. The targets are
*                expected to be around the cetral pinholes. The result of the 2D Gaussian 
*                determines the accuracy of the beam positions. For Photometry (SCI_PHOT) 
*                there are 3 groups of 3-pin-holes. Hence three sets of coordinates are always
*                given.
*
* History:      
* 18-Aug-04     (csabet) Created
******************************************************************************/
void computeTargetCentroidsRefPix (
    int                fileNumber,    //    In:    File number
    ImageFormat     *format,    //    In:    Pointer to the image format
    float             *image,        //  In: Pointer to the image data
    ReferencePixel     *refPix,    //    In:    Pointer to ref pix structure                        
    int                *error)        //    Ou:    Error status
{
  
    //  Local Declarations
    //    ------------------
    const char  routine[] = "computeTargetCentroidsRefPix";
    int            dimension = 2, xPinhole, yPinhole, sizeSearch;
    double        xTarget, yTarget, sizeXTarget, sizeYTarget;
    
    //  Algorithm
    //    ---------
    if (diagnostic > 4)cpl_msg_info(cpl_func,"Invoking      routine   '%s' \n", routine);
    if (diagnostic > 4) fprintf (midiReportPtr, "Invoking      routine   '%s' \n", routine);

    //    Reset status
    *error = 0;

    //    Using a 2D Gaussian fit find the coordinates of the targets for each case
    if ((strcmp (refPix->exposure[fileNumber].beamCombiner, "HIGH_SENS") == 0) || 
        (strcmp (refPix->exposure[fileNumber].beamCombiner, "OPEN") == 0))
    {
        //    Here we expect 2 beams for each case
        if (strcmp (refPix->exposure[fileNumber].shutterId, "ABOPEN") == 0)
        {
            //    Compute beam 1 coordinates
            sizeSearch = SIZE_SEARCH_REF_PIX;
            xPinhole = (int)X1_REF_PIX_ABOPEN;
            yPinhole = (int)Y1_REF_PIX_ABOPEN;
            midiGaussianFit (fileNumber, dimension, image, format->iXWidth, format->iYWidth, xPinhole, yPinhole, 
                sizeSearch, &xTarget, &yTarget, &sizeXTarget, &sizeYTarget, error);
            if (*error) 
            {
                midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot perform a Gaussian fit");
                return;
            }
            //    Save output values
            refPix->exposure[fileNumber].centroid[0].xCoord = xTarget;
            refPix->exposure[fileNumber].centroid[0].yCoord = yTarget;
            refPix->exposure[fileNumber].centroid[0].size = 0.5 * (sizeXTarget + sizeYTarget);

            //    Compute beam 2 coordinates
            xPinhole = (int)X2_REF_PIX_ABOPEN;
            yPinhole = (int)Y2_REF_PIX_ABOPEN;
            midiGaussianFit (fileNumber, dimension, image, format->iXWidth, format->iYWidth, xPinhole, yPinhole, 
                sizeSearch, &xTarget, &yTarget, &sizeXTarget, &sizeYTarget, error);
            if (*error) 
            {
                midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot perform a Gaussian fit");
                return;
            }
            //    Save output values
            refPix->exposure[fileNumber].centroid[1].xCoord = xTarget;
            refPix->exposure[fileNumber].centroid[1].yCoord = yTarget;
            refPix->exposure[fileNumber].centroid[1].size = 0.5 * (sizeXTarget + sizeYTarget);
        }
        else if (strcmp (refPix->exposure[fileNumber].shutterId, "AOPEN") == 0)
        {
            //    Compute beam 1 coordinates
            sizeSearch = SIZE_SEARCH_REF_PIX;
            xPinhole = (int)X1_REF_PIX_HS_AOPEN;
            yPinhole = (int)Y1_REF_PIX_HS_AOPEN;
            midiGaussianFit (fileNumber, dimension, image, format->iXWidth, format->iYWidth, xPinhole, yPinhole, 
                sizeSearch, &xTarget, &yTarget, &sizeXTarget, &sizeYTarget, error);
            if (*error) 
            {
                midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot perform a Gaussian fit");
                return;
            }
            //    Save output values
            refPix->exposure[fileNumber].centroid[0].xCoord = xTarget;
            refPix->exposure[fileNumber].centroid[0].yCoord = yTarget;
            refPix->exposure[fileNumber].centroid[0].size = 0.5 * (sizeXTarget + sizeYTarget);

            //    Compute beam 2 coordinates
            xPinhole = (int)X2_REF_PIX_HS_AOPEN;
            yPinhole = (int)Y2_REF_PIX_HS_AOPEN;
            midiGaussianFit (fileNumber, dimension, image, format->iXWidth, format->iYWidth, xPinhole, yPinhole, 
                sizeSearch, &xTarget, &yTarget, &sizeXTarget, &sizeYTarget, error);
            if (*error) 
            {
                midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot perform a Gaussian fit");
                return;
            }
            //    Save output values
            refPix->exposure[fileNumber].centroid[1].xCoord = xTarget;
            refPix->exposure[fileNumber].centroid[1].yCoord = yTarget;
            refPix->exposure[fileNumber].centroid[1].size = 0.5 * (sizeXTarget + sizeYTarget);
        }
        else if (strcmp (refPix->exposure[fileNumber].shutterId, "BOPEN") == 0)
        {
            //    Compute beam 1 coordinates
            sizeSearch = SIZE_SEARCH_REF_PIX;
            xPinhole = (int)X1_REF_PIX_HS_BOPEN;
            yPinhole = (int)Y1_REF_PIX_HS_BOPEN;
            midiGaussianFit (fileNumber, dimension, image, format->iXWidth, format->iYWidth, xPinhole, yPinhole, 
                sizeSearch, &xTarget, &yTarget, &sizeXTarget, &sizeYTarget, error);
            if (*error) 
            {
                midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot perform a Gaussian fit");
                return;
            }
            //    Save output values
            refPix->exposure[fileNumber].centroid[0].xCoord = xTarget;
            refPix->exposure[fileNumber].centroid[0].yCoord = yTarget;
            refPix->exposure[fileNumber].centroid[0].size = 0.5 * (sizeXTarget + sizeYTarget);

            //    Compute beam 2 coordinates
            xPinhole = (int)X2_REF_PIX_HS_BOPEN;
            yPinhole = (int)Y2_REF_PIX_HS_BOPEN;
            midiGaussianFit (fileNumber, dimension, image, format->iXWidth, format->iYWidth, xPinhole, yPinhole, 
                sizeSearch, &xTarget, &yTarget, &sizeXTarget, &sizeYTarget, error);
            if (*error) 
            {
                midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot perform a Gaussian fit");
                return;
            }
            //    Save output values
            refPix->exposure[fileNumber].centroid[1].xCoord = xTarget;
            refPix->exposure[fileNumber].centroid[1].yCoord = yTarget;
            refPix->exposure[fileNumber].centroid[1].size = 0.5 * (sizeXTarget + sizeYTarget);
        }
        else
        {
            midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Unknown Shutter ID");
            *error = 1;
            return;
        }
    }
    else if (strcmp (refPix->exposure[fileNumber].beamCombiner, "SCI_PHOT") == 0)
    {
        //    Here we expect 3 beams for each case
        if (strcmp (refPix->exposure[fileNumber].shutterId, "AOPEN") == 0)
        {
            sizeSearch = SIZE_SEARCH_REF_PIX;
            xPinhole = (int)X1_REF_PIX_SP_AOPEN;
            yPinhole = (int)Y1_REF_PIX_SP_AOPEN;
            midiGaussianFit (fileNumber, dimension, image, format->iXWidth, format->iYWidth, xPinhole, yPinhole, 
                sizeSearch, &xTarget, &yTarget, &sizeXTarget, &sizeYTarget, error);
            if (*error) 
            {
                midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot perform a Gaussian fit");
                return;
            }
            //    Save output values
            refPix->exposure[fileNumber].centroid[0].xCoord = xTarget;
            refPix->exposure[fileNumber].centroid[0].yCoord = yTarget;
            refPix->exposure[fileNumber].centroid[0].size = 0.5 * (sizeXTarget + sizeYTarget);
            
            xPinhole = (int)X2_REF_PIX_SP_AOPEN;
            yPinhole = (int)Y2_REF_PIX_SP_AOPEN;
            midiGaussianFit (fileNumber, dimension, image, format->iXWidth, format->iYWidth, xPinhole, yPinhole, 
                sizeSearch, &xTarget, &yTarget, &sizeXTarget, &sizeYTarget, error);
            if (*error) 
            {
                midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot perform a Gaussian fit");
                return;
            }
            //    Save output values
            refPix->exposure[fileNumber].centroid[1].xCoord = xTarget;
            refPix->exposure[fileNumber].centroid[1].yCoord = yTarget;
            refPix->exposure[fileNumber].centroid[1].size = 0.5 * (sizeXTarget + sizeYTarget);
            
            xPinhole = (int)X3_REF_PIX_SP_AOPEN;
            yPinhole = (int)Y3_REF_PIX_SP_AOPEN;
            midiGaussianFit (fileNumber, dimension, image, format->iXWidth, format->iYWidth, xPinhole, yPinhole, 
                sizeSearch, &xTarget, &yTarget, &sizeXTarget, &sizeYTarget, error);
            if (*error) 
            {
                midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot perform a Gaussian fit");
                return;
            }
            //    Save output values
            refPix->exposure[fileNumber].centroid[2].xCoord = xTarget;
            refPix->exposure[fileNumber].centroid[2].yCoord = yTarget;
            refPix->exposure[fileNumber].centroid[2].size = 0.5 * (sizeXTarget + sizeYTarget);
        }
        else if (strcmp (refPix->exposure[fileNumber].shutterId, "BOPEN") == 0)
        {
            sizeSearch = SIZE_SEARCH_REF_PIX;
            xPinhole = (int)X1_REF_PIX_SP_BOPEN;
            yPinhole = (int)Y1_REF_PIX_SP_BOPEN;
            midiGaussianFit (fileNumber, dimension, image, format->iXWidth, format->iYWidth, xPinhole, yPinhole, 
                sizeSearch, &xTarget, &yTarget, &sizeXTarget, &sizeYTarget, error);
            if (*error) 
            {
                midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot perform a Gaussian fit");
                return;
            }
            //    Save output values
            refPix->exposure[fileNumber].centroid[0].xCoord = xTarget;
            refPix->exposure[fileNumber].centroid[0].yCoord = yTarget;
            refPix->exposure[fileNumber].centroid[0].size = 0.5 * (sizeXTarget + sizeYTarget);
            
            xPinhole = (int)X2_REF_PIX_SP_BOPEN;
            yPinhole = (int)Y2_REF_PIX_SP_BOPEN;
            midiGaussianFit (fileNumber, dimension, image, format->iXWidth, format->iYWidth, xPinhole, yPinhole, 
                sizeSearch, &xTarget, &yTarget, &sizeXTarget, &sizeYTarget, error);
            if (*error) 
            {
                midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot perform a Gaussian fit");
                return;
            }
            //    Save output values
            refPix->exposure[fileNumber].centroid[1].xCoord = xTarget;
            refPix->exposure[fileNumber].centroid[1].yCoord = yTarget;
            refPix->exposure[fileNumber].centroid[1].size = 0.5 * (sizeXTarget + sizeYTarget);
            
            xPinhole = (int)X3_REF_PIX_SP_BOPEN;
            yPinhole = (int)Y3_REF_PIX_SP_BOPEN;
            midiGaussianFit (fileNumber, dimension, image, format->iXWidth, format->iYWidth, xPinhole, yPinhole, 
                sizeSearch, &xTarget, &yTarget, &sizeXTarget, &sizeYTarget, error);
            if (*error) 
            {
                midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Cannot perform a Gaussian fit");
                return;
            }
            //    Save output values
            refPix->exposure[fileNumber].centroid[2].xCoord = xTarget;
            refPix->exposure[fileNumber].centroid[2].yCoord = yTarget;
            refPix->exposure[fileNumber].centroid[2].size = 0.5 * (sizeXTarget + sizeYTarget);
        }
        else
        {
            midiReportWarning (midiReportPtr, routine, __FILE__, __LINE__, "Unknown Shutter ID");
            *error = 1;
            return;
        }
    }
    
    return; 
}
//*****************************************************************************




/******************************************************************************
*               European Southern Observatory
*          VLTI MIDI Maintenance Templates Software
*
* Module name:  evaluateBeamPosition
* Input/Output: See function arguments to avoid duplication
* Description:  Determine the reference pixel values
*
* History:      
* 23-Jun-04     (csabet) Created
******************************************************************************/
void evaluateBeamPosition (void)
{
  
    /*  Local Declarations
    --------------------*/
    const char  routine[] = "evaluateBeamPosition";
    
    /*  Algorithm
    -----------*/
    if (diagnostic > 4)cpl_msg_info(cpl_func,"Invoking      routine   '%s' \n", routine);
    if (diagnostic > 4) fprintf (midiReportPtr, "Invoking      routine   '%s' \n", routine);

    midiReportTbd (midiReportPtr, routine, __FILE__, __LINE__, "Entire routine to be written");

    return; 
}
/*****************************************************************************/

