/* $Id: midi_profile.c,v 1.32 2012-04-10 13:55:16 agabasch Exp $
 *
 * This file is part of the MIDI Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: agabasch $
 * $Date: 2012-04-10 13:55:16 $
 * $Revision: 1.32 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>

#include "midi_utils.h"
#include "midi_pfits.h"
#include "midi_dfs.h"
#include "string.h"
#include "midiTableToFits.h"
#include <fitsio.h>
#include "midi_cplupgrade.h"

/*-----------------------------------------------------------------------------
                            Private function prototypes
 -----------------------------------------------------------------------------*/

static int midi_profile_create(cpl_plugin *);
static int midi_profile_exec(cpl_plugin *);
static int midi_profile_destroy(cpl_plugin *);
static int midi_profile(cpl_frameset *, const cpl_parameterlist *);
static int midi_normalize_image_y(cpl_image * image);
static int table_to_imglst_sky_target(const char * ,
                           const char * ,
                           cpl_imagelist * ,
                           cpl_imagelist * ,
                           cpl_table *);
static int append_image_to_table(cpl_table * table, const char * columname, 
                                 cpl_image * image, int row);
static int midi_copy_extension(const char * infile , const char * outfile, 
                               const char * extension_name); 
static void midi_profile_qc_barycenter(cpl_propertylist * pro_list, int index,
		cpl_image * image, const int * x_start, const int * x_end);
static void midi_profile_qc_maximum(cpl_propertylist * pro_list, int index,
		cpl_image * image, const int * x_start, const int * x_end);

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static char midi_profile_description[] =
"This recipe derives the 2D profile of the spectral dispersed signal in\n"
"the SCI_PHOT and HIGH_SENS mode. It combines a set of AOPEN and BOPEN\n"
"photometry files with the following tags:\n"
"\n"
"PHOTOM_SP_CALIB/PHOTOM_SP_SCIENCE \n"
"or\n"
"PHOTOM_HS_CALIB/PHOTOM_HS_SCIENCE.  \n"
"\n"
"The sky subtracted data are co-added and then a two pass threshold is\n"
"applied in order to get a clean profile of the signal: In a first pass\n"
"the standard deviation of the image is derived and all the pixels\n"
"exceeding the latter are marked as they are supposed to belong to the\n"
"source and not to the background. In a second pass the standard\n"
"deviation is recalculated excluding these source-pixels. After that,\n"
"all pixels of the image not exceeding (<threshold> times\n"
"standard-deviation) are set to zero. Please note, that the <threshold>\n"
"variable can be set as a recipe option in sigma units. Finally the\n"
"integral of the profile perpendicular to the dispersion direction is\n"
"normalized to unity.\n"

"Input files:\n\n"
"  DO category:                Type:       Explanation:         Required:\n"
"  PHOTOM_SP_CALIB             Raw         Raw data frame            \n"
"  or                                                                \n"
"  PHOTOM_SP_SCIENCE           Raw         Raw data frame            \n"
"  or                                                               Y\n"
"  PHOTOM_HS_CALIB             Raw         Raw data frame            \n"
"  or                                                                \n"
"  PHOTOM_HS_SCIENCE           Raw         Raw data frame          \n\n"
"Output files:\n\n"
"  DO category:                Data type:  Explanation:\n"
"  KAPPA_HS_MASK_PRISM         FITS table  Spectral profile: Main product\n"
"  or \n"
"  KAPPA_HS_MASK_GRISM         FITS table  Spectral profile: Main product \n\n"
"  KAPPA_HS_MASK_PRISM_PROFILE FITS image  single profile: For diagnostics\n"
"  or \n"
"  KAPPA_HS_MASK_GRISM_PROFILE FITS image  single profile: For diagnostics\n\n"
"  KAPPA_HS_MASK_PRISM_SIGNAL  FITS image  single signal:  For diagnostics\n"
"  or \n"
"  KAPPA_HS_MASK_GRISM_SIGNAL  FITS image  single signal:  For diagnostics\n";

/*-----------------------------------------------------------------------------
                                Function code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
 * @defgroup midi_profile    Midi Profile
 *
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module. 
  @param    list    the plugin list
  @return   0 if everything is ok, 1 otherwise
  @note     Only this function is exported

  Create the recipe instance and make it available to the application using the 
  interface. 
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist * list)
{
    cpl_recipe  *   recipe = cpl_calloc(1, sizeof *recipe );
    cpl_plugin  *   plugin = &recipe->interface;

    if (cpl_plugin_init(plugin,
                    CPL_PLUGIN_API,
                    MIDI_BINARY_VERSION,
                    CPL_PLUGIN_TYPE_RECIPE,
                    "midi_profile",
                    "Derives the Profile of the spectrum",
                    midi_profile_description,
                    "Armin Gabasch",
                    PACKAGE_BUGREPORT,
                    midi_get_license(),
                    midi_profile_create,
                    midi_profile_exec,
                    midi_profile_destroy)) {    
        cpl_msg_error(cpl_func, "Plugin initialization failed");
        (void)cpl_error_set_where(cpl_func);                          
        return 1;                                               
    }                                                    

    if (cpl_pluginlist_append(list, plugin)) {                 
        cpl_msg_error(cpl_func, "Error adding plugin to list");
        (void)cpl_error_set_where(cpl_func);                         
        return 1;                                              
    }                                                          
    
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options    
  @param    plugin  the plugin
  @return   0 if everything is ok

  Defining the command-line/configuration parameters for the recipe.
 */
/*----------------------------------------------------------------------------*/
static int midi_profile_create(cpl_plugin * plugin)
{
    cpl_recipe    * recipe;                                               
    cpl_parameter * p;
                                                                       
    /* Do not create the recipe if an error code is already set */     
    if (cpl_error_get_code() != CPL_ERROR_NONE) {                      
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());      
        return (int)cpl_error_get_code();                              
    }                                                                  
                                                                       
    if (plugin == NULL) {                                              
        cpl_msg_error(cpl_func, "Null plugin");                        
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                 
    }                                                                  
                                                                       
    /* Verify plugin type */                                           
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {       
        cpl_msg_error(cpl_func, "Plugin is not a recipe");             
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);              
    }                                                                  
                                                                       
    /* Get the recipe */                                               
    recipe = (cpl_recipe *)plugin;                                     
                                                                       
    /* Create the parameters list in the cpl_recipe object */          
    recipe->parameters = cpl_parameterlist_new();                      
    if (recipe->parameters == NULL) {                                  
        cpl_msg_error(cpl_func, "Parameter list allocation failed");   
        cpl_ensure_code(0, (int)CPL_ERROR_ILLEGAL_OUTPUT);             
    }                                                                  

    /* Fill the parameters list */

    p = cpl_parameter_new_value("midi.midi_profile.threshold",
            CPL_TYPE_DOUBLE, "The threshold to discriminate between source and background pixels in sigma units", 
                                "midi.midi_profile",1.7);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "threshold");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);


    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int midi_profile_exec(cpl_plugin * plugin)
{

    cpl_recipe * recipe;                                                   
    int recipe_status;                                                     
    cpl_errorstate initial_errorstate = cpl_errorstate_get();              
                                                                           
    /* Return immediately if an error code is already set */               
    if (cpl_error_get_code() != CPL_ERROR_NONE) {                          
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",    
                      cpl_func, __LINE__, cpl_error_get_where());          
        return (int)cpl_error_get_code();                                  
    }                                                                      
                                                                           
    if (plugin == NULL) {                                                  
        cpl_msg_error(cpl_func, "Null plugin");                            
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                     
    }                                                                      
                                                                           
    /* Verify plugin type */                                               
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {           
        cpl_msg_error(cpl_func, "Plugin is not a recipe");                 
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);                  
    }                                                                      
                                                                           
    /* Get the recipe */                                                   
    recipe = (cpl_recipe *)plugin;                                         
                                                                           
    /* Verify parameter and frame lists */                                 
    if (recipe->parameters == NULL) {                                      
        cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                     
    }                                                                      
    if (recipe->frames == NULL) {                                          
        cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");     
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                     
    }                                                                      
                                                                           
    /* Invoke the recipe */                                                
    recipe_status = midi_profile(recipe->frames, recipe->parameters);
                                                                           
    /* Ensure DFS-compliance of the products */                            
    if (cpl_dfs_update_product_header(recipe->frames)) {                   
        if (!recipe_status) recipe_status = (int)cpl_error_get_code();                         
    }                                                                      
                                                                           
    if (!cpl_errorstate_is_equal(initial_errorstate)) {                    
        /* Dump the error history since recipe execution start.            
           At this point the recipe cannot recover from the error */       
        cpl_errorstate_dump(initial_errorstate, CPL_FALSE, NULL);          
    }                                                                      
                                                                           
    return recipe_status;                                                  
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int midi_profile_destroy(cpl_plugin * plugin)
{
    cpl_recipe * recipe;                                          
                                                                  
    if (plugin == NULL) {                                         
        cpl_msg_error(cpl_func, "Null plugin");                   
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);            
    }                                                             
                                                                  
    /* Verify plugin type */                                      
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {  
        cpl_msg_error(cpl_func, "Plugin is not a recipe");        
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);         
    }                                                             
                                                                  
    /* Get the recipe */                                          
    recipe = (cpl_recipe *)plugin;                                
                                                                  
    cpl_parameterlist_delete(recipe->parameters);             
                                                                  
    return 0;                                                    
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    frameset   the frames list
  @param    parlist    the parameters list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int midi_profile(cpl_frameset            * frameset,
                            const cpl_parameterlist * parlist)
{
   cpl_propertylist * plist=NULL;
   cpl_propertylist * pro_list=NULL;
   cpl_frameset_iterator * it = NULL;
   cpl_propertylist * qclist_all_extension;
   cpl_propertylist **pHeader_ext= NULL;
   
   const cpl_parameter   * param;

   cpl_frame        * cur_frame=NULL;

   cpl_table        * table=NULL;
   cpl_table        * profile_table=NULL;
   cpl_table       ** pTable_ext= NULL;

   cpl_imagelist    * imglst_AOPEN_DATA_S[4];
   cpl_imagelist    * imglst_AOPEN_DATA_T[4];
   cpl_imagelist    * imglst_BOPEN_DATA_S[4];
   cpl_imagelist    * imglst_BOPEN_DATA_T[4];


   cpl_image        * image_AOPEN_DATA_T[4];       
   cpl_image        * image_BOPEN_DATA_T[4];       
   cpl_image        * image_AOPEN_DATA_T_float[4];       
   cpl_image        * image_BOPEN_DATA_T_float[4];       

   cpl_mask         * image_AOPEN_DATA_T_mask[4];
   cpl_mask         * image_BOPEN_DATA_T_mask[4];

   cpl_image        * dummy_image;

   cpl_errorstate     prestate = cpl_errorstate_get();

   char             * tag=NULL;
   char             * dataname=NULL;
   char             * dataname1=NULL;
   char               gris_name[100]= ""; 
   char             * first_valid_frame=NULL;


   const char       * shutter_id =NULL; 

   int                i=0;
   int                ext_imaging_data=0;       
   int                ext_imaging_detector=0;      
   int                isPHOTAdata=0;
   int                isPHOTBdata=0;
   int                isSP=0;   
   int                isHS=0;   
   int                dimenDATA=0;
   int                extnum=1;
   int                isFirst=0;

   double             std_AOPEN [4];       
   double             std_BOPEN[4];       
   double             threshold=0.;

   int                maxloop=0;
   int                image_size_y=0;

   const int x_pos_grism_start[]={38, 52, 116, 176, 217};
   const int x_pos_grism_end[]  ={51, 69, 137, 201, 246};

   const int x_pos_prism_start[]={120, 113, 88, 60, 36};
   const int x_pos_prism_end[]  ={127, 122, 99, 73, 51};

   param = cpl_parameterlist_find_const(parlist,
                                        "midi.midi_profile.threshold");
   threshold = cpl_parameter_get_double(param);
  
   /* Check if SOF is empty */
   cur_frame = cpl_frameset_get_position(frameset, 0);
   if (cur_frame == NULL) {
	   return (int)cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
			   "SOF does not have any file");
   }


   pTable_ext  = cpl_malloc((extnum) * sizeof(cpl_table *));
   pHeader_ext = cpl_malloc((extnum) * sizeof(cpl_propertylist *));


   /* Identify the RAW and CALIB frames in the input frameset */
   cpl_ensure_code(midi_dfs_set_groups(frameset) == CPL_ERROR_NONE,
		   cpl_error_get_code());

   /* Check if SCI_PHOT or HIGH_SENS */
   it = cpl_frameset_iterator_new(frameset);
   cur_frame = cpl_frameset_iterator_get(it);

   while(cur_frame)
   {
      /*      Check the right tags */
      tag = (char*)cpl_frame_get_tag(cur_frame);
      if (strcmp(tag,MIDI_PHOTOM_SP_CALIB) ==0 || strcmp(tag,MIDI_PHOTOM_SP_SCIENCE)==0) {
         isSP=1;
         dimenDATA=4;   
      }
      if (strcmp(tag,MIDI_PHOTOM_HS_CALIB) ==0 || strcmp(tag,MIDI_PHOTOM_HS_SCIENCE)==0) {
         isHS=1;
         dimenDATA=2;   
      }
      /* cpl_msg_info(cpl_func,"tag: %s SP: %d HS:%d ",tag,isSP,isHS); */

      /* Get next frame from SOF */
      cpl_frameset_iterator_advance(it, 1);
      cur_frame = cpl_frameset_iterator_get(it);
   }
   cpl_frameset_iterator_delete(it);
   if (isHS>0 && isSP>0) {
      return (int)cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                                        "SOF has mixed data: SCI_PHOT and HIGH_SENS. This is not allowed!");
   }


   for (i=0; i<dimenDATA;i++){

      imglst_AOPEN_DATA_S[i]=cpl_imagelist_new();
      imglst_AOPEN_DATA_T[i]=cpl_imagelist_new();
      
      imglst_BOPEN_DATA_S[i]=cpl_imagelist_new();
      imglst_BOPEN_DATA_T[i]=cpl_imagelist_new();

   }
   
  

/*       Walk through the whole Set of Frames SOF  */


/*    Start  1   */
   
   cur_frame = cpl_frameset_get_position(frameset, 0);
   if (cur_frame == NULL) {
      return (int)cpl_error_set_message(cpl_func, CPL_ERROR_DATA_NOT_FOUND,
                                        "SOF does not have any file");
   }
   /* Append the images from the tables to the various imagelists */
   it = cpl_frameset_iterator_new(frameset);
   cur_frame = cpl_frameset_iterator_get(it);
   while(cur_frame)
   {
      /*      Check the right tags */
      tag = (char*)cpl_frame_get_tag(cur_frame);
      if (strcmp(tag, MIDI_PHOTOM_SP_SCIENCE) && strcmp(tag, MIDI_PHOTOM_SP_CALIB) && 
          strcmp(tag, MIDI_PHOTOM_HS_SCIENCE) && strcmp(tag, MIDI_PHOTOM_HS_CALIB)) {
          cpl_frameset_iterator_advance(it, 1);
          cur_frame = cpl_frameset_iterator_get(it);

         continue;
      }
      
      cpl_msg_info(cpl_func, "Processing file  %s",cpl_frame_get_filename(cur_frame));
      ext_imaging_data=cpl_fits_find_extension(cpl_frame_get_filename(cur_frame),"IMAGING_DATA");
      ext_imaging_detector=cpl_fits_find_extension(cpl_frame_get_filename(cur_frame),"IMAGING_DETECTOR");


      /* Read the IMAGING_DETECTOR extension from the first valid image */
      
      if(isFirst<1){
         isFirst=1;

      /* Save the name of the first valid frame */
         first_valid_frame=cpl_sprintf("%s", cpl_frame_get_filename(cur_frame));         

         for (i=0; i<extnum; i++)
         {
            pHeader_ext[i] = cpl_propertylist_load(cpl_frame_get_filename(cur_frame), ext_imaging_detector );
            pTable_ext[i]  = cpl_table_load(cpl_frame_get_filename(cur_frame), ext_imaging_detector, 1 );
            
            if (pHeader_ext[i] ==NULL || pTable_ext[i]==NULL)
            {
               cpl_msg_error(cpl_func,"Error loading the extended header or table of %s",cpl_frame_get_filename(cur_frame)); 
               
               for (i=0; i<extnum; i++)
               {
                  cpl_propertylist_delete(pHeader_ext[i]);
                  cpl_table_delete(pTable_ext[i]); 
               }
               cpl_free (pHeader_ext);
               cpl_free (pTable_ext); 
               
               
               
               return -1;
            }
         }
      }

      
/*       Check if the file is AOPEN or BOPEN */
      
      plist = cpl_propertylist_load(cpl_frame_get_filename(cur_frame), 0);
      if (cpl_propertylist_has(plist, "ESO INS SHUT ID") == 1)
      {
         shutter_id=(cpl_propertylist_get_string(plist, "ESO INS SHUT ID"));
      }
      
/*       Checking if PRISM or GRISM */
      if (cpl_propertylist_has(plist, "ESO INS GRIS NAME") == 1)
      {
         strcpy(gris_name,cpl_propertylist_get_string(plist, "ESO INS GRIS NAME"));
      }
      
      if (!cpl_errorstate_is_equal(prestate)) {
         return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(), "Could not ...");
      }
      
      if (strcmp(shutter_id,"AOPEN")==0)
      {
/*          PhotA Data found */
         isPHOTAdata=1;
         
         /* Load extension  Imaging Data  */ 
         table = cpl_table_load(cpl_frame_get_filename(cur_frame), ext_imaging_data, 1);
         if (table == NULL) {
            return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
                                              "Could not load the table");
         }
         cpl_ensure_code(cpl_errorstate_is_equal(prestate), cpl_error_get_code());      


         for (i=0; i<dimenDATA;i++){
 
            dataname=cpl_sprintf("DATA%d",i+1);           

            /*         cpl_msg_info(cpl_func, "Scanning for dataname ..."); */
            if (cpl_table_has_column(table,dataname)){
               table_to_imglst_sky_target(dataname,"TARTYP2",imglst_AOPEN_DATA_S[i],imglst_AOPEN_DATA_T[i],table);
            }
            
            cpl_msg_info(cpl_func, "Number of so far processed AOPEN %s patches: % " CPL_SIZE_FORMAT "",dataname,cpl_imagelist_get_size(imglst_AOPEN_DATA_T[i]));
            cpl_free(dataname);
         }
         
         cpl_ensure_code(cpl_errorstate_is_equal(prestate), cpl_error_get_code());
         
         cpl_table_delete(table);
      }
      
      if (strcmp(shutter_id,"BOPEN")==0)
      {
/*          PhotB Data found */
         isPHOTBdata=1;
         
         /* Load extension  Imaging Data  */ 
         table = cpl_table_load(cpl_frame_get_filename(cur_frame), ext_imaging_data, 1);
         if (table == NULL) {
            return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
                                              "Could not load the table");
         }
         cpl_ensure_code(cpl_errorstate_is_equal(prestate), cpl_error_get_code());      
         

         for (i=0; i<dimenDATA;i++){
 
            dataname=cpl_sprintf("DATA%d",i+1);           
            /*         cpl_msg_info(cpl_func, "Scanning for dataname ..."); */
            if (cpl_table_has_column(table,dataname)){
               table_to_imglst_sky_target(dataname,"TARTYP2",imglst_BOPEN_DATA_S[i],imglst_BOPEN_DATA_T[i],table);
            }
            
            cpl_msg_info(cpl_func, "Number of so far processed BOPEN %s patches: % " CPL_SIZE_FORMAT "",dataname,cpl_imagelist_get_size(imglst_BOPEN_DATA_T[i]));
            cpl_free(dataname);
         }
         
         cpl_ensure_code(cpl_errorstate_is_equal(prestate), cpl_error_get_code());
         
         cpl_table_delete(table);
      }
      
      
      cpl_propertylist_delete(plist);
      
      /* Get next frame from SOF */
      cpl_frameset_iterator_advance(it, 1);
      cur_frame = cpl_frameset_iterator_get(it);
   } /* while more frames */
   cpl_frameset_iterator_delete(it);
   
   if (isPHOTAdata==0 || isPHOTBdata==0)
   {
      cpl_msg_error(cpl_func, "No suitable SetOfFrame fround");              
      (void)cpl_error_set_where(cpl_func);
      return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(), "No suitable SetOfFrame fround");
      
   }


/* /\* Save the single sky and target cubes *\/ */


/*    for (i=0; i<dimenDATA;i++){ */
/*       if (strcmp(gris_name,"GRISM")==0 && isSP>0) dataname=cpl_sprintf("KAPPA_SP_MASK_GRISM"); */
/*       if (strcmp(gris_name,"PRISM")==0 && isSP>0) dataname=cpl_sprintf("KAPPA_SP_MASK_PRISM"); */
/*       if (strcmp(gris_name,"GRISM")==0 && isHS>0) dataname=cpl_sprintf("KAPPA_HS_MASK_GRISM"); */
/*       if (strcmp(gris_name,"PRISM")==0 && isHS>0) dataname=cpl_sprintf("KAPPA_HS_MASK_PRISM"); */

           
/*       dataname1=cpl_sprintf("imglst_AOPEN_DATA%d_T.fits",i+1);            */
/*       if (cpl_dfs_save_imagelist(frameset, parlist, frameset, imglst_AOPEN_DATA_T[i], */
/*                              CPL_BPP_IEEE_FLOAT, "midi_profile", */
/*                              dataname, NULL, NULL, */
/*                              PACKAGE "/" PACKAGE_VERSION, */
/*                              dataname1)) { */
/*          /\* Propagate the error *\/ */
/*          (void)cpl_error_set_where(cpl_func); */
/*       } */
/*       cpl_free(dataname1);       */
/*       dataname1=cpl_sprintf("imglst_AOPEN_DATA%d_S.fits",i+1);            */
/*       if (cpl_dfs_save_imagelist(frameset, parlist, frameset, imglst_AOPEN_DATA_S[i], */
/*                              CPL_BPP_IEEE_FLOAT, "midi_profile", */
/*                              dataname, NULL, NULL, */
/*                              PACKAGE "/" PACKAGE_VERSION, */
/*                              dataname1)) { */
/*          /\* Propagate the error *\/ */
/*          (void)cpl_error_set_where(cpl_func); */
/*       } */
/*       cpl_free(dataname1);       */


/*       dataname1=cpl_sprintf("imglst_BOPEN_DATA%d_T.fits",i+1);            */
/*       if (cpl_dfs_save_imagelist(frameset, parlist, frameset, imglst_BOPEN_DATA_T[i], */
/*                              CPL_BPP_IEEE_FLOAT, "midi_profile", */
/*                              dataname, NULL, NULL, */
/*                        PACKAGE "/" PACKAGE_VERSION, */
/*                              dataname1)) { */
/*          /\* Propagate the error *\/ */
/*          (void)cpl_error_set_where(cpl_func); */
/*       } */
/*       cpl_free(dataname1); */
/*       dataname1=cpl_sprintf("imglst_BOPEN_DATA%d_S.fits",i+1);            */
/*       if (cpl_dfs_save_imagelist(frameset, parlist, frameset, imglst_BOPEN_DATA_S[i], */
/*                                  CPL_BPP_IEEE_FLOAT, "midi_profile", */
/*                                  dataname, NULL, NULL, */
/*                                  PACKAGE "/" PACKAGE_VERSION, */
/*                                  dataname1)) { */
/*          /\* Propagate the error *\/ */
/*          (void)cpl_error_set_where(cpl_func); */
/*       } */
/*       cpl_free(dataname); */
      
/*       cpl_free(dataname1);       */
      

/*    } */






/*    Subtract the sky frame from the target frame*/
   
   cpl_msg_info(cpl_func, "Doing sky subtraction ...");


   for (i=0; i<dimenDATA;i++){
      
      if(cpl_imagelist_get_size(imglst_AOPEN_DATA_T[i])==cpl_imagelist_get_size(imglst_AOPEN_DATA_S[i])
         && cpl_imagelist_get_size(imglst_BOPEN_DATA_T[i])==cpl_imagelist_get_size(imglst_BOPEN_DATA_S[i])
         )
      {
         cpl_imagelist_subtract(imglst_AOPEN_DATA_T[i],imglst_AOPEN_DATA_S[i]); 
         cpl_imagelist_subtract(imglst_BOPEN_DATA_T[i],imglst_BOPEN_DATA_S[i]); 
         
      }
      else 
      {
         cpl_msg_info(cpl_func, "The number of Skyframes and Targetframes differ");
         return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
                                           "The number of Skyframes and Targetframes differ"); 
      }
      
   }
   

   cpl_msg_info(cpl_func, "Collapsing the images in the time domain ...");

/*    Collapse the sky-subtracted targets */
   for (i=0; i<dimenDATA;i++){

      dummy_image=cpl_imagelist_collapse_create(imglst_AOPEN_DATA_T[i]);
      image_AOPEN_DATA_T[i]=cpl_image_cast(dummy_image,CPL_TYPE_DOUBLE);
      cpl_image_delete(dummy_image);

      dummy_image=cpl_imagelist_collapse_create(imglst_BOPEN_DATA_T[i]);
      image_BOPEN_DATA_T[i]=cpl_image_cast(dummy_image,CPL_TYPE_DOUBLE);
      cpl_image_delete(dummy_image);      

   }

   
/* For HIGH_SENS: stack AOPEN and BOPEN DATA1/DATA2 frames */

   if (isHS>0) {
      cpl_image_add(image_AOPEN_DATA_T[0], image_BOPEN_DATA_T[0]);  
      cpl_image_add(image_AOPEN_DATA_T[1], image_BOPEN_DATA_T[1]);  
   }
 
/* For SCI_PHOT: stack AOPEN and BOPEN DATA2/DATA3 frames */

   if (isSP>0) {
      cpl_image_add(image_AOPEN_DATA_T[1], image_BOPEN_DATA_T[1]);  
      cpl_image_add(image_AOPEN_DATA_T[2], image_BOPEN_DATA_T[2]);  
   }


   pro_list = cpl_propertylist_new();

/* Save the images for diagnostics */

   if(1){
      for (i=0; i<dimenDATA;i++){
         if (strcmp(gris_name,"GRISM")==0 && isSP>0) cpl_propertylist_update_string(pro_list, CPL_DFS_PRO_CATG, "KAPPA_SP_MASK_GRISM_SIGNAL");
         if (strcmp(gris_name,"PRISM")==0 && isSP>0) cpl_propertylist_update_string(pro_list, CPL_DFS_PRO_CATG, "KAPPA_SP_MASK_PRISM_SIGNAL");
         if (strcmp(gris_name,"GRISM")==0 && isHS>0) cpl_propertylist_update_string(pro_list, CPL_DFS_PRO_CATG, "KAPPA_HS_MASK_GRISM_SIGNAL");
         if (strcmp(gris_name,"PRISM")==0 && isHS>0) cpl_propertylist_update_string(pro_list, CPL_DFS_PRO_CATG, "KAPPA_HS_MASK_PRISM_SIGNAL");
         
         
         dataname1=cpl_sprintf("image_AOPEN_DATA%d_signal.fits",i+1);           
         if (cpl_dfs_save_image(frameset, NULL, parlist, frameset, NULL,  image_AOPEN_DATA_T[i],
                                CPL_BPP_IEEE_FLOAT, "midi_profile",
                                pro_list, NULL,
                                PACKAGE "/" PACKAGE_VERSION,
                                dataname1)) {
            /* Propagate the error */
            (void)cpl_error_set_where(cpl_func);
         }
         cpl_free(dataname1);      
         
         
         dataname1=cpl_sprintf("image_BOPEN_DATA%d_signal.fits",i+1);           
         if (cpl_dfs_save_image(frameset, NULL, parlist, frameset, NULL,  image_BOPEN_DATA_T[i],
                                CPL_BPP_IEEE_FLOAT, "midi_profile",
                                pro_list, NULL,
                                PACKAGE "/" PACKAGE_VERSION,
                                dataname1)) {
            /* Propagate the error */
            (void)cpl_error_set_where(cpl_func);
         }
         cpl_free(dataname1);      
         
      }
   }
  

/* Calculate the standard deviation */

   for (i=0; i<dimenDATA;i++){
      std_AOPEN[i]=cpl_image_get_stdev(image_AOPEN_DATA_T[i]);
      std_BOPEN[i]=cpl_image_get_stdev(image_BOPEN_DATA_T[i]);
      cpl_msg_info(cpl_func,"First pass STDEV of AOPEN DATA%d: %f", i+1, std_AOPEN[i]);
      cpl_msg_info(cpl_func,"First pass STDEV of BOPEN DATA%d: %f", i+1, std_BOPEN[i]);

   }
   


/* Create a bad pixel mask and connect it with the image*/
   for (i=0; i<dimenDATA;i++){
      image_AOPEN_DATA_T_mask[i] = cpl_mask_threshold_image_create(image_AOPEN_DATA_T[i], std_AOPEN[i], FLT_MAX); 	
      image_BOPEN_DATA_T_mask[i] = cpl_mask_threshold_image_create(image_BOPEN_DATA_T[i], std_BOPEN[i], FLT_MAX); 	
      cpl_image_reject_from_mask(image_AOPEN_DATA_T[i], image_AOPEN_DATA_T_mask[i]) ;    
      cpl_image_reject_from_mask(image_BOPEN_DATA_T[i], image_BOPEN_DATA_T_mask[i]) ;    
      
   }

/* Re Calculate the standard deviation using the mask file*/

   cpl_msg_info(cpl_func," ");
   for (i=0; i<dimenDATA;i++){
      std_AOPEN[i]=cpl_image_get_stdev(image_AOPEN_DATA_T[i]);
      std_BOPEN[i]=cpl_image_get_stdev(image_BOPEN_DATA_T[i]);
      cpl_msg_info(cpl_func,"Second pass STDEV of AOPEN DATA%d: %f", i+1, std_AOPEN[i]);
      cpl_msg_info(cpl_func,"Second pass STDEV of BOPEN DATA%d: %f", i+1, std_BOPEN[i]);
   }

/* un-connect the mask and free the memory*/
   for (i=0; i<dimenDATA;i++){
      cpl_image_accept_all(image_AOPEN_DATA_T[i]) ; 
      cpl_image_accept_all(image_BOPEN_DATA_T[i]) ; 

      cpl_mask_delete(image_AOPEN_DATA_T_mask[i]);
      cpl_mask_delete(image_BOPEN_DATA_T_mask[i]);
   }

   
   


/* Threshold the image */

   for (i=0; i<dimenDATA;i++){
      cpl_image_threshold(image_AOPEN_DATA_T[i],threshold*std_AOPEN[i],FLT_MAX,0.,FLT_MAX);
      cpl_image_threshold(image_BOPEN_DATA_T[i],threshold*std_BOPEN[i],FLT_MAX,0.,FLT_MAX);

   }
   

/* Normalize the profile */

   for (i=0; i<dimenDATA;i++){
      midi_normalize_image_y(image_AOPEN_DATA_T[i]);
      midi_normalize_image_y(image_BOPEN_DATA_T[i]);
      
   }

/* Cast images do float as the midi core library expects a float profile*/
/* ONLY THE TABLE IS SAVED AS FLOAT !!!! */

   for (i=0; i<dimenDATA;i++){
      image_AOPEN_DATA_T_float[i]=cpl_image_cast(image_AOPEN_DATA_T[i],CPL_TYPE_FLOAT);
      image_BOPEN_DATA_T_float[i]=cpl_image_cast(image_BOPEN_DATA_T[i],CPL_TYPE_FLOAT);
      
   }

/* Derive QC parameters */

   cpl_msg_info(cpl_func, "Deriving QC parameters ...");
   image_size_y=cpl_image_get_size_y(image_AOPEN_DATA_T[0]);

   if (isHS>0) { 
      maxloop=2;
   }
   else {
      maxloop=4;
   }

   /* Compute the barycenter in y and the maximum... */
   for (i = 0; i < maxloop; i++){

	   if (strcmp(gris_name,"GRISM")==0)
	   {
		   if(isSP>0 && i==0){
			   /* Take data from BOBEN image as AOPEN has no flux */
			   midi_profile_qc_barycenter(pro_list, i, image_BOPEN_DATA_T[i],
					   x_pos_grism_start,x_pos_grism_end);
			   midi_profile_qc_maximum(pro_list, i, image_BOPEN_DATA_T[i],
					   x_pos_grism_start,x_pos_grism_end);
		   }
		   else{
			   midi_profile_qc_barycenter(pro_list, i, image_AOPEN_DATA_T[i],
					   x_pos_grism_start,x_pos_grism_end);
			   midi_profile_qc_maximum(pro_list, i, image_AOPEN_DATA_T[i],
					   x_pos_grism_start,x_pos_grism_end);
		   }
	   }

	   if (strcmp(gris_name,"PRISM")==0)
	   {
		   if(isSP>0 && i==0){
			   /* Take data from BOBEN image as AOPEN has no flux */
			   midi_profile_qc_barycenter(pro_list, i, image_BOPEN_DATA_T[i],
					   x_pos_prism_start,x_pos_prism_end);
			   midi_profile_qc_maximum(pro_list, i, image_BOPEN_DATA_T[i],
					   x_pos_prism_start,x_pos_prism_end);
		   }
		   else{
			   midi_profile_qc_barycenter(pro_list, i, image_AOPEN_DATA_T[i],
					   x_pos_prism_start,x_pos_prism_end);
			   midi_profile_qc_maximum(pro_list, i, image_AOPEN_DATA_T[i],
					   x_pos_prism_start,x_pos_prism_end);
		   }
	   }
   }


      
/* Save the result */


/* Save it to a fits table */
   
   profile_table=cpl_table_new(1);
   
/* Propertylist to write to the extension */
   qclist_all_extension = cpl_propertylist_new();
   cpl_propertylist_append_string (qclist_all_extension, "EXTNAME", "IMAGING_DATA");
   
   if (isHS>0) { 
      append_image_to_table(profile_table,"DATA1",image_AOPEN_DATA_T_float[0],0);
      append_image_to_table(profile_table,"DATA2",image_AOPEN_DATA_T_float[1],0);
   }
   
   if (isSP>0) { 
      append_image_to_table(profile_table,"DATA1",image_BOPEN_DATA_T_float[0],0);
      append_image_to_table(profile_table,"DATA2",image_AOPEN_DATA_T_float[1],0);
      append_image_to_table(profile_table,"DATA3",image_AOPEN_DATA_T_float[2],0);
      append_image_to_table(profile_table,"DATA4",image_AOPEN_DATA_T_float[3],0);
   }
   
   
   if (strcmp(gris_name,"GRISM")==0 && isSP>0) cpl_propertylist_update_string(pro_list, CPL_DFS_PRO_CATG, "KAPPA_SP_MASK_GRISM");
   if (strcmp(gris_name,"PRISM")==0 && isSP>0) cpl_propertylist_update_string(pro_list, CPL_DFS_PRO_CATG, "KAPPA_SP_MASK_PRISM");
   if (strcmp(gris_name,"GRISM")==0 && isHS>0) cpl_propertylist_update_string(pro_list, CPL_DFS_PRO_CATG, "KAPPA_HS_MASK_GRISM");
   if (strcmp(gris_name,"PRISM")==0 && isHS>0) cpl_propertylist_update_string(pro_list, CPL_DFS_PRO_CATG, "KAPPA_HS_MASK_PRISM");
   
   if (strcmp(gris_name,"GRISM")==0)
   {
      cpl_dfs_save_table(frameset, NULL, parlist, frameset, NULL, profile_table, 
                         qclist_all_extension, "midi_profile",
                         pro_list, NULL, 
                         PACKAGE "/" PACKAGE_VERSION,
                         "midi_profile_grism.fits");
      cpl_table_delete(profile_table);   
      
      /* Append the IMAGING_DETECTOR table from the first valid image */
      /*  midi_copy_extension() is a workaround to cpl_table_save() as
       *  the latter is modifiing the content during saving */

      midi_copy_extension(first_valid_frame, "midi_profile_grism.fits", "IMAGING_DETECTOR");

/*       if (CPL_ERROR_NONE != cpl_table_save(pTable_ext[0], NULL, pHeader_ext[0], "midi_profile_grism.fits", CPL_IO_EXTEND )) */
/*       { */
/*          cpl_msg_error(cpl_func,"Error when saving Extension: %s %s",cpl_error_get_message(),cpl_error_get_where()); */
/*       } */
   }
   
   if (strcmp(gris_name,"PRISM")==0)
   {
      cpl_dfs_save_table(frameset, NULL, parlist, frameset, NULL, profile_table, 
                         qclist_all_extension, "midi_profile",
                         pro_list, NULL, 
                         PACKAGE "/" PACKAGE_VERSION,
                         "midi_profile_prism.fits");
      cpl_table_delete(profile_table);    
      
      /* Append the IMAGING_DETECTOR table from the first valid image */
      /*  midi_copy_extension() is a workaround to cpl_table_save() as
       *  the latter is modifiing the content during saving */
      
      midi_copy_extension(first_valid_frame, "midi_profile_prism.fits", "IMAGING_DETECTOR");
      
/*       if (CPL_ERROR_NONE != cpl_table_save(pTable_ext[0], NULL, pHeader_ext[0], "midi_profile_prism.fits", CPL_IO_EXTEND )) */
/*       { */
/*          cpl_msg_error(cpl_func,"Error when saving Extension: %s %s",cpl_error_get_message(),cpl_error_get_where()); */
/*       } */
   }
   

   cpl_propertylist_delete(qclist_all_extension);
/*    cpl_free(dataname);  */
   
   for (i=0; i<extnum; i++)
   {
      cpl_propertylist_delete(pHeader_ext[i]);
      cpl_table_delete(pTable_ext[i]); 
   }
   cpl_free (pHeader_ext);
   cpl_free (pTable_ext); 
   cpl_free(first_valid_frame);   
   
   
/* Save the images for diagnostics */
   if(1){
      for (i=0; i<dimenDATA;i++){
         if (strcmp(gris_name,"GRISM")==0 && isSP>0) cpl_propertylist_update_string(pro_list, CPL_DFS_PRO_CATG, "KAPPA_SP_MASK_GRISM_PROFILE");
         if (strcmp(gris_name,"PRISM")==0 && isSP>0) cpl_propertylist_update_string(pro_list, CPL_DFS_PRO_CATG, "KAPPA_SP_MASK_PRISM_PROFILE");
         if (strcmp(gris_name,"GRISM")==0 && isHS>0) cpl_propertylist_update_string(pro_list, CPL_DFS_PRO_CATG, "KAPPA_HS_MASK_GRISM_PROFILE");
         if (strcmp(gris_name,"PRISM")==0 && isHS>0) cpl_propertylist_update_string(pro_list, CPL_DFS_PRO_CATG, "KAPPA_HS_MASK_PRISM_PROFILE");
         
         
         dataname1=cpl_sprintf("image_AOPEN_DATA%d_profile.fits",i+1);           
         if (cpl_dfs_save_image(frameset, NULL, parlist, frameset, NULL,  image_AOPEN_DATA_T[i],
                                CPL_BPP_IEEE_FLOAT, "midi_profile",
                                pro_list, NULL,
                                PACKAGE "/" PACKAGE_VERSION,
                                dataname1)) {
            /* Propagate the error */
            (void)cpl_error_set_where(cpl_func);
         }
         cpl_free(dataname1);      
         
         
         dataname1=cpl_sprintf("image_BOPEN_DATA%d_profile.fits",i+1);           
         if (cpl_dfs_save_image(frameset, NULL, parlist, frameset, NULL,  image_BOPEN_DATA_T[i],
                                CPL_BPP_IEEE_FLOAT, "midi_profile",
                                pro_list, NULL,
                                PACKAGE "/" PACKAGE_VERSION,
                                dataname1)) {
            /* Propagate the error */
            (void)cpl_error_set_where(cpl_func);
         }
/*          cpl_free(dataname); */
         cpl_free(dataname1);      
         
      }
   }


/* Free the memory */
/*    cpl_propertylist_delete(qclist_all_extension); */
   cpl_propertylist_delete(pro_list);


   for (i=0; i<dimenDATA;i++){
      cpl_image_delete(image_AOPEN_DATA_T[i]);
      cpl_image_delete(image_BOPEN_DATA_T[i]);
      cpl_image_delete(image_AOPEN_DATA_T_float[i]);
      cpl_image_delete(image_BOPEN_DATA_T_float[i]);
   }
   





/* /\* Save the image lists for diagnostics *\/ */
/*    if(0){ */
/*       for (i=0; i<dimenDATA;i++){ */
/*          if (strcmp(gris_name,"GRISM")==0 && isSP>0) dataname=cpl_sprintf("KAPPA_SP_MASK_GRISM_DIAGNOSTIC"); */
/*          if (strcmp(gris_name,"PRISM")==0 && isSP>0) dataname=cpl_sprintf("KAPPA_SP_MASK_PRISM_DIAGNOSTIC"); */
/*          if (strcmp(gris_name,"GRISM")==0 && isHS>0) dataname=cpl_sprintf("KAPPA_HS_MASK_GRISM_DIAGNOSTIC"); */
/*          if (strcmp(gris_name,"PRISM")==0 && isHS>0) dataname=cpl_sprintf("KAPPA_HS_MASK_PRISM_DIAGNOSTIC"); */

           
/*          dataname1=cpl_sprintf("imglst_AOPEN_DATA%d.fits",i+1);            */
/*          if (cpl_dfs_save_imagelist(frameset, parlist, frameset, imglst_AOPEN_DATA_T[i], */
/*                                     CPL_BPP_IEEE_FLOAT, "midi_profile", */
/*                                     dataname, NULL, NULL, */
/*                                     PACKAGE "/" PACKAGE_VERSION, */
/*                                     dataname1)) { */
/*             /\* Propagate the error *\/ */
/*             (void)cpl_error_set_where(cpl_func); */
/*          } */
/*          cpl_free(dataname1);       */


/*          dataname1=cpl_sprintf("imglst_BOPEN_DATA%d.fits",i+1);            */
/*          if (cpl_dfs_save_imagelist(frameset, parlist, frameset, imglst_BOPEN_DATA_T[i], */
/*                                     CPL_BPP_IEEE_FLOAT, "midi_profile", */
/*                                     dataname, NULL, NULL, */
/*                                     PACKAGE "/" PACKAGE_VERSION, */
/*                                     dataname1)) { */
/*             /\* Propagate the error *\/ */
/*             (void)cpl_error_set_where(cpl_func); */
/*          } */
/*          cpl_free(dataname); */
/*          cpl_free(dataname1);       */
      
      


/*       } */

/*    } */


/* Free the memory */
   for (i=0; i<dimenDATA;i++){
      cpl_imagelist_delete(imglst_AOPEN_DATA_T[i]);
      cpl_imagelist_delete(imglst_AOPEN_DATA_S[i]);
      
      cpl_imagelist_delete(imglst_BOPEN_DATA_T[i]);
      cpl_imagelist_delete(imglst_BOPEN_DATA_S[i]);
      
   }
   



   /* Free memory */

   
   return (int)cpl_error_get_code();
}



static int table_to_imglst_sky_target(const char * columname,
                    const char * columntype,
                    cpl_imagelist * imglst_sky,
                    cpl_imagelist * imglst_target,
                    cpl_table * table)
{

   int dimenDATA; 
   int   i,csky=0, ctarget=0,csky_tmp=0, ctarget_tmp=0,first_unknown=0;
   cpl_array * array_data=NULL;
   cpl_image * image_data_int=NULL;
   cpl_errorstate          prestate = cpl_errorstate_get();
   char ** target_type; 	
   cpl_type ctype;   
   int   flag_issky=0;
   int   flag_istarget=0;

   cpl_imagelist * imglst_sky_tmp=NULL;
   cpl_imagelist * imglst_target_tmp=NULL;

   imglst_sky_tmp=cpl_imagelist_new();
   imglst_target_tmp=cpl_imagelist_new();

   /* Load extension  Imaging Data  */ 
   
   dimenDATA=cpl_table_get_column_dimensions(table, columname);
   cpl_ensure_code(cpl_errorstate_is_equal(prestate), cpl_error_get_code());
   if (dimenDATA != 2) {
      return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
                                        "DATA has a wrong dimension");
   }
   /* Read target type T(arget),S(ky),B(ackground) and store them in the image list */
   /* Loop over all "images" stored in DATA and add them to the imagelist */
   
   
   if (cpl_table_has_column(table, columntype))
   {
      target_type=cpl_table_get_data_string(table, columntype);
      
   }
   else 
   {
      return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
                                        "TYPE of the Column not found");
   }
/*    cpl_msg_info(cpl_func, "Number of images in the Table: %d",cpl_table_get_nrow(table)); */
/*    cpl_msg_info(cpl_func, "Extracting %d images: please wait ...",cpl_table_get_nrow(table)); */
   
   
   ctype=cpl_table_get_column_type(table, columname);   
/*    cpl_msg_info(cpl_func, "Type of the table column: %d",ctype); */



   csky=cpl_imagelist_get_size(imglst_sky);
   
   ctarget=cpl_imagelist_get_size(imglst_target);
   
/*    cpl_msg_info(cpl_func, "Length of imagelist S T : %d %d",csky,ctarget ); */



   flag_issky=0;
   flag_istarget=0;

/* Find the first "U" image and discard the images befor  */
   
   for (i=0; i<cpl_table_get_nrow(table);i++){
      if(strcmp(target_type[i],"U")== 0){
         first_unknown=i;
         break;
      }
   }

/*    cpl_msg_info(cpl_func, "First unknown: %d",first_unknown); */
   
   for (i=first_unknown; i<cpl_table_get_nrow(table);i++){
      
      array_data=(cpl_array *)cpl_table_get_array(table,columname, i); 
      if(ctype&CPL_TYPE_INT){
         image_data_int=cpl_image_wrap_int(cpl_table_get_column_dimension(table,columname,0 ), cpl_table_get_column_dimension(table,columname,1 ), cpl_array_get_data_int(array_data) );
      }
      if(ctype&CPL_TYPE_FLOAT){
         image_data_int=cpl_image_wrap_float(cpl_table_get_column_dimension(table,columname,0 ), cpl_table_get_column_dimension(table,columname,1 ), cpl_array_get_data_float(array_data) );
      }
      /* Cast the image to float */
/*       image_data=cpl_image_cast(image_data_int, CPL_TYPE_FLOAT); */
      
/*        Append the image to the imagelists */
      
      if(strcmp(target_type[i],"S")== 0){
         cpl_imagelist_set(imglst_sky_tmp,cpl_image_cast(image_data_int, CPL_TYPE_FLOAT),csky_tmp++);
         flag_issky=1;
         cpl_image_unwrap(image_data_int);
         continue;
      }
      if(strcmp(target_type[i],"T")== 0){
         cpl_imagelist_set(imglst_target_tmp,cpl_image_cast(image_data_int, CPL_TYPE_FLOAT),ctarget_tmp++);
         flag_istarget=1;
         cpl_image_unwrap(image_data_int);
         continue;
      }

      /*Unwrap processed image manualy if nor Target nor Sky*/
      if(image_data_int!=NULL){
         cpl_image_unwrap(image_data_int);
      }
      if(flag_issky >0 && flag_istarget >0){
         /*Collapse the temporary imagelist and add the image to the "final" imagelist */
         cpl_imagelist_set(imglst_sky,cpl_imagelist_collapse_create(imglst_sky_tmp),csky++);
         cpl_imagelist_set(imglst_target,cpl_imagelist_collapse_create(imglst_target_tmp),ctarget++);
         csky_tmp=0;
         ctarget_tmp=0;
         flag_issky=0;
         flag_istarget=0;

         /*Empty imagelist and free the image buffer*/
         while(cpl_imagelist_get_size(imglst_sky_tmp)>0){
            cpl_image_delete(cpl_imagelist_unset(imglst_sky_tmp,0));
         }
         while(cpl_imagelist_get_size(imglst_target_tmp)>0){
            cpl_image_delete(cpl_imagelist_unset(imglst_target_tmp,0));
         }

      }
      
   }
 
   
/*          Empty imagelist and free the image buffer*/
   while(cpl_imagelist_get_size(imglst_sky_tmp)>0){
      cpl_image_delete(cpl_imagelist_unset(imglst_sky_tmp,0));
   }
   while(cpl_imagelist_get_size(imglst_target_tmp)>0){
      cpl_image_delete(cpl_imagelist_unset(imglst_target_tmp,0));
   }
   
   cpl_imagelist_delete(imglst_sky_tmp);
   cpl_imagelist_delete(imglst_target_tmp);
   
   
   
   return (int)cpl_error_get_code();
}


/*----------------------------------------------------------------------------*/
/**
  @brief    Normalizes an image to unity in the y-direction
  @param    image   the image to normalize
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/

int midi_normalize_image_y(cpl_image * image)
   {

   int         nx_image=0;
   int         ny_image=0;
   int         xpos=0; 
   int         ypos=0;

   double *    pimage=NULL;
   double *    pimage_norm=NULL;

   cpl_image * image_norm;
   
   cpl_msg_info(cpl_func, "Normalizing the integral of the profile to unity ...");         

   image_norm=cpl_image_collapse_create(image,0);
   
   nx_image=cpl_image_get_size_x(image);
   ny_image=cpl_image_get_size_y(image);
   
   pimage     = cpl_image_get_data_double(image);
   pimage_norm= cpl_image_get_data_double(image_norm);

   
   for (xpos=0; xpos<nx_image; ++xpos)
   {
      for(ypos=0; ypos<ny_image;++ypos)
      {
         if (pimage_norm[xpos]>0)  pimage[xpos+nx_image*ypos]  /= pimage_norm[xpos];
         if (pimage_norm[xpos]<=0) pimage[xpos+nx_image*ypos] = 0.;
      }
   }
   
   
   cpl_image_delete(image_norm);
   
   return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Inserts an image into a table
  @param    table       the table
  @param    columname   the name of the column
  @param    image       the image to insert
  @param    raw         the insert raw 
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/

static int append_image_to_table(cpl_table * table, const char * columname, cpl_image * image, int row)
{
   
   cpl_array        * array_dimension=NULL;   
   cpl_array        * array_dummy=NULL;
   
   array_dimension=cpl_array_new(2,CPL_TYPE_INT);
   cpl_array_set(array_dimension, 0,cpl_image_get_size_x(image));
   cpl_array_set(array_dimension, 1,cpl_image_get_size_y(image));
   
   cpl_table_new_column_array(table, columname, CPL_TYPE_FLOAT, cpl_image_get_size_x(image)*cpl_image_get_size_y(image));
   cpl_table_set_column_dimensions(table,columname,array_dimension);
   array_dummy = cpl_array_wrap_float(cpl_image_get_data_float(image), cpl_image_get_size_x(image)*cpl_image_get_size_y(image));    
   cpl_table_set_array(table, columname, row, array_dummy);
   cpl_array_unwrap(array_dummy);
   
   
   cpl_array_delete(array_dimension);
   
   return 0;  
}

/*----------------------------------------------------------------------------*/
/**
  @brief    copy/appends an extension of infile to outfile
  @param    infile         the filename with the extension
  @param    outfile        the filename to which the extension should be copied 
  @param    extension_name the name of the extension
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/

static int midi_copy_extension(const char * infile , const char * outfile, const char * extension_name)   
{
   fitsfile * fptrin=NULL;
   fitsfile * fptrout=NULL;
   int status=0;
   
   fits_open_diskfile(&fptrin, infile, READONLY, &status);
   fits_open_diskfile(&fptrout, outfile, READWRITE, &status);
   fits_movnam_hdu(fptrin, ANY_HDU, (char *)extension_name, 0, &status);
   fits_copy_hdu(fptrin, fptrout, 0, &status);
   fits_close_file(fptrin, &status);
   fits_close_file(fptrout, &status);
   
   if  (status != 0){
      cpl_msg_error(cpl_func,"A problem occured while copying the EXTENSION: %s", extension_name);      
      return 1;
   }
   else{
      return 0;
   }
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Calculates the y-barycenter and writes corresponding QC parameters.
  @param    pro_list    property list where the QC parameters are appended
  @param    index       corresponding QC Y index for DATA1/2/3/4, respectively
  @param    image       image where the barycenter is calculated
  @param    x_start     starting x-coordinate where QCs are calculated
  @param    x_end       ending   x-coordinate where QCs are calculated
  @return   NONE

 */
/*----------------------------------------------------------------------------*/
static void midi_profile_qc_barycenter(cpl_propertylist * pro_list, int index,
		cpl_image * image, const int * x_start, const int * x_end)
{
	char * qcname=NULL;
	int i=0;
	int image_size_y=0;

	image_size_y=cpl_image_get_size_y(image);
	for (i = 0; i < 5; i++){

		qcname=cpl_sprintf("ESO QC Y%d CENT BINNED%d",index+1,i+1);
		cpl_propertylist_update_double(pro_list, qcname,
				cpl_image_get_centroid_y_window(image, x_start[i], 1,
						x_end[i], image_size_y));
		cpl_free(qcname);
	}
	return;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Calculates the maximum and writes corresponding QC parameters.
  @param    pro_list    property list where the QC parameters are appended
  @param    index       corresponding QC Y index for DATA1/2/3/4, respectively
  @param    image       image where the maximum is calculated
  @param    x_start     starting x-coordinate where QCs are calculated
  @param    x_end       ending   x-coordinate where QCs are calculated
  @return   NONE

 */
/*----------------------------------------------------------------------------*/
static void midi_profile_qc_maximum(cpl_propertylist * pro_list, int index,
		cpl_image * image, const int * x_start, const int * x_end)
{
	char * qcname=NULL;
	char * qcname_collapsed=NULL;
	int i=0;
	int image_size_y=0;
	cpl_image * image_collapsed=NULL;

	image_size_y=cpl_image_get_size_y(image);

	for (i = 0; i < 5; i++){
		image_collapsed=cpl_image_collapse_window_create(image, x_start[i], 1,
				x_end[i], image_size_y,1);
		qcname=cpl_sprintf("ESO QC Y%d MAX BIN%d",index+1,i+1);
		qcname_collapsed=cpl_sprintf("ESO QC Y%d MAXSUM BIN%d",index+1,i+1);

		cpl_propertylist_update_double(pro_list, qcname,
				cpl_image_get_max_window(image, x_start[i], 1,
						x_end[i], image_size_y));
		cpl_propertylist_update_double(pro_list, qcname_collapsed,
				cpl_image_get_max_window(image_collapsed, 1, 1,
						1, image_size_y));


		cpl_free(qcname);
		cpl_free(qcname_collapsed);
		cpl_image_delete(image_collapsed);
	}
	return;
}

/**@}*/
